(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DepGraph
imports Graph Edg SemanticLabeling
begin 

fun
  check_dep_graph_proc ::
    "('f::show,'v::show)rules \<Rightarrow> (bool\<times>('f,'v)rules)list
     \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rule_map \<Rightarrow> shows check"
where
  "check_dep_graph_proc P dps rm rrm = do {
     check_subseteq P (concat (map snd dps))
       <+? (\<lambda>dp. ''Dependency Pair '' +#+ shows_rule dp +@+ '' is missing in decomposition''
         +#+ shows_nl);
     check_graph_decomp
       (shows \<circ> fst)
       (\<lambda>(((_, t), (_, ct)), ((u, _), (cu, _))). gcMatchI ct u \<and> gcMatchI cu t)
       (map (\<lambda>(real, Cs). (real, (map (\<lambda>(s, t). ((s, t), (capRM rrm s, capRM rm t))) Cs))) dps)
       <+? (\<lambda>s. ''our estimation (EDG***) could not show that you have a valid decomposition ''
         +#+ ''due to the following reason'' +#+ shows_nl +@+ s)
   }"

lemma check_dep_graph_proc_sound[simp]: 
  assumes check: "isOK (check_dep_graph_proc P edps (computeRuleMap R) (computeRuleMap (reverse_trs R)))" 
    and chain: "ichain (set P, set R) s t \<sigma> "
  shows "\<exists> edp \<in> set edps. fst edp \<and> (\<exists> i. ichain (set (snd edp), set R) (shift s i) (shift t i) (shift \<sigma> i))"
proof -
  let ?rm = "computeRuleMap R"
  let ?rrm = "computeRuleMap (reverse_trs R)"
  from check 
  have subset: "set P \<subseteq> set (concat (map snd edps))" (is "_ \<subseteq> ?dps") and scc: "isOK (check_graph_decomp (sshow \<circ> fst) (\<lambda> (((_,t),(_,ct)),((u,_),(cu,_))). gcMatchI ct u \<and> gcMatchI cu t) 
    (map (\<lambda> (real,Cs). (real, (map (\<lambda> (s,t). ((s,t),(capRM ?rrm s, capRM ?rm t))) Cs))) edps))" (is "isOK(check_graph_decomp ?sshow ?echeck (map ?conv edps))")
    by (auto)
  let ?quads = "map ?conv edps"
  from chain have
    inP:  "\<forall>i. (s i, t i) \<in> set P" and 
    inBetween: "\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (set R))^* " by auto 
  let ?p = "\<lambda> i. ((s i, t i), (capRM ?rrm (s i), capRM ?rm (t i)))"
  let ?ps = "\<lambda> i. (s i \<cdot> \<sigma> i, t i \<cdot> \<sigma> i)"
  fix whatever
  have one: "is_path (graph2 whatever ?echeck) ?p"
  proof (unfold edges_def graph2_def is_path_def, simp, clarify, unfold mem_def, simp)
    fix n
    from inBetween have steps: "(t n \<cdot> \<sigma> n, s (Suc n) \<cdot> \<sigma> (Suc n)) \<in> (rstep (set R))^* " ..
    show "gcMatch (capRM ?rm (t n)) (s (Suc n)) \<and> gcMatch (capRM ?rrm (s (Suc n))) (t n)" 
      by (rule conjI, rule gcMatchCapRMSound, rule steps, rule gcMatchCapRRMSound, rule steps)
  qed
  
  from inP subset have two: "path_nodes ?p \<subseteq> set (concat (map snd ?quads))" unfolding path_nodes_def by force
    thm check_graph_decomp_sound
  with one two scc 
  obtain q D where 
    suffix: "?p =\<dots> q" and 
    mem: "D \<in> set (map snd (filter fst ?quads))" and 
    ending: "path_nodes q \<subseteq> set D" 
    using check_graph_decomp_sound by blast
    
  from mem obtain C where C: "(True, C) \<in> set edps \<and> D = snd (?conv (True, C))" by force
  from suffix obtain n where "\<forall> m. (?p (m + n) = q m)" unfolding suffix_def by blast
  hence ichain: "ichain (set C, set R) (shift s n) (shift t n) (shift \<sigma> n)"
    using inP inBetween ending C unfolding path_nodes_def by force
  from ichain C have "\<exists> edp \<in> set edps. fst edp \<and> ichain (set (snd edp), set R) (shift s n) (shift t n) (shift \<sigma> n)" by force
  thus ?thesis by blast
qed

declare check_dep_graph_proc.simps[simp del]

lemma dep_graph_proc :
  assumes check: "isOK (check_dep_graph_proc P edps (computeRuleMap R) (computeRuleMap (reverse_trs R)))" 
  and rec: "\<forall> edp \<in> set edps. fst edp \<longrightarrow> finite_dpp (set (snd edp), set R)"
  shows "finite_dpp (set P, set R)"
unfolding finite_dpp_def
proof 
  assume "\<exists> s t \<sigma>. min_ichain (set P, set R) s t \<sigma>"
  from this obtain s t \<sigma> where mc: "min_ichain (set P, set R) s t \<sigma>" by auto
  hence "ichain (set P,set R) s t \<sigma>" by auto
  from check this have "\<exists> edp \<in> set edps. fst edp \<and> (\<exists> i. ichain (set (snd edp), set R) (shift s i) (shift t i) (shift \<sigma> i))" 
    by (rule check_dep_graph_proc_sound)
  from this obtain edp i where nearly: "edp \<in> set edps \<and> fst edp \<and> ichain (set (snd edp), set R) (shift s i) (shift t i) (shift \<sigma> i)" by blast 
  with rec have contra: "finite_dpp (set (snd edp), set R)" by auto
  from nearly mc have "min_ichain (set (snd edp), set R) (shift s i) (shift t i) (shift \<sigma> i)" by (auto simp: minimal_cond_def ci_subset_refl[of "set R"])
  with contra show False unfolding finite_dpp_def by blast
qed


lemma dep_graph_lab_proc :
  assumes check: "isOK (check_dep_graph_proc P edps (computeRuleMap R) (computeRuleMap (reverse_trs R)))" 
  and rec: "\<forall> edp \<in> set edps. fst edp \<longrightarrow> finite_lab_dpp (set (snd edp), set R) m"
  shows "finite_lab_dpp (set P, set R) m"
unfolding finite_lab_dpp_def
proof 
  assume "\<exists> s t \<sigma>. min_lab_ichain (set P, set R) m s t \<sigma>"
  from this obtain s t \<sigma> where mc: "min_lab_ichain (set P, set R) m s t \<sigma>" by auto
  hence "ichain (set P,set R) s t \<sigma>" by auto
  from check this have "\<exists> edp \<in> set edps. fst edp \<and> (\<exists> i. ichain (set (snd edp), set R) (shift s i) (shift t i) (shift \<sigma> i))" 
    by (rule check_dep_graph_proc_sound)
  from this obtain edp i where nearly: "edp \<in> set edps \<and> fst edp \<and> ichain (set (snd edp), set R) (shift s i) (shift t i) (shift \<sigma> i)" by blast 
  with rec have contra: "finite_lab_dpp (set (snd edp), set R) m " by auto
  from nearly mc have "min_lab_ichain (set (snd edp), set R) m (shift s i) (shift t i) (shift \<sigma> i)" by (auto simp: lab_minimal_cond_def ci_subset_refl[of "set R"])
  with contra show False unfolding finite_lab_dpp_def by blast
qed


end