theory EfficientLPO imports RPO begin

types ('i,'o)memory = "('i \<times> 'o)list"

types ('f,'v)termmemory = "((('f,'v)term \<times>('f,'v)term), (bool \<times> bool))memory"

definition store :: "('i,'o)memory \<Rightarrow> 'i \<Rightarrow> 'o \<Rightarrow> ('i,'o)memory"
where
  "store c k v = (k,v) # c"

fun anyNstriE :: "('b \<Rightarrow> 'a  \<Rightarrow> ((bool \<times> bool) \<times> 'b)) \<Rightarrow> 'b \<Rightarrow> 'a list \<Rightarrow> (bool \<times> 'b)"
where
  "anyNstriE f c [] = (False,c)" |
  "anyNstriE f c (s # ss) = (let ((stri,nstri),c') = f c s in 
     if nstri then (nstri,c') else anyNstriE f c' ss)"
 
lemma anyNstriE [fundef_cong] : 
  assumes "\<And> s c. s \<in> set ss \<Longrightarrow> f c s = g c s"
  shows "anyNstriE f c ss = anyNstriE g c ss"
using assms proof (induct ss arbitrary: c, simp)
  case (Cons s ss)
  show ?case 
  proof (cases "f c s")
    case (Pair res c')
    with Cons have rec: "anyNstriE f c' ss = anyNstriE g c' ss" by auto
    from Pair rec show ?thesis 
      by (simp add: Cons(2)[of s c, simplified], cases res, auto)    
  qed
qed

fun forAllStriE :: "('b \<Rightarrow> 'a  \<Rightarrow> ((bool \<times> bool) \<times> 'b)) \<Rightarrow> 'b \<Rightarrow> 'a list \<Rightarrow> (bool \<times> 'b)"
where
  "forAllStriE f c [] = (True,c)" |
  "forAllStriE f c (t # ts) = (let ((stri,nstri),c') = f c t in
     if stri then forAllStriE f c' ts else (False,c'))"

lemma forAllStriE [fundef_cong] : 
  assumes "\<And> t c. t \<in> set ts \<Longrightarrow> f c t = g c t"
  shows "forAllStriE f c ts = forAllStriE g c ts"
using assms proof (induct ts arbitrary: c, simp)
  case (Cons t ts)
  show ?case 
  proof (cases "f c t")
    case (Pair res c')
    with Cons have rec: "forAllStriE f c' ts = forAllStriE g c' ts" by auto
    from Pair rec show ?thesis 
      by (simp add: Cons(2)[of t c, simplified], cases res, auto)    
  qed
qed

fun lex_extE :: "('b \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> ((bool \<times> bool) \<times> 'b)) \<Rightarrow> 'b \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> ((bool \<times> bool) \<times> 'b)"
where
  "lex_extE f c [] [] = ((False,True),c)" |
  "lex_extE f c (_#_) [] = ((True,True),c)" |
  "lex_extE f c [] (_#_) = ((False,False),c)" |
  "lex_extE f c (a#as) (b#bs) = (let ((stri,nstri),c') = f c a b in
     if stri then ((True,True),c')
     else if nstri then lex_extE f c' as bs 
     else ((False,False),c'))"

lemma lex_extE_cong[fundef_cong] : fixes f g ss ts c
  assumes "\<And> s t c. \<lbrakk>s \<in> set ss; t \<in> set ts\<rbrakk> \<Longrightarrow> f c s t = g c s t" 
  shows "lex_extE f c ss ts = lex_extE g c ss ts" 
using assms 
proof (induct ss arbitrary: ts c)
  case Nil
  thus ?case by (cases ts, auto)
next
  case (Cons s ss) note oCons = this
  show ?case 
  proof (cases ts)
    case Nil
    with Cons show ?thesis by auto
  next
    case (Cons t ts')
    with oCons have id: "g c s t = f c s t" by simp
    show ?thesis
    proof (cases "f c s t")
      case (Pair res c')
      from oCons Cons have id2: "lex_extE f c' ss ts' = lex_extE g c' ss ts'" by auto
      show ?thesis
        by (simp add: Pair Cons id, cases res, auto simp: id2)
    qed
  qed
qed

function lpoE :: "('f,'v)termmemory \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> ((bool \<times> bool) \<times> ('f,'v)termmemory)"
where
  "lpoE c pr s t = (case lookup (s,t) c of
       Some (stri,nstri) \<Rightarrow> ((stri,nstri),c)
     | None \<Rightarrow> let (v,new_c) = (case (s,t) of
         ((Var x), (Var y)) \<Rightarrow> ((False, x=y),c)
       | ((Var x), (Fun g ts)) \<Rightarrow> ((False, ts=[] \<and> pr g 0 = 0),c)
       | ((Fun f ss), (Var y)) \<Rightarrow> let (con,c') = anyNstriE (\<lambda> lc s. lpoE lc pr s t) c ss in ((con,con), c')
       | ((Fun f ss), (Fun g ts)) \<Rightarrow> (let (ssNstri,c') = anyNstriE (\<lambda> lc s. lpoE lc pr s t) c ss in
           if ssNstri then ((True,True), c')
           else (let prl = pr f (length ss); prr = pr g (length ts) in
             (let (sAllStri,c'') = forAllStriE (\<lambda>lc. lpoE lc pr s) c' ts in
             if prl \<ge> prr \<and> sAllStri then
               if prl > prr then ((True,True), c'')
               else lex_extE (\<lambda>lc. lpoE lc pr) c'' ss ts
             else ((False,False), c'')
          )))
   )   in (v, store new_c (s,t) v)
         )"
by pat_completeness auto

termination lpoE
proof (relation "measures [\<lambda> (c,pr,s,t). size s + size t]", (auto)[4])
  fix "pr" s' t' c a b f ss g ts x xa y xb xc xd xe ya s t ca
  assume ab: "(s',t') = (a,b)" and a: "a = Fun f ss" and b: "b = Fun g ts" and s: "s \<in> set ss" and t: "t \<in> set ts"
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 ab a b show "((ca,pr,s,t), c, pr, s', t') \<in> measures [\<lambda>(c, pr, s, t). size s + size t]" by simp
qed

definition lpo_E
where "lpo_E pr s t = fst(lpoE [] pr s t)"

definition valid_term_order_memory where 
  "valid_term_order_memory f c \<equiv> \<forall>s t v. ((s,t),v) \<in> set c \<longrightarrow> f s t = v"

definition valid_lpo_memory where "valid_lpo_memory pr c \<equiv> valid_term_order_memory (lpo_unbounded pr) c"

declare valid_term_order_memory_def[simp]
declare valid_lpo_memory_def[simp]

lemma empty_memory_is_valid_lpo_memory : "valid_lpo_memory pr [] " by auto

lemma lpoE_memoryd_sound: 
  assumes "valid_lpo_memory pr c \<and> lookup (s,t) c = Some v"
  shows "fst(lpoE c pr s t) = lpo_unbounded pr s t \<and> valid_lpo_memory pr (snd (lpoE c pr s t))"
proof (cases v)
  case (Pair stri nstri)
  from assms have Some: "lookup (s,t) c = Some v" ..
  from Pair assms have "snd(lpoE c pr s t) = c" by simp
  from this assms have valid_memory: "valid_lpo_memory pr (snd(lpoE c pr s t))" by simp
  from Pair assms lookup_sound[OF Some] have correct_value: "(stri,nstri) = lpo_unbounded pr s t" by simp
  from assms Pair have "fst(lpoE c pr s t) = (stri,nstri)" by simp
  from this correct_value have "fst(lpoE c pr s t) = lpo_unbounded pr s t" by simp
  from this valid_memory show ?thesis by simp
qed

lemma anyNstriE_sound:
  assumes "\<And>s c. s \<in> set ss \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s) = f s t \<and> valid_term_order_memory f (snd (fE c s))"
  and "valid_term_order_memory f c"
  shows "valid_term_order_memory f (snd (anyNstriE fE c ss)) \<and> fst (anyNstriE fE c ss) = list_ex (\<lambda>s. snd (f s t)) ss"
using assms proof (induct ss arbitrary: c)
  case Nil
  hence "valid_term_order_memory f c" by simp
  moreover have "anyNstriE fE c [] = (False,c)" by simp
  moreover have "list_ex (\<lambda>s. snd (f s t)) [] = False" by simp
  ultimately show ?case by simp
next
  case (Cons x xs)
  thus ?case proof (cases "fE c x")
    case (Pair v c') note fE_val=this
    hence c':"snd (fE c x) = c'" by simp
    have "x \<in> set (x # xs)" by simp
    from this Cons have x_vals_sound:"fst (fE c x) = f x t \<and> valid_term_order_memory f (snd (fE c x))" by blast
    from this c' have c'_valid:"valid_term_order_memory f c'" by (simp only: assms)
    show ?thesis proof (cases v) 
      case (Pair stri nstri) 
      from x_vals_sound have "fst (fE c x) = f x t" by simp
      from this fE_val Pair have f_val:"f x t = (stri,nstri)" by simp
      show ?thesis proof (cases nstri)
        case True
        from True Pair fE_val have "snd (anyNstriE fE c (x # xs) ) = c'" by simp
        from c'_valid this have valid_memory:"valid_term_order_memory f (snd (anyNstriE fE c (x # xs)))" by simp
        from True Pair fE_val have anyNstriE_val:"fst (anyNstriE fE c (x # xs)) = True" by simp
        from this True f_val have any_val:"list_ex (\<lambda>s. snd (f s t)) (x # xs) = True" by simp
        from any_val anyNstriE_val valid_memory show ?thesis by simp
      next
        case False
        from False Pair fE_val have anyNstriE_val:"anyNstriE fE c (x # xs) = anyNstriE fE c' xs" by simp
        have "\<And>s c. s \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s) = f s t \<and> valid_term_order_memory f (snd (fE c s ))" proof
          fix s c
          have "s \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> s \<in> set (x # xs) \<and> valid_term_order_memory f c" by simp     
          moreover from Cons(2) 
          have "s \<in> set (x # xs) \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s) = f s t \<and> valid_term_order_memory f (snd (fE c s))" by blast
          ultimately show "s \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s) = f s t" 
            and "s \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> valid_term_order_memory f (snd (fE c s))" by (blast, simp)
        qed
        from this Cons(1)[of c'] c'_valid 
        have rec:"valid_term_order_memory f (snd (anyNstriE fE c' xs)) \<and> fst (anyNstriE fE c' xs) = list_ex (\<lambda>s. snd (f s t)) xs" by blast
        from rec anyNstriE_val have valid_memory: "valid_term_order_memory f (snd (anyNstriE fE c (x # xs)))" by simp
        from f_val False have any_val:"list_ex (\<lambda>s. snd (f s t)) (x # xs) = list_ex (\<lambda>s. snd (f s t)) xs" by simp
        from rec anyNstriE_val any_val valid_memory show ?thesis by simp 
      qed
    qed
  qed
qed

lemma forAllStriE_sound:
  assumes "\<And>t c. t \<in> set ts \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c t) = f s t \<and> valid_term_order_memory f (snd (fE c t))"
  and "valid_term_order_memory f c"
  shows "valid_term_order_memory f (snd (forAllStriE fE c ts)) \<and> fst (forAllStriE fE c ts) = list_all (\<lambda>t. fst (f s t)) ts"
using assms proof (induct ts arbitrary: c)
  case Nil
  hence "valid_term_order_memory f c" by simp
  moreover have "forAllStriE fE c [] = (True,c)" by simp
  moreover have "list_all (\<lambda>t. fst (f s t)) [] = True" by simp
  ultimately show ?case by simp
next
  case (Cons x xs)
  thus ?case proof (cases "fE c x")
    case (Pair v c') note fE_val=this
    hence c':"snd (fE c x) = c'" by simp
    have "x \<in> set (x # xs)" by simp
    from this Cons have x_vals_sound:"fst (fE c x) = f s x \<and> valid_term_order_memory f (snd (fE c x))" by blast
    from this c' have c'_valid:"valid_term_order_memory f c'" by (simp only: assms)
    show ?thesis proof (cases v) 
      case (Pair stri nstri) 
      from x_vals_sound have "fst (fE c x) = f s x" by simp
      from this fE_val Pair have f_val:"f s x = (stri,nstri)" by simp
      show ?thesis proof (cases stri)
        case False
        from False Pair fE_val have "snd (forAllStriE fE c (x # xs) ) = c'" by simp
        from c'_valid this have valid_memory:"valid_term_order_memory f (snd (forAllStriE fE c (x # xs)))" by simp
        from False Pair fE_val have forAllStriE_val:"fst (forAllStriE fE c (x # xs)) = False" by simp
        from this False f_val have forall_val:"list_all (\<lambda>t. fst (f s t)) (x # xs) = False" by simp
        from forall_val forAllStriE_val valid_memory show ?thesis by simp
      next
        case True
        from True Pair fE_val have forAllStriE_val:"forAllStriE fE c (x # xs) = forAllStriE fE c' xs" by simp
        have "\<And>t c. t \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c t) = f s t \<and> valid_term_order_memory f (snd (fE c t))" proof
          fix t c
          have "t \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> t \<in> set (x # xs) \<and> valid_term_order_memory f c" by simp     
          moreover from Cons(2) 
          have "t \<in> set (x # xs) \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c t) = f s t \<and> valid_term_order_memory f (snd (fE c t))" by blast
          ultimately show "t \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c t) = f s t" 
            and "t \<in> set xs \<and> valid_term_order_memory f c \<Longrightarrow> valid_term_order_memory f (snd (fE c t))" by (blast, simp)
        qed
        from this Cons(1)[of c'] c'_valid 
        have rec:"valid_term_order_memory f (snd (forAllStriE fE c' xs)) \<and> fst (forAllStriE fE c' xs) = list_all (\<lambda>t. fst (f s t)) xs" by blast
        from rec forAllStriE_val have valid_memory: "valid_term_order_memory f (snd (forAllStriE fE c (x # xs)))" by simp
        from f_val True have forall_val:"list_all (\<lambda>t. fst (f s t)) (x # xs) = list_all (\<lambda>t. fst (f s t)) xs" by simp
        from rec forAllStriE_val forall_val valid_memory show ?thesis by simp 
      qed
    qed
  qed
qed

declare lex_ext_unbounded.simps[simp]

lemma lex_extE_sound:
  assumes "\<And>t s c.  s \<in> set ss \<and> t \<in> set ts \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s t) = f s t \<and> valid_term_order_memory f (snd (fE c s t))"
  and "valid_term_order_memory f c"
  shows "valid_term_order_memory f (snd (lex_extE fE c ss ts)) \<and> fst (lex_extE fE c ss ts) = lex_ext_unbounded f ss ts"
using assms proof (induct ss ts arbitrary: c rule: List.list_induct2')
  case (1 c)
  from this have lex_extE_val:"lex_extE fE c [] [] = ((False,True),c)" by simp
  from Nil have lex_ext_unbounded_val:"lex_ext_unbounded f [] [] = (False, True)" by simp
  from 1(2) lex_extE_val lex_ext_unbounded_val show ?case by simp
next
  case (2 a as c)
  from this have lex_extE_val:"lex_extE fE c (a # as) [] = ((True,True),c)" by simp
  from Nil have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) [] = (True, True)" by simp
  from 2(2) lex_extE_val lex_ext_unbounded_val show ?case by simp
next
  case (3 b bs c)
  from this have lex_extE_val:"lex_extE fE c [] (b # bs)  = ((False,False),c)" by simp
  from Cons have lex_ext_unbounded_val:"lex_ext_unbounded f [] (b # bs)  = (False, False)" by simp
  from 3(2) lex_extE_val lex_ext_unbounded_val show ?case by simp
next
  case (4 a as b bs c)
  show ?case proof (cases "fE c a b")
    case (Pair v c') note oPair=this
    from 4(2)[of a b c] 4(3) have fE_sound:"fst (fE c a b) = f a b \<and> valid_term_order_memory f (snd (fE c a b))" by (simp only: assms, auto)
    from this Pair have c'_valid: "valid_term_order_memory f c'" by simp
    show ?thesis proof (cases v)
      case (Pair stri nstri)
      from fE_sound Pair oPair have f_val:"f a b = (stri,nstri)" by simp
      show ?thesis proof (cases stri)
        case True
        from oPair Pair True have lex_extE_val:"lex_extE fE c (a # as) (b # bs) = ((True,True),c')" by simp
        from f_val True have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) (b # bs) = (True,True)" by simp 
        from c'_valid lex_extE_val lex_ext_unbounded_val show ?thesis by simp
      next
        case False note striFalse = this
        show ?thesis proof (cases nstri)
          case True
          have "\<And>s t c. s \<in> set as \<and> t \<in> set bs \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s t) = f s t \<and> valid_term_order_memory f (snd (fE c s t))" proof
            fix t c s
            have "s \<in> set as \<and> t \<in> set bs \<and> valid_term_order_memory f c \<Longrightarrow> s \<in> set (a # as) \<and> t \<in> set (b # bs) \<and> valid_term_order_memory f c" by simp     
            from this 4(2)[of s t c] show "s \<in> set as \<and> t \<in> set bs \<and> valid_term_order_memory f c \<Longrightarrow> fst (fE c s t) = f s t" 
              and "s \<in> set as \<and> t \<in> set bs \<and> valid_term_order_memory f c \<Longrightarrow> valid_term_order_memory f (snd (fE c s t))" by (blast, simp)
          qed
          from this 4(1)[of c']  c'_valid have rec:
            "valid_term_order_memory f (snd (lex_extE fE c' as bs)) \<and> fst (lex_extE fE c' as bs) = lex_ext_unbounded f as bs" by (simp only: assms, blast)
          from striFalse True Pair oPair have lex_extE_val:"lex_extE fE c (a # as) (b # bs) = lex_extE fE c' as bs" by simp
          from striFalse True f_val have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) (b # bs) = lex_ext_unbounded f as bs" by simp
          from lex_extE_val lex_ext_unbounded_val rec show ?thesis by simp
        next
          case False
          from striFalse False Pair oPair have lex_extE_val:"lex_extE fE c (a # as) (b # bs) = ((False,False),c')" by simp
          from striFalse False f_val have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) (b # bs) = (False,False)" by simp
          from lex_extE_val lex_ext_unbounded_val c'_valid show ?thesis by simp
        qed
      qed
    qed
  qed
qed

declare lex_ext_unbounded.simps[simp del]


lemma lpoE_sound :
assumes "valid_lpo_memory pr c"
shows "fst(lpoE c pr s t) = lpo_unbounded pr s t \<and> valid_lpo_memory pr (snd (lpoE c pr s t))"
using assms proof (induct s arbitrary: c t rule: term_induct)
  case (Var x) note oVar=this
  show ?case proof (cases "lookup ((Var x), t) c")
    case (Some v)
    from this Var lpoE_memoryd_sound[of "pr" c "(Var x)" t v] show ?thesis by simp
  next
    case None
    show ?thesis proof (cases t)
      case (Var y)
      from Var None have lpoE_val:"fst(lpoE c pr (Var x) t) = (False,x=y)" by simp
      from Var None have "(False,x=y) = lpo_unbounded pr (Var x) t" by simp
      from lpoE_val this have correct_val: "fst(lpoE c pr (Var x) t) = lpo_unbounded pr (Var x) t" by simp
      from Var None have "snd(lpoE c pr (Var x) t) = store c ((Var x),t) (False,x=y)" by simp
      from oVar Var None this have "valid_lpo_memory pr (snd (lpoE c pr (Var x) t))" unfolding store_def by simp
      with correct_val show ?thesis by simp
    next
      case (Fun g ts)
      from Fun None have lpoE_val:"fst(lpoE c pr (Var x) t) = (False, ts=[] \<and> pr g 0 = 0)" by simp
      from Fun None have "(False,ts=[] \<and> pr g 0 = 0) = lpo_unbounded pr (Var x) t" by simp
      from this lpoE_val have correct_val: "fst(lpoE c pr (Var x) t) = lpo_unbounded pr (Var x) t" by simp
      from Fun None have "snd(lpoE c pr (Var x) t) = store c ((Var x), t) (False, ts=[] \<and> pr g 0 = 0)" by simp
      from Fun None this oVar have "valid_lpo_memory pr (snd (lpoE c pr (Var x) t))" unfolding store_def by simp
      with correct_val show ?thesis by simp
    qed
  qed
next
  case (Fun f ss) note oFun = this
  thus ?case proof (induct t arbitrary: c)
    case (Var y)
    show ?case proof (cases "lookup ((Fun f ss), (Var y)) c")
      case (Some v)
      from this Var(2) lpoE_memoryd_sound[of "pr" c "(Fun f ss)" "(Var y)" v] show ?thesis by simp
    next
      case None
      have "\<And>s c. s \<in> set ss \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr s (Var y)) = lpo_unbounded pr s (Var y) 
        \<and> valid_lpo_memory pr (snd (lpoE c pr s (Var y)))" proof 
        fix s c
        from Var(1)[of s c] show "s \<in> set ss \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr s (Var y)) = lpo_unbounded pr s (Var y)"
          and "s \<in> set ss \<and> valid_lpo_memory pr c \<Longrightarrow> valid_lpo_memory pr (snd (lpoE c pr s (Var y)))" by auto
      qed
      from anyNstriE_sound[of ss "lpo_unbounded pr" "(\<lambda>lc s. lpoE lc pr s (Var y))" "(Var y)" c] this Var(2) 
      have anyNstriE:"valid_lpo_memory pr (snd (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss)) \<and>
        fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss) = list_ex (\<lambda>s. snd (lpo_unbounded pr s (Var y))) ss" by (simp only: valid_lpo_memory_def, blast)
      from this have any:"fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss) = list_ex (\<lambda>s. snd (lpo_unbounded pr s (Var y))) ss" by simp
      have lpoE_val:
        "fst (lpoE c pr (Fun f ss) (Var y)) = (fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss),fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss))" 
        by (simp only: lpoE.simps[of c "pr" "(Fun f ss)" "(Var y)"], simp del: lpoE.simps add: None, simp add: split_def)
      have lpo_val: 
        "lpo_unbounded pr (Fun f ss) (Var y) = (fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss), fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss))" 
        by (simp only: "lpo_unbounded.simps", simp only: Let_def, simp only: any)
      from lpoE_val lpo_val have correct_value: "fst (lpoE c pr (Fun f ss) (Var y)) = lpo_unbounded pr (Fun f ss) (Var y)" by simp
      have lpoE_memory: 
        "snd (lpoE c pr (Fun f ss) (Var y)) = 
        store (snd (anyNstriE (\<lambda>lc s. lpoE lc pr s (Var y)) c ss)) (Fun f ss, Var y) (lpo_unbounded pr (Fun f ss) (Var y))" 
        by (simp only: lpoE.simps[of c "pr" "(Fun f ss)" "(Var y)"], 
          simp del: lpo_unbounded.simps lpoE.simps add: None, simp del: lpoE.simps lpo_unbounded.simps add:split_def lpo_val)
      from this anyNstriE have "valid_lpo_memory pr (snd (lpoE c pr (Fun f ss) (Var y)))" unfolding store_def by simp 
      from this correct_value show ?thesis by simp
    qed
  next 
    case (Fun g ts)
    show ?case proof (cases "lookup ((Fun f ss), (Fun g ts)) c")
      case (Some v)
      from this Fun(3) lpoE_memoryd_sound[of "pr" c "(Fun f ss)" "(Fun g ts)" v] show ?thesis by simp
    next
      case None
      have "\<And>s c. s \<in> set ss \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr s (Fun g ts)) = lpo_unbounded pr s (Fun g ts) 
        \<and> valid_lpo_memory pr (snd (lpoE c pr s (Fun g  ts)))" proof 
        fix s c
        from Fun(2)[of s c] show "s \<in> set ss \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr s (Fun g ts)) = lpo_unbounded pr s (Fun g ts)"
          and "s \<in> set ss \<and> valid_lpo_memory pr c \<Longrightarrow> valid_lpo_memory pr (snd (lpoE c pr s (Fun g ts)))" by auto
      qed
      from anyNstriE_sound[of ss "lpo_unbounded pr" "(\<lambda>lc s. lpoE lc pr s (Fun g ts))" "(Fun g ts)" c] this Fun(3) 
      have anyNstriE:"valid_lpo_memory pr (snd (anyNstriE (\<lambda>lc s. lpoE lc pr s (Fun g ts)) c ss)) \<and>
        fst (anyNstriE (\<lambda>lc s. lpoE lc pr s (Fun g ts)) c ss) = list_ex (\<lambda>s. snd (lpo_unbounded pr s (Fun g ts))) ss" by (simp only:valid_lpo_memory_def, blast)
      show ?thesis proof (cases "anyNstriE (\<lambda>lc s. lpoE lc pr s (Fun g ts)) c ss")
        case (Pair ssNstri c') note anyNstriE_Pair=this
        from this anyNstriE have c'_valid:"valid_lpo_memory pr c'" by simp
        show ?thesis proof (cases ssNstri)
          case True
          from None Pair True have lpoE_val: "fst (lpoE c pr (Fun f ss) (Fun g ts)) = (True,True)" by simp
          from this None Pair True  have lpoE_memory: "snd (lpoE c pr (Fun f ss) (Fun g ts)) = store c' (Fun f ss, Fun g ts) (True,True)" by simp
          from Pair anyNstriE True have lpo_val:"lpo_unbounded pr (Fun f ss) (Fun g ts) = (True,True)" by simp
          from lpoE_val lpo_val lpoE_memory c'_valid show ?thesis unfolding store_def by simp
        next
          case False note ssNstri_False = this
          have "\<And>t c. t \<in> set ts \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr (Fun f ss) t) = lpo_unbounded pr (Fun f ss) t 
            \<and> valid_lpo_memory pr (snd (lpoE c pr (Fun f ss) t))" proof 
            fix t c
            from Fun(1)[of t c] Fun(2) show "t \<in> set ts \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr (Fun f ss) t) = lpo_unbounded pr (Fun f ss) t" 
              by (simp only: assms, blast)
            from Fun(1)[of t c] Fun(2) show "t \<in> set ts \<and> valid_lpo_memory pr c \<Longrightarrow> valid_lpo_memory pr (snd (lpoE c pr (Fun f ss) t))" 
              by (simp only: assms, blast)
          qed
          from forAllStriE_sound[of ts "lpo_unbounded pr" "(\<lambda>lc. lpoE lc pr (Fun f ss))" "(Fun f ss)" c'] this c'_valid
          have forAllStriE:"valid_lpo_memory pr (snd (forAllStriE (\<lambda>lc. lpoE lc pr (Fun f ss)) c' ts)) \<and>
            fst (forAllStriE (\<lambda>lc. lpoE lc pr (Fun f ss)) c' ts) = list_all (\<lambda>t. fst (lpo_unbounded pr (Fun f ss) t)) ts" 
            by (simp only: valid_lpo_memory_def, blast)
          show ?thesis proof (cases "forAllStriE (\<lambda>lc. lpoE lc pr (Fun f ss)) c' ts")
            case (Pair sAllStri c'')
            from this forAllStriE have c''_valid: "valid_lpo_memory pr c''" by simp
            show ?thesis proof (cases "pr f (length ss) \<ge> pr g (length ts) \<and> sAllStri")
              case True note geqTrue=this
              show ?thesis proof (cases "pr f (length ss) > pr g (length ts)")
                case True note geTrue=this
                from None Pair anyNstriE_Pair geqTrue geTrue ssNstri_False have lpoE_val: "fst (lpoE c pr (Fun f ss) (Fun g ts)) = (True,True)" by simp
                from this None Pair anyNstriE_Pair geqTrue geTrue ssNstri_False 
                have lpoE_memory: "snd (lpoE c pr (Fun f ss) (Fun g ts)) = store c'' (Fun f ss, Fun g ts) (True,True)" by simp 
                from Pair forAllStriE geTrue geqTrue have lpo_val:"lpo_unbounded pr (Fun f ss) (Fun g ts) = (True,True)" by simp
                from lpoE_val lpo_val lpoE_memory c''_valid show ?thesis unfolding store_def by simp
              next
                case False note geFalse=this
                have "\<And>s c t. s \<in> set ss \<and> t \<in> set ts \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr s t) = lpo_unbounded pr s t 
                  \<and> valid_lpo_memory pr (snd (lpoE c pr s t))" proof 
                  fix s c t
                  from oFun(1)[of s c t] show "s \<in> set ss \<and> t \<in> set ts \<and> valid_lpo_memory pr c \<Longrightarrow> fst (lpoE c pr s t) = lpo_unbounded pr s t"
                    and "s \<in> set ss \<and> t \<in> set ts \<and> valid_lpo_memory pr c \<Longrightarrow> valid_lpo_memory pr (snd (lpoE c pr s t))" by auto
                qed
                from lex_extE_sound[of ss ts "(lpo_unbounded pr)" "(\<lambda>lc. lpoE lc pr)" c''] this c''_valid
                have lex_extE:"valid_lpo_memory pr (snd (lex_extE (\<lambda>lc. lpoE lc pr) c'' ss ts)) \<and>
                  fst (lex_extE (\<lambda>lc. lpoE lc pr) c'' ss ts) = lex_ext_unbounded (lpo_unbounded pr) ss ts" by (simp only: valid_lpo_memory_def, blast)
                from None anyNstriE_Pair ssNstri_False Pair geqTrue geFalse have 
                  lpoE_val:"fst (lpoE c pr (Fun f ss) (Fun g ts)) = fst (lex_extE (\<lambda>lc. lpoE lc pr) c'' ss ts)" 
                  by (simp only: lpoE.simps[of c "pr" "Fun f ss" "Fun g ts"], simp del: lpoE.simps, simp add: split_def)
                from anyNstriE ssNstri_False anyNstriE_Pair have anyFalse:"list_ex (\<lambda>s. snd (lpo_unbounded pr s (Fun g ts))) ss = False" by simp
                from forAllStriE Pair geqTrue have forallTrue:
                  "pr g (length ts) \<le> pr f (length ss) \<and> list_all (\<lambda>t. fst (lpo_unbounded pr (Fun f ss) t)) ts = True" by simp
                have lpo_val: "lpo_unbounded pr (Fun f ss) (Fun g ts) = fst (lex_extE (\<lambda>lc. lpoE lc pr) c'' ss ts)" 
                  by (simp only: "lpo_unbounded.simps", simp only: anyFalse Let_def forallTrue, simp add: geFalse lex_extE)
                from lpoE_val lpo_val have correct_value: "fst (lpoE c pr (Fun f ss) (Fun g ts)) = lpo_unbounded pr (Fun f ss) (Fun g ts)" by simp
                have lpoE_memory:
                  "snd (lpoE c pr (Fun f ss) (Fun g ts)) = 
                  store (snd (lex_extE (\<lambda>lc. lpoE lc pr) c'' ss ts)) (Fun f ss, Fun g ts) (lpo_unbounded pr (Fun f ss) (Fun g ts))" 
                  by (simp only: lpoE.simps[of c "pr" "(Fun f ss)" "(Fun g ts)"], 
                    simp del: lpo_unbounded.simps lpoE.simps add: None ssNstri_False Pair anyNstriE_Pair geqTrue geFalse, 
                    simp del: lpoE.simps lpo_unbounded.simps add:split_def lpo_val)
                from this lex_extE have "valid_lpo_memory pr (snd (lpoE c pr (Fun f ss) (Fun g ts)))" unfolding store_def by simp 
                from this correct_value show ?thesis by simp
              qed
            next
              case False note geqFalse = this
              from None Pair anyNstriE_Pair geqFalse ssNstri_False have lpoE_val: "fst (lpoE c pr (Fun f ss) (Fun g ts)) = (False,False)" by auto 
              from this None Pair anyNstriE_Pair geqFalse ssNstri_False 
              have lpoE_memory: "snd (lpoE c pr (Fun f ss) (Fun g ts)) = store c'' (Fun f ss, Fun g ts) (False,False)" by auto
              from Pair anyNstriE_Pair forAllStriE geqFalse anyNstriE ssNstri_False 
              have lpo_val:"lpo_unbounded pr (Fun f ss) (Fun g ts) = (False,False)" by (simp only: lpo_unbounded.simps, simp only: Let_def, auto)
              from lpoE_val lpo_val lpoE_memory c''_valid show ?thesis unfolding store_def by simp
            qed
          qed
        qed
      qed
    qed
  qed
qed

theorem lpo_E_to_lpo_unbounded : "lpo_E pr s t \<equiv> lpo_unbounded pr s t"
using lpoE_sound[of "pr" "[]" s t] empty_memory_is_valid_lpo_memory[of "pr"] unfolding lpo_E_def by simp

declare lpo_unbounded.simps[code del]
lemma [code]: "lpo_unbounded = lpo_E"
  by (intro ext, simp add: lpo_E_to_lpo_unbounded)


end