(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory ProofChecker
imports SubtermCriterionImpl DepGraph RedPairImpl SCTProcessors DpTransImpl Nontermination
  RootLabeling StringReversal
begin

types
  ('f, 'l, 'v) trsLL = "(('f, 'l) lab, 'v) rules"
  ('f, 'l, 'v) rseqL = "((('f, 'l) lab, 'v) rule \<times> (('f, 'l) lab, 'v) rseq) list"

datatype ('f, 'l, 'v) "DPProof" =
  PisEmpty |
  RedPairProc "('f,'l)lab redpair" "('f,'l,'v)trsLL"  "('f,'l,'v)DPProof" |
  RedPairUrProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v) DPProof" |
  DepGraphProc "(('f,'l,'v)DPProof option \<times> ('f,'l,'v)trsLL) list" |
  MonoRedPairProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  MonoRedPairUrProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  UnlabProc "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  SemlabProc "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  RootlabProc "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  SubtermCriterionProc "('f, 'l) lab projL" "('f, 'l, 'v) rseqL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  SizeChangeSubTerm "((('f,'l)lab,'v)rule \<times> ((nat\<times>nat)list \<times> (nat\<times>nat)list))list" |
  SizeChangeRedPair "('f,'l)lab redpair" "('f,'l,'v)trsLL" "((('f,'l)lab,'v)rule \<times> ((nat\<times>nat)list \<times> (nat\<times>nat)list))list" |
  FccProc "('f, 'l)lab" "(('f,'l)lab,'v)ctxt list" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof"

definition
  indent :: "shows \<Rightarrow> shows"
where
  "indent p s \<equiv> flat_map (\<lambda>c. (if c = newline then newline # '' '' else [c])) (p s)"

types
  ('f, 'l, 'v) ruleMap = "(('f,'l)lab,'v)rule_map"


function
  checkDPProof ::
    "shows \<Rightarrow> ('f, 'l, string) trsLL \<Rightarrow> ('f, 'l, string) trsLL \<Rightarrow> nat
     \<Rightarrow> ('f, 'l, string) ruleMap
     \<Rightarrow> ('f, 'l, string) ruleMap \<Rightarrow> ('f::show, 'l::show, string) DPProof
     \<Rightarrow> shows check" 
where 
  "checkDPProof i P R m rm rrm PisEmpty = debug (i []) ''P is empty''
     (case P of
       Nil \<Rightarrow> return ()
     | Cons _ _ \<Rightarrow> 
       error (i +@+ '': P is not empty in the following DP-problem '' +#+
         shows_nl +@+ shows_dpp P R))" |
  "checkDPProof i P R m rm rrm (RedPairProc redp rP prf) = debug (i []) ''RedPairProc'' (do {
         check_redpair_proc (get_redpair redp) P rP R 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor to switch from''
             +#+ shows_nl +@+ shows_dpp P R +@+ shows_nl +@+
             ''to the DP problem'' +#+ shows_nl +@+ shows_dpp rP R +@+ shows_nl +@+ s);
         checkDPProof (i +@+ shows_string ''.1'') rP R m rm rrm prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" |
  "checkDPProof i P R m rm rrm (RedPairUrProc redp rP ur prf)  =
     debug (i []) ''RedPairProc (usable rules + af)'' (do {
       check_redpair_ur_af_proc (get_redpair redp) P rP R rm ur
         <+? (\<lambda>s. i +@+ '': error when applying the reduction pair '' +#+
           ''processor (with usable rules and af) to switch from '' +#+ shows_nl +@+
           shows_dpp P R +@+ shows_nl +@+ ''to the DP problem '' +#+ shows_nl +@+
           shows_dpp rP R +@+ shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP R m rm rrm prf
         <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+
           shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (DepGraphProc edpts) =
     debug (i []) ''DepGraphProc'' (do {
       check_dep_graph_proc P (map (\<lambda>x.
         ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts) rm rrm
         <+? (\<lambda>s. i +@+ '': error while trying to perform SCC-decomposition  on '' +#+
           shows_nl +@+ shows_dpp P R +@+ shows_nl +@+ s);
        check_allm_gen_index
          (\<lambda>rp j. (case fst rp of Some prf \<Rightarrow> Suc j | None \<Rightarrow> j))
          (\<lambda>rp j.
            (case fst rp of
              Some prf \<Rightarrow> checkDPProof (i +@+ ''.'' +#+ shows j) (snd rp) R m rm rrm prf
            | None \<Rightarrow> OK))
          1 edpts
          <+? (\<lambda>s. i +@+ '': error below the dependency graph processor'' +#+ shows_nl
            +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (MonoRedPairProc redp rP rR prf)  =
     debug (i []) ''MonoRedPairProc'' (do {
       check_mono_redpair_proc (get_redpair redp) P rP R rR
         <+? (\<lambda>s. i +@+ '': error when applying the monotonic reduction pair processor ''
           +#+ ''to switch from '' +#+ shows_nl +@+ shows_dpp P R +@+ shows_nl +@+
           ''to the DP problem '' +#+ shows_nl +@+ shows_dpp rP rR +@+ shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP rR m
         (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
         <+? (\<lambda>s. i +@+ '': error below the monotonic reduction pair processor'' +#+
           shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (MonoRedPairUrProc redp rP rR ur prf) =
     debug (i []) ''MonoRedPairProc (usable rules)'' (do {
       check_mono_redpair_ur_proc (get_redpair redp) P rP R rm rR ur
         <+? (\<lambda>s. i +@+ '': error when applying the monotonic reduction pair '' +#+
           ''processor (with usable rules) to switch from '' +#+ shows_nl +@+ shows_dpp P R
           +@+ shows_nl +@+ ''to the DP problem'' +#+ shows_nl +@+ shows_dpp rP rR +@+
           shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP rR m
         (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
         <+? (\<lambda>s. i +@+ '': error below the monotonic reduction pair '' +#+
           ''(with usable rules) processor'' +#+ shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R (Suc m) rm rrm (UnlabProc ulP ulR prf) =
     debug (i []) ''UnlabProc'' (do {
       check_unlab_proc P R ulP ulR;
       checkDPProof (i +@+ shows_string ''.1'') ulP ulR m
         (computeRuleMap ulR) (computeRuleMap (reverse_trs ulR)) prf
         <+? (\<lambda>s. i +@+ '': error below the unlabeling processor'' +#+ shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R 0 rm rrm (UnlabProc ulP ulR prf) =
     error (''unlabeling not allowed due to label-depth of 0'' +#+ shows_nl)" |
  "checkDPProof i P R m rm rrm (SemlabProc sli L lP lR prf) =
     debug (i []) ''SemlabProc'' (do {
       create_sem_lab_proc sli L P R lP lR
         <+? (\<lambda>s. i +@+ '': error when applying the semantic labeling processor'' +#+
           shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') lP lR (Suc m)
         (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf
         <+? (\<lambda>s. i +@+ '': error below the semantic labeling processor'' +#+ shows_nl
           +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (RootlabProc lP lR prf) =
     debug (i []) ''RootlabProc'' (do {
       rl_proc P R lP lR
         <+? (\<lambda>s. i +@+ '': error when applying the root labeling processor'' +#+ shows_nl
           +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') lP lR (Suc m)
         (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf
           <+? (\<lambda>s. i +@+ '': error below the root-labeling processor'' +#+ shows_nl
             +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (SubtermCriterionProc p rseq rP prf) =
     debug (i []) ''SubtermProc'' (do {
       create_subterm_criterion_proc p rseq P rP R rm
         <+? (\<lambda>s. i +@+ '': error when applying the subterm criterion processor''
           +#+ shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP R m rm rrm prf
         <+? (\<lambda>s. i +@+ '': error below the subterm criterion processor'' +#+ shows_nl
           +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (SizeChangeSubTerm graphs) =
     debug (i []) ''SizeChangeSubTerm'' (check_sct_subterm_proc P graphs rm rrm)" |
  "checkDPProof i P R m rm rrm (SizeChangeRedPair redp ur graphs) =
     debug (i []) ''SizeChangeRedPair'' (check_sct_ur_af_proc (get_redpair redp) P  graphs rm rrm ur)" |
  "checkDPProof i P R m rm rrm (FccProc f fcs rP rR prf) =
     debug (i []) ''FccProc'' (do {
       fcc_proc P R f fcs rP rR
         <+? (\<lambda>s. i +@+ '': error when applying the closure under flat contexts'' +#+
           shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP rR m
         (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
         <+? (\<lambda>s. i +@+ '': error below flat context closure processor'' +#+ shows_nl
           +@+ indent s)
     })"
by (pat_completeness) auto

termination "checkDPProof"
proof 
  let ?R = "measures [(\<lambda> (i,P,R,m,rm,rrm,prf). size prf)]"
  show "wf ?R" by auto
next
  fix i P R m rm rrm redp rP "prf" 
  show "((i +@+ shows_string ''.1'', rP, R, m, rm, rrm, prf), i, P, R, m, rm, rrm, RedPairProc redp rP prf) \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm redp rP ur "prf"
  show "((i +@+ shows_string ''.1'', rP, R, m, rm, rrm, prf), i, P, R, m, rm, rrm, RedPairUrProc redp rP ur prf) \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i and P R::"('f::show,'l::show,string)trsLL" and m rm rrm and  
    edpts :: "(('f,'l,string)DPProof option \<times> ('f,'l,string)trsLL) list" and n x and "prf"::"('f,'l,string)DPProof"
  assume elem: "x \<in> set edpts" and "fst x = Some prf"
  hence "(option_size size) (fst x) = Suc (size prf)" by auto
  hence "prod_size (option_size size) (list_size
                (prod_size (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0)) 
                (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0)))) x > size prf" 
    (is "?m x > size prf")
    by (cases x, auto)   
  with elem have "size prf < ((list_size ?m) edpts)" 
  proof (induct edpts, simp)
    case (Cons y ys)
    thus ?case by (cases "x = y", auto)
  qed
  with elem show "((i +@+ ''.'' +#+ shows n, snd x, R, m, rm, rrm, prf), i, P, R, m, rm, rrm, DepGraphProc edpts) \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm redp rP rR "prf"
  show "((i +@+ shows_string ''.1'', rP, rR, m, computeRuleMap rR,
         computeRuleMap (reverse_trs rR), prf),
        i, P, R, m, rm, rrm, MonoRedPairProc redp rP rR prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm redp rP rR ur "prf"
  show "((i +@+ shows_string ''.1'', rP, rR, m, computeRuleMap rR,
         computeRuleMap (reverse_trs rR), prf),
        i, P, R, m, rm, rrm, MonoRedPairUrProc redp rP rR ur prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm ulP ulR "prf" 
  show " ((i +@+ shows_string ''.1'', ulP, ulR, m, computeRuleMap ulR,
         computeRuleMap (reverse_trs ulR), prf),
        i, P, R, Suc m, rm, rrm, UnlabProc ulP ulR prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm sli L lP lR "prf"
  show "((i +@+ shows_string ''.1'', lP, lR, Suc m, computeRuleMap lR,
         computeRuleMap (reverse_trs lR), prf),
        i, P, R, m, rm, rrm, SemlabProc sli L lP lR prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm lP lR "prf"
  show "((i +@+ shows_string ''.1'',lP,lR,Suc m,computeRuleMap lR,computeRuleMap(reverse_trs lR),prf),
        i,P,R,m,rm,rrm,RootlabProc lP lR prf)
        \<in> measures [\<lambda>(i,P,R,m,rm,rrm,prf). size prf]" by auto
next
  fix i P R m rm rrm p rseq rP "prf"
  show "((i +@+ shows_string ''.1'',rP,R,m,rm,rrm,prf),i,P,R,m, rm,rrm,
    SubtermCriterionProc p rseq rP prf)
    \<in> measures [\<lambda>(i,P,R,m,rm,rrm,prf). size prf]" by auto
next
  fix i P R m rm rrm f fcs rP rR prof
  show "((i +@+ shows_string ''.1'', rP, rR, m, computeRuleMap rR, computeRuleMap (reverse_trs rR), prof),
    i, P, R, m, rm, rrm, FccProc f fcs rP rR prof)
    \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prof). size prof]" by auto
qed

lemma checkDPProofSoundI:
  fixes P::"('f::show,'l::show,string)trsLL"
  shows "isOK(checkDPProof i P R m rm rrm prf) \<and>
    rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R) \<longrightarrow> finite_lab_dpp (set P, set R) m"
proof (induct rule:checkDPProof.induct[where P = "\<lambda> i P R m rm rrm prf. isOK(checkDPProof i P R m rm rrm prf) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R) \<longrightarrow> finite_lab_dpp (set P, set R) m"])
  case (1 i P R m rm rrm)  thus ?case unfolding finite_lab_dpp_def by (cases P, auto)
next
  case (2 i P R m rm rrm redp rP prof) 
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (RedPairProc redp rP prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_only_pairs_lab_proc[where rP = rP], rule generic_redpair_impl.check_redpair_proc_sound[OF get_redpair])
      from rm 2 show "finite_lab_dpp (set rP, set R) m" by force
    next
      from rm 2 show "isOK (check_redpair_proc (get_redpair redp) P rP R)" by auto
    qed
  }
  thus ?case by blast
next
  case (3 i P R m rm rrm redp rP ur prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (RedPairUrProc redp rP ur prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_only_pairs_lab_proc[where rP = rP], rule generic_redpair_impl.check_redpair_ur_af_proc_sound[OF get_redpair])
      from rm 3 show "isOK (check_redpair_ur_af_proc (get_redpair redp) P rP R (computeRuleMap R) ur)" by auto
    next
      from rm 3 show "finite_lab_dpp (set rP, set R) m" by force
    qed
  }
  thus ?case by blast
next
  case (4 i P R m rm rrm edpts) 
  let ?mc = "\<lambda> edpts. (map (\<lambda> x. ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts)"
  let ?edps = "?mc edpts"
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (DepGraphProc edpts)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule dep_graph_lab_proc, auto)
      from 4 rm show "isOK(check_dep_graph_proc P ?edps (computeRuleMap R) (computeRuleMap (reverse_trs R)))" by (auto)
    next
      fix a rp
      assume elem: "(True, rp) \<in> set ?edps"
      from 4 rm have "isOK (check_allm_gen_index
        (\<lambda>rp j. case fst rp of Some prf \<Rightarrow> Suc j | None \<Rightarrow> j)
        (\<lambda>rp j.
          (case fst rp of
            Some prf \<Rightarrow> checkDPProof (i +@+ ''.'' +#+ shows j) (snd rp) R m rm rrm prf
          | None \<Rightarrow> OK))
        1 edpts)" (is "isOK (check_allm_gen_index ?g ?f 1 edpts)") by auto
      hence "\<forall>prfrP \<in> set edpts. \<exists>n. isOK (?f prfrP n)" by (rule isOK_check_allm_gen_index)
      with 4 rm have fin: "\<forall> prf rP. (Some prf, rP) \<in> set edpts \<longrightarrow> finite_lab_dpp (set rP, set R) m" by auto
      from elem obtain sub where "set sub \<subseteq> set edpts \<and> (True, rp) \<in> set (?mc sub)" by blast
      thus "finite_lab_dpp (set rp, set R) m"
      proof (induct sub, simp)
	case (Cons edp subb)
	with fin show ?case 
	  by (cases "edp", cases "fst edp", simp, cases "snd edp = rp", auto)
      qed
    qed
  }
  thus ?case by blast
next
  case (5 i P R m rm rrm redp rP rR prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (MonoRedPairProc redp rP rR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_lab_proc[where rP = rP], rule generic_redpair_impl.check_mono_redpair_proc_sound[OF get_redpair])
      from rm 5 show "isOK (check_mono_redpair_proc (get_redpair redp) P rP R rR)" by auto
    next
      from rm 5 show "finite_lab_dpp (set rP, set rR) m" by force
    qed
  }
  thus ?case by blast
next
  case (6 i P R m rm rrm redp rP rR ur prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (MonoRedPairUrProc redp rP rR ur prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_lab_proc[where rP = rP], rule generic_redpair_impl.check_mono_redpair_ur_proc_sound[OF get_redpair])
      from rm 6 show "isOK (check_mono_redpair_ur_proc (get_redpair redp) P rP R (computeRuleMap R) rR ur)" by auto
    next
      from rm 6 show "finite_lab_dpp (set rP, set rR) m" by force
    qed
  }
  thus ?case by blast
next
  case (7 i P R m rm rrm ulP ulR prof)
  {
    assume rm: "isOK(checkDPProof i P R (Suc m) rm rrm (UnlabProc ulP ulR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) (Suc m)"
    proof (rule apply_sound_lab_proc[where rP = ulP and rR = ulR], rule check_unlab_proc_sound)
      from rm 7 show "finite_lab_dpp (set ulP, set ulR) m" by (force simp: Let_def)
    next
      from rm 7 show "Suc m = Suc m \<and> isOK(check_unlab_proc P R ulP ulR)" by auto
    qed
  }
  thus ?case by blast
next
  case (8 i P R rm rrm ulP ulR prof)
  thus ?case by auto
next
  case (9 i P R m rm rrm sli L lP lR prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (SemlabProc sli L lP lR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m" 
    proof (rule apply_sound_lab_proc[where rP = lP], rule create_sem_lab_proc_sound)
      from rm 9 show "finite_lab_dpp (set lP, set lR) (Suc m)" by force
    next
      from rm 9 show "Suc m = Suc m \<and> isOK (create_sem_lab_proc sli L P R lP lR)" by auto
    qed
  }
  thus ?case by blast
next
  case (10 i P R m rm rrm lP lR prof)
  {
    assume rm: "isOK(checkDPProof i P R m rm rrm (RootlabProc lP lR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap(reverse_trs R)"
    have "finite_lab_dpp(set P,set R) m"
    proof (rule apply_sound_lab_proc[where rP = lP],rule rl_proc_sound)
      from rm 10 show "finite_lab_dpp(set lP,set lR) (Suc m)" by force
    next
      from rm 10 show "Suc m = Suc m \<and> isOK(rl_proc P R lP lR)" by auto
    qed
  }
  thus ?case by blast
next
  case (11 i P R m rm rrm p rseq rP prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (SubtermCriterionProc p rseq rP prof))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P,set R) m"
    proof (rule apply_ci_only_pairs_lab_proc[where rP = rP], rule subterm_criterion_proc_ci)
      from rm 11 show "isOK(create_subterm_criterion_proc p rseq P rP R (computeRuleMap R))" by auto
    next
      from rm 11 show "finite_lab_dpp(set rP,set R) m" by force
    qed
  }
  thus ?case by blast
next 
  case (12 i P R m rm rrm graphs)
  {
    assume rm: "isOK(checkDPProof i P R m rm rrm (SizeChangeSubTerm graphs))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P, set R) m"
    proof (rule apply_ci_only_pairs_lab_proc[OF check_sct_subterm_proc_sound, of "[]"])
      show "finite_lab_dpp (set [], set R) m" unfolding finite_lab_dpp_def by auto
    next
      from rm 12 show "isOK(check_sct_subterm_proc P graphs (computeRuleMap R) (computeRuleMap (reverse_trs R)))" by force
    qed
  }
  thus ?case by blast
next
  case (13 i P R m rm rrm redp ur graphs)
  {
    assume rm: "isOK(checkDPProof i P R m rm rrm (SizeChangeRedPair redp ur graphs))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P, set R) m"
    proof (rule apply_ci_only_pairs_lab_proc[OF generic_redpair_impl.check_sct_ur_af_proc_sound[OF get_redpair], of "[]"])
      show "finite_lab_dpp (set [], set R) m" unfolding finite_lab_dpp_def by auto
    next
      from rm 13 show "isOK(check_sct_ur_af_proc (get_redpair redp) P graphs (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur )" by force
    qed
  }
  thus ?case by blast
next
  case (14 i P R m rm rrm f fcs rP rR prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (FccProc f fcs rP rR prof))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_sound_lab_proc, rule fcc_proc_sound)
      from 14 rm show "finite_lab_dpp (set rP, set rR) m" by force
    next
      from rm show "isOK (fcc_proc P R f fcs rP rR) \<and> m = m" by auto
    qed
  }
  thus ?case by blast
qed

declare checkDPProof.simps[simp del]

lemma checkDPProofSound: "isOK(checkDPProof i P R m (computeRuleMap R) (computeRuleMap (reverse_trs R)) prf) \<Longrightarrow> finite_lab_dpp (set P, set R) m"
  using checkDPProofSoundI by blast

datatype ('f,'l,'v)"TRSProof" =
  DPTrans "(('f,'l)lab shp,'v) rules"  "(('f,'l)lab shp,'l,'v) DPProof" |
  RuleRemoval "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Unlab "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Semlab "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Rootlab "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Fcc "(('f,'l)lab,'v)ctxt list" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  StringReversal "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  RisEmpty

fun
  checkTRSProof ::
    "shows \<Rightarrow> ('f::show, 'l::show, string) trsLL \<Rightarrow> nat \<Rightarrow> ('f, 'l, string) TRSProof
     \<Rightarrow> shows check"
where
  "checkTRSProof i r m (DPTrans P prf) = debug (i []) ''DP trans'' ((\<lambda>IR. ((\<lambda>ILR. do {
     check_DP r P
       <+? (\<lambda>s. i +@+ '': error when switching from the TRS'' +#+ shows_nl +@+ shows_trs r
         +@+ shows_nl +@+ ''to the initial DP problem'' +#+ shows_nl +@+ shows_dpp P IR +@+ shows_nl +@+ s);
     checkDPProof (i +@+ shows_string ''.1'') (init_lab_rules P) ILR 0
       (computeRuleMap ILR) (computeRuleMap (reverse_trs ILR)) prf
       <+? (\<lambda>s. i +@+ '': error below switch to dependency pairs'' +#+ shows_nl +@+ indent s)
   })
   (init_lab_rules IR)))
   (ID_list r))" |
  "checkTRSProof i r m (RuleRemoval rp rR prf) = debug (i []) ''rule removal'' (do {
     check_rule_removal_proc (get_redpair rp) r rR
       <+? (\<lambda>s. i +@+ '': error in rule removal when switching from the TRS'' +#+ shows_nl
         +@+ shows_trs r +@+ shows_nl +@+ ''to the TRS'' +#+ shows_nl +@+ shows_trs rR
         +@+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') rR m prf
       <+? (\<lambda>s. i +@+ '': error below the rule removal processor'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i r (Suc m) (Unlab uR prf) = debug (i []) ''unlabel'' (do {
     unlab_trs_tt r uR
       <+? (\<lambda>s. i +@+ '': error in unlabeling of TRS when switching from the TRS''
         +#+ shows_nl +@+ shows_trs r +@+ shows_nl +@+ ''to the TRS'' +#+ shows_nl +@+
         shows_trs uR +@+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') uR m prf
       <+? (\<lambda>s. i +@+ '': error below the unlabeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i r 0 (Unlab uR prf) = debug (i []) ''unlabel'' (do {
     check_subseteq (map_trsI unlab r) uR 
       <+? (\<lambda>lr. i +@+ '': error in unlabeling of TRS when switching from the TRS''
         +#+ shows_nl +@+ shows_trs r +@+ shows_nl +@+ ''to the TRS'' +#+ shows_nl +@+
         shows_trs uR +@+ shows_nl +@+ shows_string ''unlabeled rule '' +@+ shows_rule lr +@+ shows_string '' is not present'');
     checkTRSProof (i +@+ shows_string ''.1'') uR 0 prf
       <+? (\<lambda>s. i +@+ '': error below the unlabeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i r m RisEmpty =
     debug (i []) ''R is empty'' (check (r = [])
       (i +@+ '': R is not empty, but it is the TRS'' +#+ shows_trs r +@+ shows_nl))" |
  "checkTRSProof i R m (Semlab sli L lR prf) = debug (i []) ''Semlab'' (do {
     create_sem_lab_tt sli L R lR
       <+? (\<lambda>s. i +@+ '': error when applying semantic labeling'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prf
       <+? (\<lambda>s. i +@+ '': error below semantic labeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i R m (Rootlab lR prf) = debug (i []) ''Rootlab'' (do {
     rl_tt R lR
       <+? (\<lambda>s. i +@+ '': error when applying root-labeling'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prf
       <+? (\<lambda>s. i +@+ '': error below root-labeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i R m (Fcc fcs rR prf) = debug (i []) ''Fcc'' (do {
     fcc_tt R fcs rR
       <+? (\<lambda>s. i +@+ '': error when applying flat context closure'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') rR m prf
       <+? (\<lambda>s. i +@+ '': error below flat context closure'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i R m (StringReversal rR prf) = debug (i []) ''String Reversal'' (do {
     check_string_rev R rR
       <+? (\<lambda>s. i +@+ '': error when applying string reversal'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') rR m prf
       <+? (\<lambda>s. i +@+ '': error below string reversal'' +#+ shows_nl +@+ indent s)
   })"

lemma checkTRSProofSound:
  fixes prof :: "('f::show,'l::show,string)TRSProof"
  assumes check: "isOK(checkTRSProof i r m prof)"
  shows "decrSN (set r) m"
using assms
proof (induct "prof" arbitrary: i r m)
  case (DPTrans P dpprf)
  let ?il = "init_lab_rules :: (('f,'l)lab shp,'v) rules \<Rightarrow> (('f,'l)lab shp,'l,'v)trsLL"
  let ?ilr = "?il (ID_list r)"
  from DPTrans have dpsSound: "isOK(check_DP r P)" and rec: "isOK(checkDPProof (i +@+ shows_string ''.1'') (?il P) ?ilr 0 (computeRuleMap ?ilr) (computeRuleMap (reverse_trs ?ilr)) dpprf)"
    by (auto simp: Let_def)
  from rec have "finite_lab_dpp (set (?il P), set (?il (ID_list r))) 0" 
    by (rule checkDPProofSound) 
  with init_lab_sound have "finite_dpp (set P, set (ID_list r))" by auto
  with dpsSound check_DP have "SN (rstep (set r))" by simp
  thus ?case by (rule SN_to_decr_SN)
next
  case (RuleRemoval rp rR prof')
  hence ok:"isOK (checkTRSProof i r m (RuleRemoval rp rR prof'))" and IH:"\<And>r' i' m'. isOK (checkTRSProof i' r' m' prof') \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have rr_ok:"isOK(check_rule_removal_proc (get_redpair rp) r rR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') rR m prof')" by auto
  from rec_ok IH have SN: "decrSN (set rR) m" by auto
  show "decrSN (set r) m"
    by (rule apply_ci_lab_tt, rule generic_redpair_impl.rule_removal_sound[OF get_redpair], rule SN, rule rr_ok)
next
  case RisEmpty
  hence "SN (rstep (set r))" by auto
  thus ?case by (rule SN_to_decr_SN)
next
  case (Unlab uR prof)
  thus ?case 
  proof (cases m)
    case 0
    with Unlab have ok:"isOK (checkTRSProof i r 0 (Unlab uR prof))" and IH:"\<And>r' i'. isOK (checkTRSProof i' r' 0 prof) \<Longrightarrow> decrSN (set r') 0" by auto
    from ok have ul_ok:"unlab_trs (set r) \<subseteq> set uR" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') uR 0 prof)" by auto
    from rec_ok IH have SN: "SN (rstep (set uR))" unfolding decrSN_def by auto
    from SN_subset[OF SN rstep_mono[OF ul_ok]] have SN: "SN (rstep (unlab_trs (set r)))" .
    from SN_map_imp_SN[of unlab "set r"] have "SN (rstep (set r))" using SN unfolding SN_def by auto
    thus "decrSN (set r) m" unfolding 0 unfolding decrSN_def by simp
  next
    case (Suc m')
    with Unlab have ok:"isOK (checkTRSProof i r (Suc m') (Unlab uR prof))" and IH:"\<And>r' i'. isOK (checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
    from ok have ul_ok:"isOK(unlab_trs_tt r uR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') uR m' prof)" by auto
    from rec_ok IH have SN: "decrSN (set uR) m'" by auto
    show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF unlab_trs_tt_sound], rule SN, auto simp: ul_ok Suc)
  qed
next
  case (Semlab sli L lR prof)
  hence ok:"isOK (checkTRSProof i r m (Semlab sli L lR prof))" and IH:"\<And>r' i' m'. isOK (checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have l_ok:"isOK(create_sem_lab_tt sli L r lR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prof)" by auto
  from rec_ok IH have SN: "decrSN (set lR) (Suc m)" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF create_sem_lab_tt_sound SN], simp, rule l_ok)
next
  case (Rootlab lR prof)
  hence ok:"isOK (checkTRSProof i r m (Rootlab lR prof))" and IH:"\<And>r' i' m'. isOK (checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have l_ok:"isOK(rl_tt r lR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prof)" by auto
  from rec_ok IH have SN: "decrSN (set lR) (Suc m)" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF rl_tt_sound SN], simp, rule l_ok)
next
  case (Fcc fcs rR prof)
  hence ok: "isOK (checkTRSProof i r m (Fcc fcs rR prof))" and IH: "\<And>r' i' m'. isOK(checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have r_ok: "isOK(fcc_tt r fcs rR)" and rec_ok: "isOK(checkTRSProof (i +@+ shows_string ''.1'') rR m prof)" by auto
  from rec_ok IH have SN: "decrSN (set rR) m" by auto
  show "decrSN (set r) m" by (rule apply_weak_ci_lab_tt[OF fcc_weak_ci_tt_impl SN], rule r_ok)
next
  case (StringReversal rR prof)
  hence ok: "isOK (checkTRSProof i r m (StringReversal rR prof))" and IH: "\<And>r' i' m'. isOK(checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have r_ok: "isOK(check_string_rev r rR)" and rec_ok: "isOK(checkTRSProof (i +@+ shows_string ''.1'') rR m prof)" by auto
  from rec_ok IH have SN: "decrSN (set rR) m" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF string_rev_sound_lab_tt_impl SN], simp add: r_ok)
qed

datatype ('f,'l,'v)"RelTRSProof" =
  RelRuleRemoval "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)RelTRSProof" |
  RelSisEmpty  "('f,'l,'v)TRSProof" |
  RelUnlab "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)RelTRSProof" |
  RelSemlab "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)RelTRSProof" |
  RelRootlab "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)RelTRSProof" |
  RelRisEmpty 

fun
  checkRelTRSProof ::
    "shows \<Rightarrow> ('f::show, 'l :: show, string)trsLL \<Rightarrow> ('f,'l,string)trsLL \<Rightarrow> ('f,'l, string) RelTRSProof
     \<Rightarrow> shows check"
where
  "checkRelTRSProof i r s RelRisEmpty =
     debug (i []) ''R is empty'' (check (r = [])
       (i +@+ '': R is not empty, but it is the TRS'' +#+ shows_trs r +@+ shows_nl))" |
  "checkRelTRSProof i r s (RelSisEmpty prf) =
     debug (i []) ''S is empty'' (do { 
       check (s = [])
       (i +@+ '': S is not empty, but it is the TRS'' +#+ shows_trs s +@+ shows_nl);
       checkTRSProof (i +@+ shows_string ''.1'') r 0 prf
       <+? (\<lambda>ss. i +@+ '': error below the rule removal processor'' +#+ shows_nl +@+ indent ss)
 })" | 
  "checkRelTRSProof i r s (RelRuleRemoval rp rR rS prf) = debug (i []) ''rule removal'' (do {
     check_relative_rule_removal (get_redpair rp) (r,s) (rR,rS)
       <+? (\<lambda>ss. i +@+ '': error in rule removal when switching from the TRSs'' +#+ shows_nl
         +@+ shows_trs r +@+ shows_nl +@+ '' and '' +#+ shows_nl +@+ shows_trs s 
         +@+ shows_nl +@+ ''to the TRSs'' +#+ shows_nl +@+ shows_trs rR
         +@+ shows_nl +@+ '' and '' +#+ shows_nl +@+ shows_trs rS +@+ ss);
     checkRelTRSProof (i +@+ shows_string ''.1'') rR rS prf
       <+? (\<lambda>s. i +@+ '': error below the rule removal processor'' +#+ shows_nl +@+ indent s)
   })" |
  "checkRelTRSProof i r s (RelUnlab uR uS prf) = debug (i []) ''unlabel'' (do {
     unlab_reltrs_tt (r,s) (uR,uS)
       <+? (\<lambda>m. i +@+ '': error in unlabeling of TRS when switching from the TRS R''
         +#+ shows_nl +@+ shows_trs r +@+ shows_nl +@+ shows_string ''and S'' +@+ shows_nl +@+ shows_trs s +@+ ''to the TRSs '' +#+ shows_nl +@+
         shows_trs uR +@+ shows_nl +@+ shows_trs uS +@+ m);
     checkRelTRSProof (i +@+ shows_string ''.1'') uR uS prf
       <+? (\<lambda>s. i +@+ '': error below the unlabeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkRelTRSProof i R S (RelSemlab sli L lR lS prf) = debug (i []) ''Semlab'' (do {
     create_sem_lab_rel_tt sli L (R,S) (lR,lS)
       <+? (\<lambda>s. i +@+ '': error when applying semantic labeling'' +#+ shows_nl +@+ s);
     checkRelTRSProof (i +@+ shows_string ''.1'') lR lS prf
       <+? (\<lambda>s. i +@+ '': error below semantic labeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkRelTRSProof i R S (RelRootlab lR lS prf) = debug (i []) ''Rootlab'' (do {
     rl_rel_tt (R,S) (lR,lS)
       <+? (\<lambda>s. i +@+ '': error when applying root-labeling'' +#+ shows_nl +@+ s);
     checkRelTRSProof (i +@+ shows_string ''.1'') lR lS prf
       <+? (\<lambda>s. i +@+ '': error below semantic labeling'' +#+ shows_nl +@+ indent s)
   })" 



lemma checkRelTRSProofSound:
  fixes prof :: "('f::show,'l::show,string)RelTRSProof"
  assumes check: "isOK(checkRelTRSProof i r s prof)"
  shows "rel_SN_alt (set r, set s)"
using assms
proof (induct "prof" arbitrary: i r s)
  case RelRisEmpty
  thus "rel_SN_alt (set r, set s)" by (simp add: rel_SN_alt.simps)
next
  case (RelRuleRemoval rp rr rs prof')
  hence ok:"isOK (checkRelTRSProof i r s (RelRuleRemoval rp rr rs prof'))" and IH:"\<And>r' s' i'. isOK (checkRelTRSProof i' r' s' prof') \<Longrightarrow> rel_SN_alt (set r', set s')" by auto
  from ok have rr_ok:"isOK(check_relative_rule_removal (get_redpair rp) (r,s) (rr,rs))" and rec_ok:"isOK(checkRelTRSProof (i +@+ shows_string ''.1'') rr rs prof')" 
    by auto
  from rec_ok IH have relSN: "rel_SN_alt (set rr, set rs)" by auto
  show "rel_SN_alt (set r, set s)"
    using apply_sound_rel_tt_impl[OF generic_redpair_impl.relative_rule_removal_sound[OF get_redpair] _ relSN, OF rr_ok] by simp
next
  case (RelSisEmpty prof')
  hence s: "set s = {}" and ok: "isOK(checkTRSProof (i +@+ shows_string ''.1'') r 0 prof')" by auto
  from checkTRSProofSound[OF ok, simplified] decr_SN_to_SN_II have "SN (rstep (set r))" by auto
  thus "rel_SN_alt (set r, set s)" 
    unfolding s rel_SN_s_empty .
next
  case (RelUnlab uR uS prof')
  hence ok:"isOK (checkRelTRSProof i r s (RelUnlab uR uS prof'))" and IH:"\<And>r' s' i'. isOK (checkRelTRSProof i' r' s' prof') \<Longrightarrow> rel_SN_alt (set r', set s')" by auto
  from ok have ul_ok:"isOK(unlab_reltrs_tt (r,s) (uR,uS))" and rec_ok:"isOK(checkRelTRSProof (i +@+ shows_string ''.1'') uR uS prof')" by auto
  from rec_ok IH have SN: "rel_SN_alt (set uR, set uS)" by auto
  show "rel_SN_alt (set r, set s)" by (rule apply_sound_rel_tt_impl[OF unlab_reltrs_tt _ SN], rule ul_ok)
next
  case (RelSemlab sli L lR lS prof')
  hence ok:"isOK (checkRelTRSProof i r s (RelSemlab sli L lR lS prof'))" and IH:"\<And>r' s' i'. isOK (checkRelTRSProof i' r' s' prof') \<Longrightarrow> rel_SN_alt (set r', set s')" by auto
  from ok have l_ok:"isOK(create_sem_lab_rel_tt sli L (r,s) (lR,lS))" and rec_ok:"isOK(checkRelTRSProof (i +@+ shows_string ''.1'') lR lS prof')" by auto
  from rec_ok IH have SN: "rel_SN_alt (set lR, set lS)" by auto
  show "rel_SN_alt (set r, set s)" by (rule apply_sound_rel_tt_impl[OF create_sem_lab_rel_tt_sound _ SN], rule l_ok)
next
  case (RelRootlab lR lS prof')
  hence ok:"isOK (checkRelTRSProof i r s (RelRootlab lR lS prof'))" and IH:"\<And>r' s' i'. isOK (checkRelTRSProof i' r' s' prof') \<Longrightarrow> rel_SN_alt (set r', set s')" by auto
  from ok have l_ok:"isOK(rl_rel_tt (r,s) (lR,lS))" and rec_ok:"isOK(checkRelTRSProof (i +@+ shows_string ''.1'') lR lS prof')" by auto
  from rec_ok IH have SN: "rel_SN_alt (set lR, set lS)" by auto
  show "rel_SN_alt (set r, set s)" by (rule apply_sound_rel_tt_impl[OF rl_rel_tt_sound _ SN], rule l_ok)
qed



datatype ('f,'l,'v)certificationProblem = 
  TRSProof "('f,'v)rules" "('f,'l,'v)TRSProof" | 
  TRSDisproof "('f,'v)rules" "('f,'v)TRSDisproof" | 
  RelTRSProof "('f,'v)reltrsL" "('f,'l,'v)RelTRSProof" |
  RelTRSDisproof "('f,'v)reltrsL" "('f,'v)RelTRSDisproof"

datatype ('f,'v)inputProblem = TRS "('f,'v)rules" | RelTRS "('f,'v)reltrsL"

fun input :: "('f,'l,'v)certificationProblem \<Rightarrow> ('f,'v)inputProblem"
where "input (TRSProof trs _) = TRS trs"
  |   "input (TRSDisproof trs _) = TRS trs"
  |   "input (RelTRSProof trs _) = RelTRS trs"
  |   "input (RelTRSDisproof trs _) = RelTRS trs"

fun desired_property :: "('f,'l,'v)certificationProblem \<Rightarrow> bool"
where "desired_property (TRSProof trs _) = SN (rstep (set trs))"
   |  "desired_property (TRSDisproof trs _) = (\<not> (SN (rstep (set trs))))"
   |  "desired_property (RelTRSProof (r,s) _) = rel_SN (set r, set s)"
   |  "desired_property (RelTRSDisproof (r,s) _) = (\<not> rel_SN (set r, set s))"

fun xml_tag :: "('f,'l,'v)certificationProblem \<Rightarrow> string"
where "xml_tag (TRSProof _ _) = ''<trsTerminationProof>''"
  |   "xml_tag (TRSDisproof _ _) = ''<trsNonterminationProof>''"
  |   "xml_tag (RelTRSProof _ _) = ''<relativeTerminationProof>''"
  |   "xml_tag (RelTRSDisproof _ _) = ''<relativeNonterminationProof>''"

datatype certificationResult = Certified string | Error string

fun
  certifyCertProblem ::
    "('f::show,'l::show,string)certificationProblem \<Rightarrow> certificationResult"
where
  "certifyCertProblem (TRSProof trs prf) =
     (case checkTRSProof (shows_string ''1'') (init_lab_rules trs) 0 prf of
       Inl err \<Rightarrow> Error((''termination proof not accepted'' +#+ shows_nl +@+ err) [])
     | Inr _ \<Rightarrow> Certified ''<trsTerminationProof>'')" |
  "certifyCertProblem (TRSDisproof trs prf) =
     (case checkTRSDisproof (shows_string ''1'') trs prf of
       Inl err \<Rightarrow> Error((''non-termination proof not accepted'' +#+ shows_nl +@+ err) [])
     | Inr _ \<Rightarrow> Certified ''<trsNonterminationProof>'')" |
  "certifyCertProblem (RelTRSProof (r,s) prf) =
     (case checkRelTRSProof (shows_string ''1'') (init_lab_rules r) (init_lab_rules s) prf of
       Inl err \<Rightarrow> Error((''relative termination proof not accepted'' +#+ shows_nl +@+ err) [])
     | Inr _ \<Rightarrow> Certified ''<relativeTerminationProof>'')"  |
  "certifyCertProblem (RelTRSDisproof (r,s) prf) =
     (case checkRelTRSDisproof (shows_string ''1'') (r,s) prf of
       Inl err \<Rightarrow> Error((''relative non-termination proof not accepted'' +#+ shows_nl +@+ err) [])
     | Inr _ \<Rightarrow> Certified ''<relativeNonterminationProof>'')" 

lemma certifyCertProblemSound:
  assumes answer: "certifyCertProblem (cp :: ('f :: show,'l :: show,string)certificationProblem) = Certified info"
  shows "desired_property cp \<and> xml_tag cp = info"
proof (cases cp)
  case (TRSProof trs prof)
  let ?IL = "init_lab_rules trs :: (('f,'l)lab,string) rules"
  from TRSProof answer have ok: "isOK(checkTRSProof (shows_string ''1'') ?IL 0 prof) \<and> xml_tag cp = info" 
    by (cases "checkTRSProof (shows_string ''1'') ?IL 0 prof") auto
  have "decrSN (init_lab_trs (set trs) :: (('f,'l)lab,string)trs) 0" using checkTRSProofSound ok by force
  with TRSProof ok decr_SN_to_SN[of "set trs"] answer show ?thesis by auto
next
  case (TRSDisproof trs prof)
  with answer have "isOK(checkTRSDisproof (shows_string ''1'') trs prof) \<and> xml_tag cp = info"
    by (cases "checkTRSDisproof (shows_string ''1'') trs prof") auto
  with checkTRSDisproofSound TRSDisproof show ?thesis by auto
next
  case (RelTRSProof rs prof)
  show ?thesis
  proof (cases rs)
    case (Pair r s)
    let ?ILr = "init_lab_rules r :: (('f,'l)lab,string) rules"
    let ?ILs = "init_lab_rules s :: (('f,'l)lab,string) rules"
    from Pair RelTRSProof answer have "isOK(checkRelTRSProof (shows_string ''1'') ?ILr ?ILs prof) \<and> xml_tag cp = info" (is "?check \<and> ?tag")
      by (cases "checkRelTRSProof (shows_string ''1'') ?ILr ?ILs prof", auto)
    hence ?check and ?tag by blast+
    from checkRelTRSProofSound[OF `?check`] have "rel_SN_alt (set ?ILr, set ?ILs)" by auto
    with rel_SN_init_lab have "rel_SN_alt (set r, set s)" by auto 
    with RelTRSProof Pair `?tag` 
    show ?thesis by (simp, simp add: rel_SN_conv)
  qed
next
  case (RelTRSDisproof rs prof)
  show ?thesis
  proof (cases rs)
    case (Pair r s)
    with RelTRSDisproof answer have "isOK(checkRelTRSDisproof (shows_string ''1'') (r,s) prof) \<and> xml_tag cp = info"
      by (cases "checkRelTRSDisproof (shows_string ''1'') (r,s) prof") auto
    with checkRelTRSDisproofSound RelTRSDisproof Pair show ?thesis 
      by (auto)
  qed
qed

end
