(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SemanticLabeling
imports DpFramework RelativeRewriting
begin


lemma ichain_imp_map_ichain:
  assumes chain: "ichain (P,R) s t \<sigma>"
  shows "ichain (map_trs fg P, map_trs fg R) (\<lambda> i. map_term fg (s i)) (\<lambda> i. map_term fg (t i)) (\<lambda> i. map_subst fg (\<sigma> i))"
  (is "ichain (?lP, ?lR) ?ls ?lt ?lsig")
proof -
  have mem: "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
  proof
    fix i
    from chain have "(s i, t i) \<in> P" by auto
    thus "(?ls i, ?lt i) \<in> ?lP" by force
  qed
  have steps: "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*"
  proof
    fix i
    from chain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    hence "(map_term fg (t i \<cdot> \<sigma> i), map_term fg (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (rstep (map_trs fg R))^*" by (rule rewrites_imp_map_rewrites)
    thus "(?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*" by auto
  qed
  from mem steps show ?thesis by auto
qed
  

text {* now we head for labelings *}

datatype ('f,'l)lab = Lab "('f,'l)lab" 'l | FunLab "('f,'l)lab" "('f,'l)lab list" | UnLab 'f

text {* An alternative induction scheme for labs. *}
lemma
  fixes P :: "('f,'l)lab \<Rightarrow> bool"
  assumes "\<And>f. P(UnLab f)"
    and "\<And>f l. P f \<Longrightarrow> P(Lab f l)"
    and "\<And>f ls. P f \<Longrightarrow> (\<And>l. l \<in> set ls \<Longrightarrow> P l) \<Longrightarrow> P(FunLab f ls)"
  shows lab_induct[case_names UnLab Lab FunLab,induct type: lab]: "P f"
    and "\<And>f. f \<in> set fs \<Longrightarrow> P f"
using assms by (induct f and fs) auto

instantiation lab :: ("show", "show") "show" begin
  fun shows_lab :: "('f::show,'l::show)lab \<Rightarrow> shows"
  where "shows_lab(UnLab f) = shows f"
      | "shows_lab(Lab f l) = (shows_lab f +@+ ''['' +#+ shows l +@+ shows_string '']'')"
      | "shows_lab(FunLab f l) = (shows_lab f +@+ ''['' +#+ shows_map shows_lab l +@+ shows_string '']'')"
  definition "shows_prec (d::nat) l = shows_lab l"
  definition "shows_list_lab (fs::('f::show,'l::show)lab list) \<equiv> shows_list_aux shows fs"

  lemma assoc_lab: "shows_prec d (f::('f::show,'l::show)lab) r @ s = shows_prec d f (r @ s)" by (induct f arbitrary: r s) (auto simp: shows_prec_lab_def)
  lemma assoc_lab_elt: "\<forall>(f::('f::show,'l::show)lab)\<in>set fs. shows_prec d f r @ s = shows_prec d f (r @ s)"
    using assoc_lab by auto
  instance by
    (intro_classes,rule assoc_lab,unfold shows_list_lab_def)
    (rule shows_list_aux_assoc[OF assoc_lab_elt])
end

text {* Resolve problems with recurisve dependencies in generated code. *}
lemma [code]:
  "HOL.eq (FunLab f ls) (FunLab g ks) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ls ks"
  by (simp_all add: eq list_all2_eq [symmetric]) 

(* unlab removes one layer of labelings *)
fun unlab :: "('f,'l)lab \<Rightarrow> ('f,'l)lab"
where "unlab (Lab f l) = f"
 | "unlab (FunLab f l) = f"
 | "unlab x = x"

abbreviation unlab_term :: "(('f,'l)lab,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
where "unlab_term \<equiv> map_term unlab"

abbreviation unlab_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
where "unlab_trs \<equiv> map_trs unlab"

abbreviation unlab_rules :: "(('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules"
where "unlab_rules \<equiv> map (map_rule unlab)"

abbreviation unlab_subst :: "(('f,'l)lab,'v)subst \<Rightarrow> (('f,'l)lab,'v)subst"
where "unlab_subst \<equiv> map_subst unlab"

(* init_lab adds empty labelings to all function symbols *)
fun init_lab :: "'f \<Rightarrow> ('f,'l)lab"
where "init_lab f = UnLab f"

abbreviation init_lab_term :: "('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
where "init_lab_term \<equiv> map_term init_lab"

abbreviation init_lab_subst :: "('f,'v)subst \<Rightarrow> (('f,'l)lab,'v)subst"
where "init_lab_subst \<equiv> map_subst init_lab"

abbreviation init_lab_rule :: "('f,'v)rule \<Rightarrow> (('f,'l)lab,'v)rule"
where "init_lab_rule \<equiv> map_rule init_lab"

abbreviation init_lab_trs :: "('f,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
where "init_lab_trs \<equiv> map_trs init_lab"

abbreviation init_lab_rules :: "('f,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules"
where "init_lab_rules \<equiv> map init_lab_rule"


(* remove_lab switches back to an unlabeled signature and thereby
   removes all labelings *)
fun remove_lab :: "('f,'l)lab \<Rightarrow> 'f" 
where "remove_lab (Lab f _) = remove_lab f"
 | "remove_lab (UnLab f) = f"
 | "remove_lab (FunLab f _) = remove_lab f"

lemma remove_lab_sound[simp]: "remove_lab (init_lab f) = f" by auto

abbreviation remove_lab_term :: "(('f,'l)lab,'v)term \<Rightarrow> ('f,'v)term"
where "remove_lab_term \<equiv> map_term remove_lab"

abbreviation remove_lab_subst :: "(('f,'l)lab,'v)subst \<Rightarrow> ('f,'v)subst"
where "remove_lab_subst \<equiv> map_subst remove_lab"


lemma remove_lab_term_sound: "remove_lab_term (init_lab_term t) = t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (induct ss, auto)
qed


lemma unlab_init_lab_term_sound: "unlab_term (init_lab_term t) = init_lab_term t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (induct ss, auto)
qed

lemma unlab_init_lab_trs_sound: "unlab_trs (init_lab_trs R) = (init_lab_trs R)"
  by (force simp: unlab_init_lab_term_sound)


abbreviation remove_lab_rule :: "(('f,'l)lab,'v)rule \<Rightarrow> ('f,'v)rule"
where "remove_lab_rule \<equiv> map_rule remove_lab"

lemma remove_lab_rule_sound: "remove_lab_rule (init_lab_rule lr) = lr"
  by (auto simp: remove_lab_term_sound)

abbreviation remove_lab_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> ('f,'v)trs"
where "remove_lab_trs \<equiv> map_trs remove_lab"

lemma remove_lab_trs_sound: fixes R :: "('f,'v)trs"
  shows "remove_lab_trs (init_lab_trs R) = R"
proof 
  show "R \<subseteq> remove_lab_trs (init_lab_trs R :: (('f,'l)lab,'v)trs)" (is "_ \<subseteq> remove_lab_trs ?IR")
  proof 
    fix lr 
    assume "lr \<in> R"
    hence "remove_lab_rule ((init_lab_rule lr) :: (('f,'l)lab,'v)rule) \<in> (remove_lab_trs ?IR)" by force
    thus "lr \<in> remove_lab_trs ?IR" by (simp only: remove_lab_rule_sound)
  qed
next
  show "remove_lab_trs (init_lab_trs R) \<subseteq> R" by (auto simp: remove_lab_term_sound)
qed


definition decr_of_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
  where "decr_of_trs R \<equiv> { (l,r) | l r . (l,r) \<in> R \<and> l \<noteq> r \<and> unlab_term l = unlab_term r}"

definition unlab_of_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs" 
  where "unlab_of_trs R \<equiv> map_trs unlab (R - decr_of_trs R)"

definition decrSN :: "(('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> bool"
  where "decrSN R m \<equiv> (\<forall> n < m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))) \<longrightarrow> SN (rstep R)"

lemma unlab_of_trs_mono: "r \<subseteq> s \<Longrightarrow> unlab_of_trs r \<subseteq> unlab_of_trs s"
  unfolding unlab_of_trs_def decr_of_trs_def by auto

lemma unlab_of_trs_union[simp]: "unlab_of_trs (r \<union> s) = unlab_of_trs r \<union> unlab_of_trs s"
  unfolding unlab_of_trs_def decr_of_trs_def by auto

lemma decr_of_trs_mono: "r \<subseteq> s \<Longrightarrow> decr_of_trs r \<subseteq> decr_of_trs s"
  unfolding decr_of_trs_def by auto

lemma decr_of_trs_union[simp]: "decr_of_trs (r \<union> s) = decr_of_trs r \<union> decr_of_trs s"
  unfolding decr_of_trs_def by auto

lemma decr_of_trs_subset_trs: "decr_of_trs r \<subseteq> r"
  unfolding decr_of_trs_def by auto

lemma unlab_trs_union: "unlab_trs (R \<union> S) = unlab_trs R \<union> unlab_trs S"
  by auto

definition sound_lab_tt :: "((('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> (('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> bool) \<Rightarrow> bool"
  where "sound_lab_tt tt \<equiv> \<forall> R m rR rm. tt R m rR rm \<longrightarrow> (decrSN rR rm \<longrightarrow> decrSN R m)"


definition sound_lab_tt_impl :: "((('f,'l)lab,'v)rules \<Rightarrow> nat \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> nat \<Rightarrow> bool) \<Rightarrow> bool"
  where "sound_lab_tt_impl tt \<equiv> \<forall> R m rR rm. tt R m rR rm \<longrightarrow> (decrSN (set rR) rm \<longrightarrow> decrSN (set R) m)"


lemma SN_unlab:
  fixes R :: "(('f,'l)lab,'v)trs" and t :: "(('f,'l)lab,'v)term"
  assumes SN1: "SN (rstep (decr_of_trs R))"
  and SN2: "SN_elt (rstep (unlab_of_trs R)) (unlab_term t)"
  shows "SN_elt (rstep R) t"
proof (rule ccontr)
  let ?D = "decr_of_trs R"
  let ?ND = "R - ?D"
  have R: "R = ?D \<union> ?ND" using decr_of_trs_subset_trs by auto
  assume "\<not> SN_elt (rstep R) t"
  from this obtain tt where "\<forall> i. (tt i, tt (Suc i)) \<in> rstep R" and t: "tt 0 = t" unfolding SN_elt_def by auto
  hence steps: "\<forall> i. (tt i, tt (Suc i)) \<in> (rstep ?D \<union> rstep ?ND)" by (simp only: rstep_union[symmetric], simp only: R[symmetric])
  show False
  proof (cases "\<exists> i. (\<forall> j. j \<ge> i \<longrightarrow> (tt j, tt (Suc j)) \<in> rstep ?D)")
    case True
    from this obtain i where steps2: "\<forall> j.  j \<ge> i \<longrightarrow> (tt j, tt (Suc j)) \<in> rstep ?D" ..
    have "\<forall> j. ((shift tt i) j, (shift tt i) (Suc j)) \<in> rstep ?D" using steps2 by auto
    with SN1 show False unfolding SN_defs by blast
  next
    case False
    have id: "unlab_trs ?ND = unlab_of_trs R" unfolding unlab_of_trs_def by simp
    have "unlab_trs ?D \<subseteq> Id" unfolding decr_of_trs_def by auto
    with rstep_id rstep_mono have subset: "rstep (unlab_trs ?D) \<subseteq> Id" by blast
    have compat: "Id O rstep (unlab_of_trs R) \<subseteq> rstep (unlab_of_trs R)" by auto
    from steps False have infND: "\<forall> i. (\<exists> j. j \<ge> i \<and> (tt j, tt (Suc j)) \<in> rstep ?ND)" by force
    from steps have "\<forall> i. (unlab_term (tt i), unlab_term (tt (Suc i))) \<in> Id \<union> rstep (unlab_of_trs R)" 
      using rewrite_imp_map_rewrite subset by (simp only: id[symmetric], blast)
    from non_strict_ending[OF this compat] t SN2 have 
      "\<exists> j. \<forall> i \<ge> j. (unlab_term (tt i), unlab_term (tt (Suc i))) \<in> Id - rstep (unlab_of_trs R)" by force
    from this obtain j where nInf: "\<forall> i \<ge> j. (unlab_term (tt i), unlab_term (tt (Suc i))) \<notin> rstep (unlab_of_trs R)" by auto
    from infND have "\<forall> i. (\<exists> j. j \<ge> i \<and> (unlab_term (tt j), unlab_term (tt (Suc j))) \<in> rstep (unlab_of_trs R))"
      using rewrite_imp_map_rewrite[of _ _ ?ND] by (simp only: id[symmetric]) blast+
    with nInf show False by blast
  qed
qed

lemma rel_SN_map_imp_SN:
  assumes SN: "rel_SN_alt (map_trs fg R, map_trs fg S)"
  shows "rel_SN_alt (R,S)"
proof (unfold rel_SN_alt.simps, intro allI impI)
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep (R \<union> S)"
  obtain g where g: "\<And> i. g i = map_term fg (f i)" by auto
  have "\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R \<union> map_trs fg S)"
  proof
    fix i
    from spec[OF steps[simplified rstep_union], of i]
    show "(g i, g (Suc i)) \<in> rstep (map_trs fg R \<union> map_trs fg S)" unfolding g rstep_union
      using rewrite_imp_map_rewrite[where fg = fg] by blast
  qed
  with SN[unfolded rel_SN_alt.simps]
  obtain i where no: "\<forall> j \<ge> i. (g j, g (Suc j)) \<notin> rstep (map_trs fg R)" by blast
  show "\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R"
  proof (rule exI[of _ i], intro allI impI)
    fix j
    assume "i \<le> j"
    with no have "(g j, g (Suc j)) \<notin> rstep (map_trs fg R)" by simp
    with rewrite_imp_map_rewrite[where fg = fg] show "(f j, f (Suc j)) \<notin> rstep R" unfolding g by blast
  qed
qed

lemmas rel_SN_unlab = rel_SN_map_imp_SN[where fg = unlab]
lemmas rel_SN_init_lab = rel_SN_map_imp_SN[where fg = init_lab]


lemma rel_SN_id:
  assumes SN: "rel_SN_alt (R, S')"
  and subset: "\<And> rule. rule \<in> S \<Longrightarrow> rule \<in> S' \<or> fst rule = snd rule"
  shows "rel_SN_alt (R,S)"
oops (* will only be required after decreasing rules can be added, since these are subject to deletion *)
(* proof will be by far easier using rel_SN instead of rel_SN_alt, therefore perhaps first prove rel_SN = rel_SN_alt theorem *)

definition
  unlab_reltrs_tt :: "(('f::show,'l::show)lab,'v::show)reltrsL \<Rightarrow> (('f,'l)lab,'v)reltrsL \<Rightarrow> shows check"
where
  "unlab_reltrs_tt RS rRS \<equiv>  (check_subseteq (map_trsI unlab (fst RS)) (fst rRS) \<guillemotright> check_subseteq (map_trsI unlab (snd RS)) (snd rRS))
    <+? (\<lambda>s. shows_string ''cannot find unlabeled rule '' +@+ shows_nl +@+ shows_rule s)"

lemma unlab_reltrs_tt: "sound_rel_tt_impl (\<lambda> RS rRS. isOK(unlab_reltrs_tt RS rRS))"
unfolding sound_rel_tt_impl_def
proof (intro allI impI)
  fix R S rR rS
  assume check: "isOK (unlab_reltrs_tt (R,S) (rR, rS))" and SN: "rel_SN_alt (set rR, set rS)"
  from check have "unlab_trs (set R) \<subseteq> set rR" and "unlab_trs (set S) \<subseteq> set rS" unfolding unlab_reltrs_tt_def by auto
  from rel_SN_alt_mono[OF this SN] show "rel_SN_alt (set R, set S)" by (rule rel_SN_unlab)
qed
   

lemma unlab_tt_sound: "sound_lab_tt (\<lambda> R m rR rm. rR = unlab_of_trs R \<and> Suc rm = m)"
unfolding sound_lab_tt_def decrSN_def
proof (intro allI, intro impI, erule conjE, simp)
  fix R m rR rm
  assume imp: "(\<forall> n < rm. SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) (unlab_of_trs R))))) \<longrightarrow> SN(rstep (unlab_of_trs R))"
    and  SN: "\<forall> n < m. SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))"
    and m: "Suc rm = m"
    and unlab: "rR = unlab_of_trs R"
  have "\<forall> n < rm. SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) (unlab_of_trs R))))"
  proof (rule allI, rule impI)
    fix n
    assume "n < rm"
    with m have le: "Suc n < m" by auto
    have id: "((unlab_of_trs ^^ n) (unlab_of_trs R)) = (unlab_of_trs ^^ Suc n) R" by (simp add: funpow_swap1)
    from SN le show "SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) (unlab_of_trs R))))" by (simp only: id)
  qed
  with imp have SN2: "SN(rstep (unlab_of_trs R))" by auto
  from m have "0 < m" by auto
  with SN have "SN(rstep (decr_of_trs ((unlab_of_trs ^^ 0) R)))" by blast
  hence SN3: "SN(rstep (decr_of_trs R))" by auto
  from SN2 SN_unlab[OF SN3]
  show "SN(rstep R)" unfolding SN_defs by auto
qed


definition lab_minimal_cond :: "(('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)subst) \<Rightarrow> bool"
where "lab_minimal_cond R m s t \<sigma> \<equiv> \<forall> n \<le> m. \<forall> i. (SN_elt(rstep ((unlab_of_trs ^^ n) R)) ((unlab_term ^^ n) (t i \<cdot> \<sigma> i)))"

lemma lab_minimal_cond_imp_minimal_cond[simp]: "lab_minimal_cond R m s t \<sigma> \<Longrightarrow> minimal_cond R s t \<sigma>"
proof -
  assume "lab_minimal_cond R m s t \<sigma>"
  thus ?thesis  unfolding lab_minimal_cond_def minimal_cond_def 
    by (rule allE[where x = 0], simp)
qed

lemma funpow_mono:
  fixes f :: "'a set \<Rightarrow> 'a set"
  assumes mono: "\<And> x y. x \<subseteq> y \<Longrightarrow> f x \<subseteq> f y"
  and subset: "x \<subseteq> y"
  shows "(f ^^ n) x \<subseteq> (f ^^ n) y"
  by (induct n, simp, rule subset, auto simp: mono)

lemma funpow_ci_mono:
  fixes f :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
  assumes mono: "\<And> x y. x \<subseteq>ci y \<Longrightarrow> f x \<subseteq>ci f y"
  and subset: "x \<subseteq>ci y"
  shows "(f ^^ n) x \<subseteq>ci (f ^^ n) y"
  by (induct n, simp, rule subset, auto simp: mono)


lemma rstep_decr_of_trs_mono: assumes "rstep R \<subseteq> rstep S"
  shows "rstep (decr_of_trs R) \<subseteq> rstep (decr_of_trs S)"
(* 
   this lemma does not hold, consider R = f_1(a) \<rightarrow> f_0(a) and S = f_1(x) \<rightarrow> f_0(a) 
   then decr_of_trs R = R, whereas decr_of_trs S = {} 

   this is the reason why ci_subset is not defined as \<rightarrow>R \<subseteq> \<rightarrow>S
*)
oops

lemma decr_of_trs_mono_ci: assumes "R \<subseteq>ci S"
  shows "decr_of_trs R \<subseteq>ci decr_of_trs S"
proof (unfold ci_subset_def, intro allI impI)
  fix l r
  assume "(l,r) \<in> decr_of_trs R"
  hence inR: "(l,r) \<in> R" and noteq: "l \<noteq> r" and unlabeq: "unlab_term l = unlab_term r"
    unfolding decr_of_trs_def by auto
  from assms inR ci_subset_def
  obtain l' r' C where inS: "(l',r') \<in> S" and l: "l = C\<langle>l'\<rangle>" and r: "r = C\<langle>r'\<rangle>" by blast
  with noteq have noteq2: "l' \<noteq> r'" by auto
  from l r unlabeq have unlabeq2: "unlab_term l' = unlab_term r'" by auto
  have "(l',r') \<in> decr_of_trs S" unfolding decr_of_trs_def using inS noteq2 unlabeq2 by auto
  from this l r show "\<exists> l' r' C. (l',r') \<in> decr_of_trs S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>" by blast
qed
 
lemma unlab_of_trs_mono_ci: assumes "R \<subseteq>ci S"
  shows "unlab_of_trs R \<subseteq>ci unlab_of_trs S"
proof (unfold ci_subset_def, intro allI impI)
  fix l r
  assume "(l,r) \<in> unlab_of_trs R"  
  from this obtain ll lr where inR: "(ll,lr) \<in> R" and l: "l = unlab_term ll" and r: "r = unlab_term lr" and cond: "(ll = lr \<or> unlab_term ll \<noteq> unlab_term lr)"
    unfolding unlab_of_trs_def decr_of_trs_def by (force)+
  from assms inR ci_subset_def 
  obtain l' r' C where inS: "(l',r') \<in> S" and ll: "ll = C\<langle>l'\<rangle>" and lr: "lr = C\<langle>r'\<rangle>" by blast
  from cond have "l' = r' \<or> unlab_term l' \<noteq> unlab_term r'" (is "?l \<or> ?r")
  proof
    assume "ll = lr" with l r ll lr show ?thesis by auto
  next
    assume "unlab_term ll \<noteq> unlab_term lr" with l r ll lr show ?thesis by auto
  qed
  with inS have mem: "(unlab_term l', unlab_term r') \<in> unlab_of_trs S" unfolding unlab_of_trs_def decr_of_trs_def by (cases "l' = r'", force+)
  show "\<exists> l' r' C. (l',r') \<in> unlab_of_trs S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    by (rule exI[of _ "unlab_term l'"], 
         rule exI[of _ "unlab_term r'"],
         rule exI[of _ "map_ctxt unlab C"],
      auto simp: l r ll lr mem)
qed

lemma lab_minimal_cond_mono: 
  assumes subset: "R \<subseteq>ci R'" and cond: "lab_minimal_cond R' m s t \<sigma>" 
  shows "lab_minimal_cond R m s t \<sigma>"
unfolding lab_minimal_cond_def
proof (intro allI, intro impI, intro allI)
  fix n i
  assume "n \<le> m"
  with cond have SN: "SN_elt (rstep ((unlab_of_trs ^^ n) R')) ((unlab_term ^^ n) (t i \<cdot> \<sigma> i))"
    unfolding lab_minimal_cond_def by auto
  show "SN_elt (rstep ((unlab_of_trs ^^ n) R)) ((unlab_term ^^ n) (t i \<cdot> \<sigma> i))"
    by (rule SN_elt_subset[OF SN rstep_ci_mono[OF funpow_ci_mono[OF unlab_of_trs_mono_ci subset]]], simp)
qed

fun min_lab_ichain ::
 "(('f,'l)lab,'v)dpp \<Rightarrow> nat \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)subst) \<Rightarrow> bool"
where "min_lab_ichain (P,R) m s t \<sigma> = (ichain (P,R) s t \<sigma> \<and> lab_minimal_cond R m s t \<sigma>)"

lemma min_lab_ichainI[intro]: 
  assumes "R \<subseteq>ci R'" and "lab_minimal_cond R' m s t \<sigma>" and "ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)" 
  shows "min_lab_ichain (P,R) m (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  from assms have "lab_minimal_cond R m s t \<sigma>" using lab_minimal_cond_mono by auto
  hence "lab_minimal_cond R m (shift s i) (shift t i) (shift \<sigma> i)" 
    unfolding lab_minimal_cond_def by auto
  with assms show ?thesis by auto
qed


definition finite_lab_dpp :: "(('f,'l)lab,'v)dpp \<Rightarrow> nat \<Rightarrow> bool" where "finite_lab_dpp DPP m = (\<not>(\<exists>s t \<sigma>. min_lab_ichain DPP m s t \<sigma>))"

lemma decr_of_init_is_empty: 
  "decr_of_trs (init_lab_trs R :: (('f,'l)lab,'v)trs) = {}" (is "decr_of_trs ?IR = {}")
proof -
  {
    fix il ir :: "(('f,'l)lab,'v)term"
    assume nid: "il \<noteq> ir"
    assume id: "unlab_term il = unlab_term ir"
    assume mem: "(il,ir) \<in> ?IR"
    let ?i = "init_lab_term :: ('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
    from mem obtain l r where "(l,r) \<in> R" and il: "il = ?i l" and ir: "ir = ?i r" by force
    from id nid have "?i l \<noteq> ?i r \<and> unlab_term (?i l) = unlab_term (?i r)" by (auto simp: il ir) 
    hence False by (simp only: unlab_init_lab_term_sound, auto)
  }
  thus ?thesis unfolding decr_of_trs_def  by auto
qed

lemma unlab_of_init_is_init: 
  "(unlab_of_trs ^^ n) (init_lab_trs R :: (('f,'l)lab,'v)trs) = (init_lab_trs R :: (('f,'l)lab,'v)trs)" (is "_ = ?IR")
proof -
  {
    fix n
    have "(unlab_of_trs ^^ n) ?IR = ?IR"
    proof (induct n, simp)
      case (Suc n)
      have "(unlab_of_trs ^^ Suc n) ?IR = unlab_of_trs ((unlab_of_trs ^^ n) ?IR)" by simp
      also have "\<dots> = unlab_of_trs ?IR" by (simp only: Suc)
      also have "\<dots> = unlab_trs ?IR" unfolding unlab_of_trs_def using decr_of_init_is_empty by auto
      also have "\<dots> = ?IR" by (rule unlab_init_lab_trs_sound)
      finally show ?case .
    qed
  } 
  thus ?thesis by auto
qed

lemma unlab_term_init_is_init: 
  "(unlab_term ^^ n) (init_lab_term t :: (('f,'l)lab,'v)term) = (init_lab_term t :: (('f,'l)lab,'v)term)" (is "_ = ?it")
proof (induct n, simp)
  case (Suc n)
  with unlab_init_lab_term_sound[of t]
  show ?case by (simp del: init_lab.simps add: funpow_swap1)
qed

lemma init_lab_sound: assumes finite: "finite_lab_dpp (init_lab_trs P :: (('f,'l)lab,'v)trs, init_lab_trs R) m" (is "finite_lab_dpp (?lP, ?lR) _") 
  shows "finite_dpp (P,R)"
proof (rule ccontr)
  assume "\<not> finite_dpp (P,R)"
  from this obtain s t \<sigma> where chain: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?it = "init_lab_term :: ('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
  let ?ir = "init_lab_trs :: ('f,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
  let ?ls = "\<lambda> i. ?it (s i)"
  let ?lt = "\<lambda> i. ?it (t i)"
  let ?lsig = "\<lambda> i. init_lab_subst (\<sigma> i)" 
  let ?lP = "?ir P"
  let ?lR = "?ir R"

  have ichain2: "ichain (?lP,?lR) ?ls ?lt ?lsig" 
    by (rule ichain_imp_map_ichain, rule ichain)
  
  have "\<forall> n \<le> m. \<forall> i. SN_elt (rstep ((unlab_of_trs ^^ n) ?lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))" 
  proof (intro allI, intro impI, intro allI)
    fix n i
    from chain have SN: "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
    have id_term: "(unlab_term ^^ n) (?lt i \<cdot> ?lsig i) = ?lt i \<cdot> ?lsig i" by  (simp only: map_subst_distrib[symmetric], rule unlab_term_init_is_init)
    from SN have foo: "SN_elt (rstep (remove_lab_trs ?lR)) (remove_lab_term (?lt i \<cdot> ?lsig i))" 
      by (simp only: remove_lab_trs_sound remove_lab_term_sound map_subst_distrib[symmetric])
    have "SN_elt (rstep ?lR) (?lt i \<cdot> ?lsig i)" by (rule SN_map_imp_SN[OF foo])
    thus "SN_elt (rstep ((unlab_of_trs ^^ n) ?lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))" by (simp only: id_term unlab_of_init_is_init) 
  qed


  with ichain2 have "min_lab_ichain (?lP,?lR) m ?ls ?lt ?lsig" by (auto simp: lab_minimal_cond_def)

  with finite show False unfolding finite_lab_dpp_def by blast
qed

lemma init_lab_sound_trs:
  assumes SN: "SN (rstep (init_lab_trs R))" (is "SN (rstep ?IR)")
  shows "SN (rstep R)"
  using assms SN_map_imp_SN[of init_lab R] unfolding SN_def by auto


lemma rtrancl_rstep_unlab_of_is_unlab: "(rstep (unlab_of_trs R))^* = (rstep (unlab_trs R))^*" (is "?l = ?r")
proof -
  have "?l = (rstep (unlab_of_trs R) \<union> Id)^*" using rtrancl_reflcl by auto
  also have "\<dots> = (rstep (unlab_of_trs R \<union> Id))^*" by (simp only: rstep_id rstep_union)
  also have "\<dots> = (rstep (unlab_trs R \<union> Id))^*" 
  proof -
    have "unlab_of_trs R \<union> Id = unlab_trs R \<union> Id"
      by (rule set_ext, auto simp: split_paired_all unlab_of_trs_def decr_of_trs_def)
    thus ?thesis by auto
  qed
  also have "\<dots> = (rstep (unlab_trs R) \<union> Id)^*" by (simp only: rstep_id rstep_union)
  also have "\<dots> = (rstep (unlab_trs R))^*" by (rule rtrancl_reflcl)
  finally show ?thesis .
qed

lemma unlab_proc_sound: fixes R :: "(('f,'l)lab,'v)trs"
  assumes finite: "finite_lab_dpp (unlab_trs P, unlab_of_trs R) m"
  shows "finite_lab_dpp (P,R) (Suc m)"
proof (rule ccontr)
  assume "\<not> finite_lab_dpp (P,R) (Suc m)"
  from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) (Suc m) s t \<sigma>" unfolding finite_lab_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?ls = "\<lambda> i. unlab_term (s i)"
  let ?lt = "\<lambda> i. unlab_term (t i)"
  let ?lsig = "\<lambda> i. unlab_subst (\<sigma> i)" 
  let ?lP = "unlab_trs P"
  let ?lR = "unlab_of_trs R"
  from ichain have "ichain (?lP, unlab_trs R) ?ls ?lt ?lsig" using ichain_imp_map_ichain by blast
  hence ichain2: "ichain (?lP, ?lR) ?ls ?lt ?lsig" using rtrancl_rstep_unlab_of_is_unlab by auto

  have cond: "lab_minimal_cond ?lR m ?ls ?lt ?lsig" 
    unfolding lab_minimal_cond_def
  proof (intro allI impI)
    fix n i
    assume le: "n \<le> m"
    with chain have "SN_elt (rstep ((unlab_of_trs ^^ (Suc n)) R)) ((unlab_term ^^ (Suc n)) (t i \<cdot> \<sigma> i))"    
      by (simp only: min_lab_ichain.simps lab_minimal_cond_def)
    thus "SN_elt (rstep ((unlab_of_trs ^^ n) ?lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))"
      by (simp add: funpow_swap1)
  qed

  from ichain2 cond have "min_lab_ichain (?lP, ?lR) m ?ls ?lt ?lsig" by (auto simp: lab_minimal_cond_def)
  with finite show False unfolding finite_lab_dpp_def by blast
qed


definition sound_lab_proc 
  where "sound_lab_proc proc \<equiv> \<forall> P R m rP rR rm. (proc P R m rP rR rm \<longrightarrow> (finite_lab_dpp (rP, rR) rm \<longrightarrow> finite_lab_dpp (P, R) m))"

definition sound_lab_proc_impl
  where "sound_lab_proc_impl proc \<equiv> \<forall> P R m rP rR rm. (proc P R m (computeRuleMap R) rP rR rm \<longrightarrow> (finite_lab_dpp (set rP, set rR) rm \<longrightarrow> finite_lab_dpp (set P, set R) m))"

definition sound_lab_only_pairs_proc_impl
  where "sound_lab_only_pairs_proc_impl proc \<equiv> \<forall> P R m rP rm. (proc P R m (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP rm \<longrightarrow> (finite_lab_dpp (set rP, set R) rm \<longrightarrow> finite_lab_dpp (set P, set R) m))"


lemma decr_SN_to_SN_init: 
  assumes "decrSN (init_lab_trs R :: (('f,'l)lab,'v)trs) m" (is "decrSN ?IR _")
  shows "SN (rstep (init_lab_trs R :: (('f,'l)lab,'v)trs))"
proof -
  have SN2: "\<forall>n. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) ?IR)))" 
  proof
    fix n
    have "decr_of_trs ((unlab_of_trs ^^ n) ?IR) = decr_of_trs ?IR" by (simp only: unlab_of_init_is_init)
    also have "\<dots> = {}" by (rule decr_of_init_is_empty)
    finally show "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) ?IR)))" by auto
  qed
  with assms show ?thesis unfolding decrSN_def by auto
qed

lemma decr_SN_to_SN :
  assumes decr: "decrSN (init_lab_trs R) m"
  shows "SN (rstep R)"
  by (rule init_lab_sound_trs, rule decr_SN_to_SN_init, rule decr)

lemma decr_SN_to_SN_II :
  assumes decr: "decrSN R 0"
  shows "SN (rstep R)"
using assms unfolding decrSN_def by auto

lemma SN_to_decr_SN :
  assumes "SN (rstep R)"
  shows "decrSN R m"
using assms unfolding decrSN_def by auto


lemma weak_ci_tt_to_sound_lab_tt: assumes ci_proc: "weak_ci_tt tt"
  shows "sound_lab_tt (\<lambda> R m rR rm. tt R rR \<and> m = rm)"
unfolding sound_lab_tt_def decrSN_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume m: "m = rm" and cond: "tt R rR" and step: "(\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))) \<longrightarrow> SN (rstep rR)" and SN1: "\<forall> n < m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))"
  from cond ci_proc have subset: "rR \<subseteq>ci R" and SN: "SN (rstep rR) \<longrightarrow> SN (rstep R)" unfolding weak_ci_tt_def by auto
  have "\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))"
  proof (intro allI impI)
    fix n
    assume "n < rm"
    with m have SN2: "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))" using SN1 by auto
    show "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))"
      by (rule SN_subset, rule SN2, rule rstep_ci_mono, rule decr_of_trs_mono_ci, rule funpow_ci_mono, rule unlab_of_trs_mono_ci, auto simp: subset)
  qed
  with step have "SN (rstep rR)" by auto
  with SN show "SN (rstep R)" ..
qed

lemma weak_ci_tt_impl_to_sound_lab_tt: assumes ci_proc: "weak_ci_tt_impl tt"
  shows "sound_lab_tt_impl (\<lambda> R m rR rm. tt R rR \<and> m = rm)"
unfolding sound_lab_tt_impl_def decrSN_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume m: "m = rm" and cond: "tt R rR" and step: "(\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set rR))))) \<longrightarrow> SN (rstep (set rR))" and SN1: "\<forall> n < m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set R))))"
  from cond ci_proc have subset: "set rR \<subseteq>ci set R" and SN: "SN (rstep (set rR)) \<longrightarrow> SN (rstep (set R))" unfolding weak_ci_tt_impl_def by auto
  have "\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set rR))))"
  proof (intro allI impI)
    fix n
    assume "n < rm"
    with m have SN2: "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set R))))" using SN1 by auto
    show "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set rR))))"
      by (rule SN_subset, rule SN2, rule rstep_ci_mono, rule decr_of_trs_mono_ci, rule funpow_ci_mono, rule unlab_of_trs_mono_ci, auto simp: subset)
  qed
  with step have "SN (rstep (set rR))" by auto
  with SN show "SN (rstep (set R))" ..
qed

lemma ci_proc_to_sound_lab_proc: assumes ci_proc: "ci_proc proc"
  shows "sound_lab_proc (\<lambda> P R m rP rR rm. m = rm \<and> proc P R rP rR)"
unfolding sound_lab_proc_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume m: "m = rm" and cond: "proc P R rP rR" and finite: "finite_lab_dpp (rP, rR) rm"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (P,R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "rR \<subseteq>ci R \<and> ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_proc_def by force
    have "min_lab_ichain (rP,rR) rm (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = R and R = rR] chain ml m by auto
    hence "\<not> finite_lab_dpp (rP,rR) rm" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed

(* the following two techniques are only possible due to the explicit index m,
   they allow to use non-chain-identifying processors even in the labeled semantics,
   although afterwards one cannot perform unlabeling any more
*)
lemma sound_proc_to_sound_lab_proc: assumes sound_proc: "sound_proc proc"
  shows "sound_lab_proc (\<lambda> P R m rP rR rm. rm = 0 \<and> proc P R rP rR)"
unfolding sound_lab_proc_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume rm: "rm = 0" and cond: "proc P R rP rR" and finite: "finite_lab_dpp (rP, rR) rm"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (P,R) s t \<sigma>" by auto
    hence "\<not> finite_dpp (P,R)" unfolding finite_dpp_def by auto
    with sound_proc cond have "\<not> finite_dpp (rP, rR)" unfolding sound_proc_def by auto
    from this obtain s t \<sigma> where "min_ichain (rP, rR) s t \<sigma>" unfolding finite_dpp_def by auto
    hence "min_lab_ichain (rP, rR) 0 s t \<sigma>" by (auto simp: minimal_cond_def lab_minimal_cond_def)
    hence "\<not> finite_lab_dpp (rP,rR) 0" unfolding finite_lab_dpp_def by blast
    with finite rm show False by simp
  qed
qed

lemma sound_tt_impl_to_sound_lab_tt_impl: assumes sound: "sound_tt_impl tt"
  shows "sound_lab_tt_impl (\<lambda> R m rR rm. rm = 0 \<and> tt R rR)"
unfolding sound_lab_tt_impl_def
proof (intro allI impI, simp)
  fix R m rR rm
  assume tt: "rm = 0 \<and> tt R rR" and decr: "decrSN (set rR) 0"
  from decr have "SN (rstep (set rR))" unfolding decrSN_def by force
  with tt sound have "SN (rstep (set R))" unfolding sound_tt_impl_def by auto
  thus "decrSN (set R) m" by (rule SN_to_decr_SN)
qed
 

lemma sound_proc_to_sound_lab_proc_impl: assumes sound_proc: "sound_proc_impl proc"
  shows "sound_lab_proc_impl (\<lambda> P R m rmap rP rR rm. rm = 0 \<and> proc P R rmap rP rR)"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume rm: "rm = 0" and cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_lab_dpp (set rP, set rR) rm"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    hence "\<not> finite_dpp (set P,set R)" unfolding finite_dpp_def by auto
    with sound_proc cond have "\<not> finite_dpp (set rP, set rR)" unfolding sound_proc_impl_def by auto
    from this obtain s t \<sigma> where "min_ichain (set rP, set rR) s t \<sigma>" unfolding finite_dpp_def by auto
    hence "min_lab_ichain (set rP, set rR) 0 s t \<sigma>" by (auto simp: minimal_cond_def lab_minimal_cond_def)
    hence "\<not> finite_lab_dpp (set rP,set rR) 0" unfolding finite_lab_dpp_def by blast
    with finite rm show False by simp
  qed
qed 
 
lemma ci_proc_to_sound_lab_proc_impl[simp]: assumes ci_proc: "ci_proc_impl proc"
  shows "sound_lab_proc_impl (\<lambda> P R m rmap rP rR rm. m = rm \<and> proc P R rmap rP rR)"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume m: "m = rm" and cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_lab_dpp (set rP, set rR) rm"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "set rR \<subseteq>ci set R \<and> ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_proc_impl_def by force
    have "min_lab_ichain (set rP,set rR) rm (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = "set R" and R = "set rR"] chain ml m by auto
    hence "\<not> finite_lab_dpp (set rP,set rR) rm" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 

lemma ci_proc_to_sound_lab_only_pairs_proc_impl[simp]: assumes ci_proc: "ci_only_pairs_proc_impl proc"
  shows "sound_lab_only_pairs_proc_impl (\<lambda> P R m rmap1 rmap2 rP rm. m = rm \<and> proc P R rmap1 rmap2 rP)"
unfolding sound_lab_only_pairs_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m rP rm
  assume m: "m = rm" and cond: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP" and finite: "finite_lab_dpp (set rP, set R) rm"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_only_pairs_proc_impl_def by blast
    have "min_lab_ichain (set rP,set R) rm (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = "set R" and R = "set R"] chain ml m ci_subset_refl[of "set R"] by auto
    hence "\<not> finite_lab_dpp (set rP,set R) rm" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 


lemma apply_ci_only_pairs_lab_proc: assumes ci: "ci_only_pairs_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set R) m"
  and proc: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from ci have "sound_lab_only_pairs_proc_impl (\<lambda> P R m rmap1 rmap2 rP rm. m = rm \<and> proc P R rmap1 rmap2 rP)" by simp
  thus ?thesis unfolding sound_lab_only_pairs_proc_impl_def using fin proc by blast
qed

lemma apply_ci_lab_proc: assumes ci: "ci_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR) m"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from ci have "sound_lab_proc_impl (\<lambda> P R m rmap rP rR rm. m = rm \<and> proc P R rmap rP rR)" by simp
  thus ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_lab_proc: assumes sound: "sound_lab_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR) rm"
  and proc: "proc P R m (computeRuleMap R) rP rR rm"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from sound show ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_proc: assumes sound: "sound_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR) 0"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from sound_proc_to_sound_lab_proc_impl[OF sound]
  show ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_lab_tt: assumes sound: "sound_lab_tt_impl tt"
  and fin: "decrSN (set rR) rm"
  and proc: "tt R m rR rm"
  shows "decrSN (set R) m"
proof -
  from sound
  show ?thesis unfolding sound_lab_tt_impl_def using fin proc sound by blast
qed 

lemma apply_weak_ci_lab_tt: assumes ci: "weak_ci_tt_impl tt"
  and fin: "decrSN (set rR) m"
  and proc: "tt R rR"
  shows "decrSN (set R) m"
  using apply_sound_lab_tt[OF weak_ci_tt_impl_to_sound_lab_tt[OF ci] fin] proc by simp

lemma apply_ci_lab_tt: assumes ci: "ci_tt_impl tt"
  and fin: "decrSN (set rR) m"
  and proc: "tt R rR"
  shows "decrSN (set R) m"
  using apply_weak_ci_lab_tt[OF ci_tt_to_weak_ci_tt_impl[OF ci] fin] proc  by simp



text {* Now let us head for the semantic part *}

types ('v, 'c) assign = "'v \<Rightarrow> 'c"
types ('f, 'c, 'l) inter = "('f, 'l) lab \<Rightarrow> 'c list \<Rightarrow> 'c"
types ('f, 'l)labels = "('f, 'l) lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab) set"
types ('f, 'c, 'l)label = "('f, 'l) lab \<Rightarrow> 'c list \<Rightarrow> ('l + ('f, 'l) lab list)"

fun
  label :: "('f, 'l) lab \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> ('f, 'l) lab"
where
  "label f (Inl l) = Lab f l" |
  "label f (Inr l) = FunLab f l"

fun eval_lab :: "('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('v,'c)assign \<Rightarrow> (('f,'l)lab,'v)term \<Rightarrow> ('c \<times> (('f,'l)lab,'v)term)" 
where "eval_lab I L \<alpha> (Var x) = (\<alpha> x, Var x)"
  |   "eval_lab I L \<alpha> (Fun f ts) = 
     (let clts = map (eval_lab I L \<alpha>) ts in 
     (let cs = map fst clts in 
     (let c = I f cs in
     (let lts = map snd clts in 
     (let l = L f cs in
     (c, Fun (label f l) lts))))))"

abbreviation eval where "eval I L \<alpha> t \<equiv> fst (eval_lab I L \<alpha> t)"
abbreviation lab where "lab I L \<alpha> t \<equiv> snd (eval_lab I L \<alpha> t)"
definition wf_assign where "wf_assign C \<alpha> \<equiv> \<forall> x. \<alpha> x \<in> C"
definition lab_rule where "lab_rule I L C lr \<equiv> {(lab I L \<alpha> (fst lr), lab I L \<alpha> (snd lr)) | \<alpha> . wf_assign C \<alpha>}"
abbreviation lab_trs where "lab_trs I L C R \<equiv> \<Union>(lab_rule I L C ` R)"
definition undo_lab where "undo_lab f \<equiv> (fst f, tl (snd f))"
definition wf_inter where "wf_inter I C \<equiv> \<forall> f cs. set cs \<subseteq> C \<longrightarrow> I f cs \<in> C"
abbreviation model_rule where "model_rule I L C l r \<equiv> \<forall> \<alpha>. wf_assign C \<alpha> \<longrightarrow> eval I L \<alpha> l  = eval I L \<alpha> r"
definition model where "model I L C R \<equiv> \<forall> (l,r) \<in> R. model_rule I L C l r"

lemma [simp]: "unlab (label f l) = f" by (cases l, auto)

lemma unlab_term: "unlab_term (lab I L \<alpha> t) = t"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by (auto simp: Let_def, induct ts, auto)
qed
lemma [simp]: "remove_lab (label f l) = remove_lab f" by (cases l, auto)

lemma unlab_rule_I: 
   "unlab_trs (lab_rule I L C rule) \<subseteq> {rule}" (is "?rule \<subseteq> {rule}")
proof
  fix ulrule
  assume ulrule: "ulrule \<in> ?rule"
  show "ulrule \<in> {rule}"
  proof (cases ulrule)
    case (Pair ull ulr)
    with ulrule have "(ull,ulr) \<in> ?rule" by auto
    from this obtain ll lr where lrule: "ull = unlab_term ll \<and> ulr = unlab_term lr \<and> (ll,lr) \<in> lab_rule I L C rule" 
      by force
    from this obtain l r \<alpha> where "rule = (l,r) \<and> ll = lab I L \<alpha> l \<and> lr = lab I L \<alpha> r" unfolding lab_rule_def by (cases rule, auto)
    with Pair ulrule lrule show ?thesis by (simp add: unlab_term)
  qed
qed

fun
  get_lab :: "('f, 'l) lab \<Rightarrow> ('l + ('f, 'l) lab list)"
where
  "get_lab (Lab f l) = Inl l" |
  "get_lab (FunLab f l) = Inr l"

definition
  decr_of_ord :: "('l + ('f, 'l) lab list) ars \<Rightarrow> (('f, 'l) lab, 'v) trs"
where
  "decr_of_ord gr \<equiv> {(l,r) | l r. unlab_term l = unlab_term r \<and> \<not> is_var l \<and> \<not> is_var r \<and> (get_lab (the (root l)),get_lab (the (root r))) \<in> gr}"

definition
  R_subset_lab_decr :: "('l + ('f, 'l) lab list) ars \<Rightarrow> (('f, 'l) lab, 'v) trs \<Rightarrow> (('f, 'l) lab, 'v) trs \<Rightarrow> bool"
where
  "R_subset_lab_decr gr R rR \<equiv> \<forall> (l,r) \<in> rR. (unlab_term l, unlab_term r) \<in> R \<and> (l = r \<or> unlab_term l \<noteq> unlab_term r) \<or> (l,r) \<in> decr_of_ord gr"

locale sl_interpr = 
  fixes C :: "('c::show) set"
  and  c :: "'c"
  and  I :: "('f::show,'c::show,'l::show)inter"
  and  gr :: "('l + ('f, 'l) lab list) ars"
  and  L :: "('f,'c,'l)label"
(*  and  LS :: "('f,'l)labels" *)
  assumes c: "c \<in> C"
  and     wf_I: "wf_inter I C"  
  and Decr_SN: "SN (rstep (decr_of_ord gr :: (('f,'l)lab,'v::show)trs))"
begin
abbreviation Eval where "Eval \<equiv> eval I L"
abbreviation Lab where "Lab \<equiv> lab I L"
abbreviation Lab_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs" where "Lab_trs \<equiv> lab_trs I L C"
abbreviation wf_ass where "wf_ass \<equiv> wf_assign C"
definition subst_ass where "subst_ass \<alpha> \<sigma> \<equiv> \<lambda> x. Eval \<alpha> (get_subst \<sigma> x)"
definition lab_subst where "lab_subst \<alpha> \<sigma> \<equiv> Subst (\<lambda> x. Lab \<alpha> (get_subst \<sigma> x))"
abbreviation Decr :: "(('f,'l)lab,'v)trs" where "Decr \<equiv> decr_of_ord gr"
abbreviation R_subset_lab_Decr :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs \<Rightarrow> bool" where "R_subset_lab_Decr \<equiv> R_subset_lab_decr gr"
definition default_ass where "default_ass x = c"

lemma wf_default_ass: "wf_ass default_ass" 
  unfolding default_ass_def wf_assign_def using c by auto

abbreviation LAB where "LAB t \<equiv> Lab default_ass t"

lemma unlab_rule_II: 
   "{rule} \<subseteq> unlab_trs (lab_rule I L C rule)" (is "?rule \<supseteq> {rule}")
proof (cases rule)
  case (Pair l r)
  with wf_default_ass
  have "(LAB l, LAB r) \<in> lab_rule I L C rule" (is "(?l,?r) \<in> _") 
    unfolding lab_rule_def using Pair by auto
  hence "(unlab_term ?l, unlab_term ?r) \<in> ?rule"  by force
  hence "rule \<in> ?rule" using Pair by (simp add: unlab_term)
  thus ?thesis by auto
qed

lemma unlab_rule[simp]: 
   "unlab_trs (lab_rule I L C rule) = {rule}"
using unlab_rule_I unlab_rule_II by blast

(*
lemma remove_label_term[simp]: "remove_lab_term (Lab \<alpha> t) = remove_lab_term t"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by (auto simp: Let_def map_compose[symmetric])
qed  

lemma remove_label_rule: "remove_lab_trs (lab_rule I L C rule) = {remove_lab_rule rule}" (is "?rule = {?rrule}")
proof 
  show "?rule \<subseteq> {?rrule}"
  proof
    fix ulrule
    assume ulrule: "ulrule \<in> ?rule"
    show "ulrule \<in> {?rrule}"
    proof (cases ulrule)
      case (Pair ull ulr)
      with ulrule have "(ull,ulr) \<in> ?rule" by auto
      from this obtain ll lr where lrule: "ull = remove_lab_term ll \<and> ulr = remove_lab_term lr \<and> (ll,lr) \<in> lab_rule I L C rule" 
	by force
      from this obtain l r \<alpha> where "rule = (l,r) \<and> ll = Lab \<alpha> l \<and> lr = Lab \<alpha> r" unfolding lab_rule_def by (cases rule, auto)
      with Pair ulrule lrule show ?thesis by (simp)
    qed
  qed
next
  show "{?rrule} \<subseteq> ?rule"
  proof (cases rule)
    case (Pair l r)
    with wf_default_ass have "(LAB l, LAB r) \<in> lab_rule I L C rule" unfolding lab_rule_def by auto    
    hence "(remove_lab_term (LAB l), remove_lab_term (LAB r)) \<in> ?rule"  by force
    hence "?rrule \<in> ?rule" using Pair by (simp)
    thus ?thesis by auto
  qed
qed

lemma remove_label_trs: fixes R :: "(('f,'l)lab,'v)trs"
  shows "remove_lab_trs (Lab_trs R) = remove_lab_trs R"
proof (rule set_ext)
  fix rule :: "('f,'v)rule"
  let ?R = "remove_lab_trs R :: ('f,'v)trs"
  let ?ULR = "remove_lab_trs (Lab_trs R) :: ('f,'v)trs"
  show "(rule \<in> ?ULR) = (rule \<in> ?R)"
  proof
    assume rule: "rule \<in> ?ULR"
    from this obtain rrule where cond: "rrule \<in> R" and "rule \<in> remove_lab_trs (lab_rule I L C rrule)" by auto
    hence "rule = remove_lab_rule rrule" using remove_label_rule[where ?rule = rrule] by auto
    thus "rule \<in> ?R" using cond map_trs_mono by auto
  next
    assume rule: "rule \<in> ?R"
    from this obtain rrule where cond: "rrule \<in> R" and "rule = remove_lab_rule rrule" by auto
    with remove_label_rule[where ?rule = rrule] have "rule \<in> remove_lab_trs (lab_rule I L C rrule)" by auto  
    with map_trs_mono rule cond show "rule \<in> ?ULR" by auto
  qed
qed

lemma remove_label_trs_2: "remove_lab_trs (unlab_trs R) = remove_lab_trs R"
proof -
    show ?thesis by (simp only: map_trs_comp, force simp: remove_unlab_is_remove)
qed
*)

lemma unlab_trs: "unlab_trs (Lab_trs R) = R"
proof (rule set_ext)
  fix rule
  let ?ULR = "unlab_trs (Lab_trs R)"
  show "(rule \<in> ?ULR) = (rule \<in> R)"
  proof
    assume rule: "rule \<in> R"
    from unlab_rule[where ?rule = rule] have "rule \<in> unlab_trs (lab_rule I L C rule)" by auto  
    with map_trs_mono rule show "rule \<in> ?ULR" by auto
  next
    assume rule: "rule \<in> ?ULR"
    from this obtain rrule where "rrule \<in> R \<and> rule \<in> unlab_trs (lab_rule I L C rrule)" by auto
    thus "rule \<in> R" using rule unlab_rule[where ?rule = rrule] by auto
  qed
qed

lemma SN_imp_SN_lab_term: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trs"
  assumes SN: "SN_elt (rstep R) t"
  shows "SN_elt (rstep (Lab_trs R)) (LAB t)"
proof -
  from SN have "SN_elt (rstep (unlab_trs (Lab_trs R))) (unlab_term (LAB t))" 
    by (simp only: unlab_trs unlab_term)
  thus ?thesis by (rule SN_map_imp_SN)
qed    

lemma SN_imp_SN_lab_term_2: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trs"
  assumes SN: "SN_elt (rstep R) t"
  and subset: "unlab_trs LR \<subseteq> R"
  shows "SN_elt (rstep LR) (LAB t)"
proof -
  from SN have "SN_elt (rstep R) (unlab_term (LAB t))" 
    by (simp only: unlab_trs unlab_term)  
  show ?thesis 
    by (rule SN_map_imp_SN[where fg = unlab], simp only: unlab_term, rule SN_elt_subset, rule SN, rule rstep_mono, rule subset)
qed


lemma eval_subst[simp] :
  "Eval \<alpha> (t \<cdot> \<sigma>) = Eval (subst_ass \<alpha> \<sigma>) t"
proof (induct t, cases \<sigma>, simp add: subst_ass_def)
  case (Fun f ts)
  let ?beta = "subst_ass \<alpha> \<sigma>"
  have evalInd: "map (\<lambda> x. Eval \<alpha> (x \<cdot> \<sigma>)) ts = map (Eval ?beta) ts" using Fun  by auto
  have "Eval \<alpha> (Fun f ts \<cdot> \<sigma>) = Eval \<alpha> (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" by simp
  also have "\<dots> = Eval ?beta (Fun f ts)" by (simp add: Let_def o_def evalInd)
  finally show ?case .
qed

lemma lab_subst[simp] :
  "Lab \<alpha> (t \<cdot> \<sigma>) = Lab (subst_ass \<alpha> \<sigma>) t \<cdot> lab_subst \<alpha> \<sigma>"
  by (induct t, cases \<sigma>, simp add: lab_subst_def, simp add: Let_def o_def)

lemma wf_term[simp]: assumes "wf_ass \<alpha>"
  shows "Eval \<alpha> t \<in> C"
proof (induct t)
  case (Var x)
  with assms show ?case by (auto simp: wf_assign_def)
next
  case (Fun f ts)
  let ?cs = "map (Eval \<alpha>) ts"
  let ?c = "I f ?cs"
  from Fun have "set ?cs \<subseteq> C" by auto
  with wf_I have "?c \<in> C" unfolding wf_inter_def by (best)
  thus ?case by (simp add: Let_def o_def)
qed

lemma wf_ass_subst_ass[simp]: assumes "wf_ass \<alpha>"
  shows "wf_ass (subst_ass \<alpha> \<sigma>)"
using assms unfolding subst_ass_def by (simp add: wf_assign_def)


lemma [simp]: "Eval \<alpha> (t\<cdot>\<sigma>) = Eval (subst_ass \<alpha> \<sigma>) t" by (cases \<sigma>) (auto simp: subst_ass_def)

lemma sem_rewrite: assumes step: "(s,t) \<in> rstep R"
  and model: "model I L C R"
  and wfass: "wf_ass \<alpha>"
  shows "Eval \<alpha> s = Eval \<alpha> t"
using step
proof
  fix l r C \<sigma> assume in_R: "(l, r) \<in> R" and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>"
  let ?beta = "subst_ass \<alpha> \<sigma>"
  have wfbeta: "wf_ass ?beta" using wfass by auto
  have "Eval \<alpha> (l\<cdot>\<sigma>) = Eval ?beta l" by simp
  also have "\<dots> = Eval ?beta r" using in_R wfbeta model by (auto simp: model_def)
  also have "\<dots> = Eval \<alpha> (r\<cdot>\<sigma>)" by simp
  finally have root: "Eval \<alpha> (l\<cdot>\<sigma>) = Eval \<alpha> (r\<cdot>\<sigma>)" .
  have "Eval \<alpha> s = Eval \<alpha> (C\<langle>l\<cdot>\<sigma>\<rangle>)" by (simp add: s)
  also have "\<dots> = Eval \<alpha> (C\<langle>r\<cdot>\<sigma>\<rangle>)" using root by (induct C) (auto simp: Let_def)
  also have "\<dots> = Eval \<alpha> t" by (simp add: t)
  finally show ?thesis .
qed

lemma lab_rewrite: assumes step: "(s,t) \<in> rstep R"
  and model: "model I L C R"
  and wfass: "wf_ass \<alpha>"
  shows "(Lab \<alpha> s, Lab \<alpha> t) \<in> rstep (Lab_trs R)"
using step proof
  fix l r C \<sigma>
  assume rule: "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  have root: "(Lab \<alpha> (l \<cdot> \<sigma>), Lab \<alpha> (r \<cdot> \<sigma>)) \<in> rstep (Lab_trs R)"
  proof -
    let ?beta = "subst_ass \<alpha> \<sigma>"
    let ?delt = "lab_subst \<alpha> \<sigma>"
    have wfbeta: "wf_ass ?beta" using wfass by auto
    from rule wfbeta
      have "(Lab ?beta l, Lab ?beta r) \<in> (Lab_trs R)" by (simp add: lab_rule_def, force)
    hence "(Lab ?beta l \<cdot> ?delt, Lab ?beta r \<cdot> ?delt) \<in> CS (Lab_trs R)" by (rule subst_closureI)
    from ctxt_closureI[OF this, where C = \<box>] 
      have "(Lab ?beta l \<cdot> ?delt, Lab ?beta r \<cdot> ?delt) \<in> rstep (Lab_trs R)"
      unfolding rstep_CC_CS by simp
    thus ?thesis by simp
  qed
  have id: "(Lab \<alpha> s, Lab \<alpha> t) = (Lab \<alpha> C\<langle>l \<cdot> \<sigma>\<rangle>, Lab \<alpha> C\<langle>r \<cdot> \<sigma>\<rangle>)" using s t by auto
  have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rstep R" using rule ctxt_closureI[where C = \<box>] by auto
  hence "Eval \<alpha> (l \<cdot> \<sigma>) = Eval \<alpha> (r \<cdot> \<sigma>)" using model wfass by (rule sem_rewrite)
  have "(Lab \<alpha> C\<langle>l \<cdot> \<sigma>\<rangle>, Lab \<alpha> C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep (Lab_trs R)" 
  proof (induct C)
    case Hole
    show ?case using root by simp
  next
    case (More f ss C ts)
    let ?D = "More (label f (L f (map (Eval \<alpha>) (ss @ C\<langle>l \<cdot> \<sigma>\<rangle> # ts)))) (map (Lab \<alpha>) ss) \<box> (map (Lab \<alpha>) ts)"
    from rule have "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R" using ctxt_closureI by auto
    hence "Eval \<alpha> C\<langle>l \<cdot> \<sigma>\<rangle> = Eval \<alpha> C\<langle>r \<cdot> \<sigma>\<rangle>" using model wfass by (rule sem_rewrite)
    hence id: "(Lab \<alpha> (More f ss C ts)\<langle>l \<cdot> \<sigma>\<rangle>,Lab \<alpha> (More f ss C ts)\<langle>r \<cdot> \<sigma>\<rangle>) =
      (?D\<langle>Lab \<alpha> C\<langle>l \<cdot> \<sigma>\<rangle>\<rangle>, ?D\<langle>Lab \<alpha> C\<langle>r \<cdot> \<sigma>\<rangle>\<rangle>)" (is "_ = ?p2") 
      by (auto simp: Let_def o_def)
    with More have "?p2 \<in> CC (rstep (Lab_trs R))" using ctxt_closureI[where C = ?D] by best
    with ctxt_closed_rstep id show ?case unfolding ctxt_closed_def by auto
  qed
  with s t show ?thesis by auto
qed

lemma rewrites_imp_lab_rewrites: 
  assumes "(s,t) \<in> (rstep R)^*"
  and model: "model I L C R"
  shows "(LAB s, LAB t) \<in> (rstep (Lab_trs R))^*"
using assms unfolding rstep_CC_CS
proof (induct, simp)
  case (step y z)
  with lab_rewrite[where s=y and t=z and \<alpha> = default_ass] model wf_default_ass
    show ?case unfolding rstep_CC_CS by force
qed
 
lemma SN_lab_imp_SN_term: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trs"
  assumes SN: "SN_elt (rstep (Lab_trs R)) (LAB t)" 
  and model: "model I L C R"
  shows "SN_elt (rstep R) t"
proof (rule ccontr)
  assume "\<not> SN_elt (rstep R) t"
  from this obtain f where inf: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" unfolding SN_elt_def  by auto
  obtain g where g: "\<And> i. g i = LAB (f i)" by auto
  have one: "g 0 = LAB t" using inf g by auto
  have two: "\<forall> i. (g i, g (Suc i)) \<in> rstep (Lab_trs R)" 
    by (rule allI, simp only: g, rule lab_rewrite, auto simp: inf wf_default_ass model g)
  from one two SN show False unfolding SN_elt_def by auto
qed    

lemma rel_SN_lab_imp_rel_SN: fixes R  and S :: "(('f,'l)lab,'v)trs" 
  assumes SN: "rel_SN_alt (Lab_trs R, Lab_trs S)" 
  and model: "model I L C (R \<union> S)"
  shows "rel_SN_alt (R,S)"
proof(unfold rel_SN_alt.simps, intro allI impI)
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep (R \<union> S)"
  obtain g where g: "\<And> i. g i = LAB (f i)" by auto
  from model have rmodel: "model I L C R" unfolding model_def by auto
  from steps have "\<And> i. (g i, g (Suc i)) \<in> rstep (Lab_trs R \<union> Lab_trs S)" using lab_rewrite[OF _ model wf_default_ass]
    unfolding g by auto
  with SN obtain i where i: "\<forall> j \<ge> i. (g j, g (Suc j)) \<notin> rstep (Lab_trs R)" unfolding rel_SN_alt.simps by force
  show "\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R"
  proof 
    show "\<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R" using lab_rewrite[OF _ rmodel wf_default_ass] i unfolding g  by blast
  qed
qed

lemma decr_of_lab_is_empty: "decr_of_trs (Lab_trs R) = {}"
unfolding decr_of_trs_def by (auto simp: lab_rule_def unlab_term)

lemma gr_irreflexive: "(x,x) \<notin> gr"
proof
  fix l :: "('l + ('f, 'l) lab list)"
  assume in_gr: "(l,l) \<in> gr"
  show False
  proof (cases l)
    case (Inl ll)    
    let ?t = "Fun (SemanticLabeling.Lab (UnLab f) ll) [] :: (('f,'l)lab,'v)term"
    from in_gr Inl have "(?t, ?t) \<in> Decr"
      unfolding decr_of_ord_def by auto
    from this obtain t :: "(('f,'l)lab,'v)term" where "(t,t) \<in> Decr" by force
    hence t: "(t,t) \<in> rstep Decr" by auto
    let ?f = "\<lambda> i :: nat. t"
    from t have "\<forall> i. (?f i, ?f (Suc i)) \<in> rstep Decr" by auto
    with Decr_SN show False unfolding SN_defs by force
  next
    case (Inr ll)    
    let ?t = "Fun (FunLab (UnLab f) ll) [] :: (('f,'l)lab,'v)term"
    from in_gr Inr have "(?t, ?t) \<in> Decr"
      unfolding decr_of_ord_def by auto
    from this obtain t :: "(('f,'l)lab,'v)term" where "(t,t) \<in> Decr" by force
    hence t: "(t,t) \<in> rstep Decr" by auto
    let ?f = "\<lambda> i :: nat. t"
    from t have "\<forall> i. (?f i, ?f (Suc i)) \<in> rstep Decr" by auto
    with Decr_SN show False unfolding SN_defs by force
  qed
qed

lemma decr_of_decr_is_decr: "decr_of_trs Decr = Decr"
  using gr_irreflexive
  unfolding decr_of_trs_def decr_of_ord_def by auto
  
lemma lab_sound_tt: 
  shows "sound_lab_tt (\<lambda> R m rR rm. rm = Suc m \<and> Lab_trs R \<subseteq> rR \<and> R_subset_lab_Decr R rR \<and> model I L C R)"
unfolding sound_lab_tt_def 
proof (intro allI impI, (erule conjE)+)
  fix R :: "(('f,'l)lab,'v)trs" and m and rR :: "(('f,'l)lab,'v)trs" and rm
  assume m: "rm = Suc m" and subset: "Lab_trs R \<subseteq> rR" and decrSN: "decrSN rR rm" and subset2: "R_subset_lab_Decr R rR" and model: "model I L C R"
  from decrSN show "decrSN R m" 
    unfolding decrSN_def
  proof (intro impI)
    assume SN: "\<forall> n<m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))" (is "\<forall> n<m. ?invR n")
    assume imp: "(\<forall>n<rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))) \<longrightarrow> SN (rstep rR)" (is "(\<forall> n<rm. ?invLR n) \<longrightarrow> ?SNL")
    have "\<forall> n < rm. ?invLR n"
    proof (intro allI impI)
      fix n
      assume le: "n < rm"
      from subset2 have decr_subset: "decr_of_trs rR \<subseteq> Decr" unfolding R_subset_lab_decr_def decr_of_trs_def
	by auto
      show "?invLR n"
      proof (cases n)
	case (Suc mm)
	have "unlab_of_trs rR = R" unfolding unlab_of_trs_def 
	proof (rule set_ext, simp only: split_paired_all)
	  fix l r 
	  show "((l,r) \<in> unlab_trs (rR - decr_of_trs rR)) = ((l,r) \<in> R)"
	  proof
	    assume "(l,r) \<in> R"
	    hence "(LAB l, LAB r) \<in> Lab_trs R" by (force simp: lab_rule_def wf_default_ass)
	    with subset have one: "(LAB l, LAB r) \<in> rR" by auto
	    have "(LAB l, LAB r) \<notin> Decr" unfolding decr_of_ord_def using gr_irreflexive by (auto simp: unlab_term) 
	    with decr_subset one have "(LAB l, LAB r) \<in> rR - decr_of_trs rR" by force
	    hence "(unlab_term (LAB l), unlab_term (LAB r)) \<in> unlab_trs (rR - decr_of_trs rR)" by force
	    thus "(l,r) \<in> unlab_trs (rR - decr_of_trs rR)" by (simp only: unlab_term)
	  next
	    assume "(l,r) \<in> unlab_trs (rR - decr_of_trs rR)"
	    from this obtain ll lr where mem: "(ll,lr) \<in> rR - decr_of_trs rR" and id: "unlab_term ll = l \<and> unlab_term lr = r" by force
	    with subset2 have disj: "(l, r) \<in> R \<and> (ll = lr \<or> l \<noteq> r) \<or> (ll, lr) \<in> Decr" unfolding R_subset_lab_decr_def by auto
	    {
	      assume "\<not> ((l,r) \<in> R \<and> (ll = lr \<or> l \<noteq> r)) \<and> (ll,lr) \<in> Decr"
	      hence inDecr: "(ll,lr) \<in> Decr" and disj2: "(l,r) \<notin> R \<or> ll \<noteq> lr \<and> l = r" by blast+
	      from inDecr have "unlab_term ll = unlab_term lr" unfolding decr_of_ord_def by auto
	      with mem inDecr have "(ll,ll) \<in> Decr" unfolding decr_of_trs_def by force
	      with gr_irreflexive have False unfolding decr_of_ord_def by auto
	    }
	    with disj show "(l,r) \<in> R" by blast
	  qed
	qed
	hence id: "(unlab_of_trs ^^ (Suc mm)) rR = (unlab_of_trs ^^ mm) R" by (simp add: funpow_swap1)
	have "?invLR (Suc mm)" using le m SN Suc by (simp only: id)
	thus ?thesis using Suc by auto
      next
	case 0
	show ?thesis by (simp add: 0, rule SN_subset, rule Decr_SN, rule rstep_mono, rule decr_subset)
      qed
    qed
    with imp have ?SNL by auto
    show "SN (rstep R)"
      using SN_lab_imp_SN_term[OF _ model] SN_subset[OF `?SNL` rstep_mono[OF subset]]
      unfolding SN_defs by auto 
  qed
qed


definition sem_lab_proc :: "
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  bool"
  where "sem_lab_proc P R m lP lR rm \<equiv> rm = Suc m \<and> Lab_trs P \<subseteq> lP \<and> Lab_trs R \<subseteq> lR \<and> R_subset_lab_Decr R lR \<and> model I L C R"

lemma split_pairs_all: "(\<forall> x \<in> r :: 'a ars. P x) = (\<forall> a b . (a,b) \<in> r \<longrightarrow> P (a,b))"
  by (auto)

abbreviation lab_terms where "lab_terms s \<sigma> i \<equiv> Lab (subst_ass default_ass (\<sigma> i)) (s i)"
abbreviation lab_substs  where "lab_substs  \<sigma> i \<equiv> lab_subst default_ass (\<sigma> i)"

lemma sem_lab_proc_sound: 
  shows "sound_lab_proc sem_lab_proc"
unfolding sound_lab_proc_def
proof (intro allI impI, unfold sem_lab_proc_def, (erule conjE)+ )
  fix P R :: "(('f,'l)lab,'v)trs" and m and lP lR :: "(('f,'l)lab,'v)trs" and rm
  assume lP: "Lab_trs P \<subseteq> lP" and lR1: "Lab_trs R \<subseteq> lR" and subset2: "R_subset_lab_Decr R lR" and model: "model I L C R" and m: "rm = Suc m"
  from subset2 have lR2: "unlab_of_trs lR \<subseteq> R" unfolding R_subset_lab_decr_def unlab_of_trs_def decr_of_ord_def decr_of_trs_def by (auto simp: gr_irreflexive)
  let ?lP = "Lab_trs P"
  let ?lR = "Lab_trs R"
  assume finite: "finite_lab_dpp (lP, lR) rm"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto

    let ?ls = "lab_terms s \<sigma>"
    let ?lt = "lab_terms t \<sigma>"
    let ?lsig = "lab_substs \<sigma>"
    let ?lss = "\<lambda> i. Lab default_ass (s i \<cdot> \<sigma> i)"
    let ?lts = "\<lambda> i. Lab default_ass (t i \<cdot> \<sigma> i)"

    have s: "\<forall> i. ?lss i = ?ls i \<cdot> ?lsig i" by simp
    have t: "\<forall> i. ?lts i = ?lt i \<cdot> ?lsig i" by simp

    have rstepSub: "(rstep ?lR)^* \<subseteq> (rstep lR)^*" 
      by (rule rtrancl_mono, rule rstep_mono, rule lR1)

    from chain have "\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    hence "\<forall> i. (?lts i, ?lss (Suc i)) \<in> (rstep ?lR)^*" using rewrites_imp_lab_rewrites model by best
    with s t have "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*" by auto
    with rstepSub have steps: "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep lR)^*" by auto
    
    from chain have mem: "\<forall> i. (s i, t i) \<in> P" by simp
    have "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
    proof
      fix i
      have memI: "(s i, t i) \<in> P" using mem by simp
      have "wf_ass (subst_ass default_ass (\<sigma> i))" 
	by (rule wf_ass_subst_ass, rule wf_default_ass)
      hence "(?ls i, ?lt i) \<in> lab_rule I L C (s i, t i)"
	unfolding lab_rule_def by (auto)
      with memI show "(?ls i, ?lt i) \<in> ?lP" by auto
    qed
    hence memLab: "\<forall> i. (?ls i, ?lt i) \<in> lP" using lP by auto
    
    have minimal: "lab_minimal_cond lR rm ?ls ?lt ?lsig"
      unfolding lab_minimal_cond_def
    proof (intro allI impI)
      fix n i
      assume le: "n \<le> rm"
      show "SN_elt (rstep ((unlab_of_trs ^^ n) lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))" (is "?goal n")
      proof (cases n)
	case 0
	from subset2 have decr_subset: "decr_of_trs lR \<subseteq> Decr" unfolding R_subset_lab_decr_def decr_of_trs_def
	  by auto
	from chain have "SN_elt (rstep ((unlab_of_trs ^^ 0) R)) ((unlab_term ^^ 0) (t i \<cdot> \<sigma> i))" by (simp only: min_lab_ichain.simps lab_minimal_cond_def)
	hence SN: "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by simp
	have "SN_elt (rstep lR) (?lt i \<cdot> ?lsig i)" 
	  by (rule SN_unlab, rule SN_subset, rule Decr_SN, rule rstep_mono, rule decr_subset, simp only: lab_subst[symmetric] unlab_term, 
	  rule SN_elt_subset, rule SN, rule rstep_mono, rule lR2)	  
	with 0 show ?thesis by auto
      next
	case (Suc mm)
	with le m have "mm \<le> m" by auto
	with chain have cond: "SN_elt (rstep ((unlab_of_trs ^^ mm) R)) ((unlab_term ^^ mm) (t i \<cdot> \<sigma> i))" by (simp only: min_lab_ichain.simps lab_minimal_cond_def)
	have cond2: "SN_elt (rstep ((unlab_of_trs ^^ mm) (unlab_of_trs lR))) ((unlab_term ^^ mm) (t i \<cdot> \<sigma> i))" 
	  by (rule SN_elt_subset[OF cond rstep_mono[OF funpow_mono[OF unlab_of_trs_mono]]], auto simp: lR2)
	have "?goal (Suc mm)" 
	  by (simp add: funpow_swap1, simp only: map_subst_distrib[symmetric] lab_subst[symmetric], simp only: unlab_term cond2)	  
	with Suc show ?thesis by auto
      qed
    qed
	
    from steps minimal memLab have "min_lab_ichain (lP,lR) rm ?ls ?lt ?lsig" by auto
    with finite show False unfolding finite_lab_dpp_def by blast
  qed
qed

end

datatype arithFun = Arg nat | Const nat | Sum "arithFun list" | Max "arithFun list" | Min "arithFun list" | Prod "arithFun list"

text {* An alternative induction scheme for arithFuns. *}
lemma
  fixes P :: "arithFun \<Rightarrow> bool"
  assumes "\<And>n. P(Arg n)" and "\<And>n. P(Const n)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Sum fs)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Max fs)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Min fs)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Prod fs)"
  shows arithFun_induct[case_names Arg Const Sum Max Min Prod,induct type: arithFun]: "P f"
    and "\<And>f. f \<in> set fs \<Longrightarrow> P f"
proof (induct f and fs)
  fix n::nat show "P(Arg n)" by (simp add: assms)
next
  fix n::nat show "P(Const n)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<forall>f\<in>set fs. P f" thus "P(Sum fs)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<forall>f\<in>set fs. P f" thus "P(Max fs)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<forall>f\<in>set fs. P f" thus "P(Min fs)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<And>f. f \<in> set fs \<Longrightarrow> P f" thus "P(Prod fs)" by (simp add: assms)
next
  show "\<forall>f\<in>set []. P f" by simp
next
  fix f fs assume "P f" and "\<forall>f\<in>set fs. P f" thus "\<forall>g\<in>set(f#fs). P g" by simp
next
  show "\<forall>f\<in>set []. P f" by simp
next
  fix f fs assume "P f" and "\<forall>f\<in>set fs. P f" thus "\<forall>g\<in>set(f#fs). P g" by simp
next
  show "\<forall>f\<in>set []. P f" by simp
next
  fix f fs assume "P f" and "\<forall>f\<in>set fs. P f" thus "\<forall>g\<in>set(f#fs). P g" by simp
next
  fix f::"arithFun" assume "f \<in> set []" thus "P f" by simp
next
  fix f fs g assume "P f" and "\<And>f. f \<in> set fs \<Longrightarrow> P f" and "g \<in> set(f#fs)" thus "P g" by auto
qed

instantiation arithFun :: "show" begin
  fun shows_arithFun :: "arithFun \<Rightarrow> shows"
  where
    "shows_arithFun (Arg i) = (''x_'' +#+ shows (Suc i))" |
    "shows_arithFun (Const n) = shows n" |
    "shows_arithFun (Sum fs) = shows_list_gen shows_arithFun [] [] '' + '' [] fs" |
    "shows_arithFun (Prod fs) = shows_list_gen shows_arithFun [] [] '' * '' [] fs" |
    "shows_arithFun (Min fs) = shows_list_gen shows_arithFun [] ''min('' '','' '')'' fs" |
    "shows_arithFun (Max fs) = shows_list_gen shows_arithFun [] ''max('' '','' '')'' fs"
  definition "shows_prec (d::nat) f = shows_arithFun f"
  definition "shows_list (as::arithFun list) \<equiv> shows_list_aux shows as"

  lemma assoc_arithFun: "shows_prec d (f::arithFun) r @ s = shows_prec d f (r @ s)"
  proof (induct f arbitrary: r s)
    case (Arg n) show ?case unfolding shows_arithFun.simps shows_string_def o_def assoc[symmetric]
      unfolding shows_prec_arithFun_def by simp
  next
    case (Const n) show ?case unfolding shows_arithFun.simps shows_string_def o_def assoc[symmetric]
      unfolding shows_prec_arithFun_def by simp
  next
    case (Sum fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  next
    case (Max fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  next
    case (Min fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  next
    case (Prod fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  qed
  lemma assoc_arithFun_elt: "\<forall>(f::arithFun)\<in>set fs. shows_prec d f r @ s = shows_prec d f (r @ s)"
    using assoc_arithFun by auto
  instance by
    (intro_classes,rule assoc_arithFun,unfold shows_list_arithFun_def)
    (rule shows_list_aux_assoc[OF assoc_arithFun_elt])
end

fun take_default where "take_default def [] _ = def" | "take_default _ (x # xs) 0 = x" | "take_default def (_ # xs) (Suc i) = take_default def xs i"

fun eval_arithFun_unbound :: "nat list \<Rightarrow> arithFun \<Rightarrow> nat"
where "eval_arithFun_unbound nats (Arg i) = take_default 0 nats i"
    | "eval_arithFun_unbound nats (Const n) = n"
    | "eval_arithFun_unbound nats (Sum []) = 0"
    | "eval_arithFun_unbound nats (Sum (f # fs)) =  (eval_arithFun_unbound nats f + eval_arithFun_unbound nats (Sum fs))"
    | "eval_arithFun_unbound nats (Prod []) = 1"
    | "eval_arithFun_unbound nats (Prod (f # fs)) =  (eval_arithFun_unbound nats f * eval_arithFun_unbound nats (Prod fs))"
    | "eval_arithFun_unbound nats (Max [f]) = eval_arithFun_unbound nats f"
    | "eval_arithFun_unbound nats (Max (f # fs)) =  (max (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Max fs)))"
    | "eval_arithFun_unbound nats (Min [f]) = eval_arithFun_unbound nats f"
    | "eval_arithFun_unbound nats (Min (f # fs)) =  (min (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Min fs)))"

declare eval_arithFun_unbound.simps[simp del]

fun eval_arithFun :: "nat \<Rightarrow> arithFun \<Rightarrow> nat list \<Rightarrow> nat"
where "eval_arithFun c f nats = (eval_arithFun_unbound nats f) mod c"


datatype ('f,'l)sl_inter = SL_Inter nat "((('f,'l)lab  \<times> nat) \<times> arithFun)list"

instantiation sl_inter :: ("show", "show") "show" begin
  fun
    shows_sl_inter :: "('f::show, 'l::show) sl_inter \<Rightarrow> shows"
  where
    "shows_sl_inter(SL_Inter n fnas) = (
       ''carrier {0,...,'' +#+ shows n +@+ ''}'' +#+ shows_nl +@+
       shows_map (\<lambda>((f,n),a). ''['' +#+ shows f +@+ ''/'' +#+ shows n +@+ ''] = '' +#+ shows a +@+ shows_nl) fnas
     )"
  definition "shows_prec (d::nat) sli = shows_sl_inter sli"
  definition "shows_list_sl_inter (ss::('f::show,'l::show)sl_inter list) \<equiv> shows_list_aux shows ss"

  lemma assoc_sl_inter: "shows_prec d (sli::('f::show,'l::show)sl_inter) r @ s = shows_prec d sli (r @ s)"
  proof (cases sli)
    fix n::nat and fnas::"((('f::show,'l::show)lab\<times>nat)\<times>arithFun)list" assume assm: "sli = SL_Inter n fnas"
    let ?elt = "\<lambda>((f::('f::show,'l::show)lab,n::nat),a::arithFun). ''['' +#+ shows f +@+ ''/'' +#+ shows n +@+ ''] = '' +#+ shows a +@+ shows_nl"
    have "\<And>r s x. ?elt x r @ s = ?elt x (r@s)" unfolding show_defs o_def unfolding assoc[symmetric] by auto
    hence elt: "\<And>r s.\<forall>x\<in>set fnas. ?elt x r @ s = ?elt x (r@s)" by blast
    show ?thesis unfolding assm unfolding shows_prec_sl_inter_def shows_sl_inter.simps split_def show_defs o_def unfolding assoc[symmetric]
      unfolding shows_map_assoc[OF elt,of r s,unfolded show_defs o_def split_def,symmetric,unfolded assoc[symmetric]] by simp
  qed
  lemma assoc_sl_inter_elt: "\<forall>(sli::('f::show,'l::show)sl_inter)\<in>set slis. shows sli r @ s = shows sli (r@s)"
    using assoc_sl_inter by auto
  instance by
    (intro_classes,rule assoc_sl_inter,unfold shows_list_sl_inter_def)
    (rule shows_list_aux_assoc[OF assoc_sl_inter_elt])
end

definition unlab_of_trsI :: "(('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules"
 where "unlab_of_trsI R \<equiv> remdups (map_trsI unlab (filter (\<lambda> (l,r). l = r \<or> unlab_term l \<noteq> unlab_term r) R))"

lemma unlab_of_trsI_sound: "set (unlab_of_trsI R) = unlab_of_trs (set R)"
unfolding unlab_of_trsI_def unlab_of_trs_def decr_of_trs_def 
by auto

definition
  unlab_trs_tt :: "(('f::show,'l::show)lab,'v::show)rules \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> shows check"
where
  "unlab_trs_tt R rR \<equiv> let u =
  unlab_of_trsI R in (check_subseteq rR u \<guillemotright> check_subseteq u rR)
    <+? (\<lambda>s. shows_rules rR +@+ shows_nl +@+ '' does not correspond to unlabeled TRS ''
      +#+ shows_nl +@+ shows_rules u +@+ shows_nl)"

lemma unlab_trs_tt_sound: "sound_lab_tt_impl (\<lambda> R m rR rm. m = Suc rm \<and> isOK(unlab_trs_tt R rR))"
unfolding sound_lab_tt_impl_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume "isOK(unlab_trs_tt R rR)" and SN: "decrSN (set rR) rm" and m: "m = Suc rm"
  hence "set rR = set (unlab_of_trsI R)" unfolding unlab_trs_tt_def by (auto simp: Let_def)
  hence "set rR = unlab_of_trs (set R)" using unlab_of_trsI_sound by auto
  with SN have "decrSN(unlab_of_trs (set R)) rm" by auto
  with unlab_tt_sound m show "decrSN (set R) m" unfolding sound_lab_tt_def by auto
qed

fun get_largest_element where "get_largest_element (SL_Inter n _) = n"

fun sl_inter_to_inter :: "('f,'l)sl_inter \<Rightarrow> ('f,nat,'l)inter"
where "sl_inter_to_inter (SL_Inter c ls) fl cs = (case lookup (fl,length cs) ls of 
       None \<Rightarrow> 0
     | Some f \<Rightarrow> eval_arithFun (Suc c) f cs)"

lemma wf_sl_inter: 
  shows "wf_inter (sl_inter_to_inter sli) {x | x.  x < Suc (get_largest_element sli)}"
proof (cases sli)
  case (SL_Inter c ls)
  thus ?thesis unfolding wf_inter_def
  proof (intro allI, intro impI)
    fix f cs
    show "sl_inter_to_inter sli f cs \<in> {x | x. x < Suc (get_largest_element sli)}" using SL_Inter
      by (cases "lookup (f, length cs) ls", auto)
  qed
qed

interpretation sl_finite: sl_interpr "{x | x . x < Suc (get_largest_element sli)}" "get_largest_element sli" "sl_inter_to_inter sli" "{}" L for L
  by (unfold_locales, simp, rule wf_sl_inter, unfold decr_of_ord_def, auto)


(* now let us head for checking the model condition *)


fun enum_vectors :: "'c list \<Rightarrow> 'v list \<Rightarrow> ('v \<times> 'c)list list"
where "enum_vectors C [] = [[]]"
    | "enum_vectors C (x # xs) = (let rec = enum_vectors C xs in concat (map (\<lambda> vec. map (\<lambda> c. (x,c) # vec) C) rec))" 

fun assign_from_vector :: "('v \<times> 'c)list \<Rightarrow> 'v \<Rightarrow> 'c"
where "assign_from_vector vec x = the (lookup x vec)"

lemma enum_vectors_complete: assumes "C \<noteq> []"
  shows "\<exists> vec \<in> set (enum_vectors C xs). \<forall> x \<in> set xs. \<forall> c \<in> set C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c"
proof (induct xs, simp)
  case (Cons x xs)
  from this obtain vec where enum: "vec \<in> set (enum_vectors C xs)" and eq: "\<forall> x \<in> set xs. \<forall> c \<in> set C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c" by auto
  from enum have res: "set (map (\<lambda> c. (x,c) # vec) C) \<subseteq> set (enum_vectors C (x # xs)) " (is "_ \<subseteq> ?res") by auto
  show ?case 
  proof (cases "\<alpha> x \<in> set C")
    case False
    from assms have "hd C \<in> set C" by auto
    with res have elem: "(x,hd C) # vec \<in> ?res" (is "?vec \<in> _") by auto
    from eq False have "\<forall> y \<in> set (x # xs). \<forall> c \<in> set C. \<alpha> y = c \<longrightarrow> assign_from_vector ?vec y = c" by auto
    with elem show ?thesis by blast
  next
    case True
    with res have elem: "(x, \<alpha> x) # vec \<in> ?res" (is "?vec \<in> _") by auto
    from eq True have "\<forall> y \<in> set (x # xs). \<forall> c \<in> set C. \<alpha> y = c \<longrightarrow> assign_from_vector ?vec y = c" by auto
    with elem show ?thesis by blast
  qed
qed

declare assign_from_vector.simps[simp del]    

lemma eval_lab_independent:
  fixes t::"(('f,'l)lab,'v)term" and I::"('f,'c,'l)inter"
  assumes "\<forall>x\<in>vars_term t. \<alpha> x = \<beta> x"
  shows "eval_lab I L \<alpha> t = eval_lab I L \<beta> t"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence map: "map (eval_lab I L \<alpha>) ts = map (eval_lab I L \<beta>) ts" by auto
  thus ?case by (auto simp: Let_def map)
qed

fun
  check_sl_rule_ass ::
    "bool \<Rightarrow> ('f::show,'c::show,'l::show)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> (('f,'l)lab,'v)rules
     \<Rightarrow> ('v::show,'c)assign \<Rightarrow> (('f,'l)lab,'v)rule \<Rightarrow> shows check"
where
  "check_sl_rule_ass mc I L lR \<alpha> (l, r) = do {
     let cl_ll = eval_lab I L \<alpha> l;
     let cr_lr = eval_lab I L \<alpha> r;
     check (mc \<longrightarrow> (fst cl_ll = fst cr_lr))
       (''rule '' +#+ shows_rule (l, r) +@+ '' violates the model condition, [lhs] = ''
         +#+ shows (fst cl_ll) +@+ '', [rhs] = '' +#+ shows (fst cr_lr));
     check ((snd cl_ll, snd cr_lr) mem lR)
       (''labeled rule '' +#+ shows_rule (snd cl_ll, snd cr_lr) +@+ shows_string '' missing'')
   }"

fun
  check_sl_rule ::
    "('f::show,'c::show,'l::show)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 'c list \<Rightarrow> bool
     \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v::show)rule \<Rightarrow> shows check"
where
  "check_sl_rule I L C mc lR (l,r) = check_allm (\<lambda>\<alpha>. check_sl_rule_ass mc I L lR \<alpha> (l,r)) 
     (map assign_from_vector (enum_vectors C (list_union (vars_term_list l) (vars_term_list r))))"

lemma check_sl_rule_sound[simp]:
  fixes lR :: "(('f::show,'l::show)lab,'v::show)rules" and I :: "('f::show,'c::show,'l::show)inter"
  assumes "isOK (check_sl_rule I L C mc lR (l,r))"
  shows "(mc \<longrightarrow> model_rule I L (set C) l r) \<and> lab_rule I L (set C) (l,r) \<subseteq> set lR"
proof (cases C)
  case Nil
  thus ?thesis unfolding lab_rule_def wf_assign_def by auto
next
  case (Cons c D)
  let ?C = "Cons c D"
  let ?V = "list_union (vars_term_list l) (vars_term_list r) :: 'v list"
  from assms Cons have check: "\<forall> \<alpha> \<in> set (map assign_from_vector (enum_vectors ?C ?V)). isOK(check_sl_rule_ass mc I L lR \<alpha> (l,r))"
    by auto
  have "\<forall> \<alpha>. wf_assign (set ?C) \<alpha> \<longrightarrow> isOK(check_sl_rule_ass mc I L lR \<alpha> (l,r))"
  proof (intro allI, intro impI)
    fix \<alpha> :: "'v \<Rightarrow> 'c"
    assume wf_ass: "wf_assign (set ?C) \<alpha>"
    from enum_vectors_complete obtain vec where vec: "vec \<in> set (enum_vectors ?C ?V) \<and> (\<forall> x \<in> set ?V. \<forall> c \<in> set ?C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c)"
      by best
    let ?\<beta> = "assign_from_vector vec"
    from wf_ass vec have "\<forall> x \<in> set ?V. \<alpha> x = ?\<beta> x" unfolding wf_assign_def by auto
    hence "(\<forall> x \<in> vars_term l. ?\<beta> x = \<alpha> x) \<and> (\<forall> x \<in> vars_term r. ?\<beta> x = \<alpha> x) " by auto
    hence equal: "eval_lab I L \<alpha> l = eval_lab I L ?\<beta> l \<and> eval_lab I L \<alpha> r = eval_lab I L ?\<beta> r" 
      using eval_lab_independent by force
    from vec check have "isOK(check_sl_rule_ass mc I L lR ?\<beta> (l,r))" by auto
    with equal show "isOK(check_sl_rule_ass mc I L lR \<alpha> (l,r))" by auto
  qed
  thus ?thesis using Cons unfolding lab_rule_def by (auto simp: Let_def mem_iff)
qed

declare check_sl_rule.simps[simp del]

types ('f,'c,'l,'v)sl_check1 =
  "('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 'c list \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules
   \<Rightarrow> shows check"

types ('f,'l,'v)sl_check2 =
  "(('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> shows check"

definition
  check_sl_model_lab_trs :: "('f::show,'c::show,'l::show,'v::show)sl_check1" 
where
  "check_sl_model_lab_trs I L C lR R \<equiv> check_allm (check_sl_rule I L C True lR) R"

definition 
  check_sl_lab_trs :: "('f::show,'c::show,'l::show,'v::show)sl_check1" 
where
  "check_sl_lab_trs I L C lP P \<equiv> check_allm (check_sl_rule I L C False lP) P"

definition
  check_sl_only_lab_and_decr
where
  "check_sl_only_lab_and_decr gr R \<equiv>
     check_allm (\<lambda>(ll,lr).
       let l = unlab_term ll; r = unlab_term lr in
         check
           (if l = r
             then (\<not> is_var ll \<and> \<not> is_var lr \<and> gr (get_lab (the (root ll)),
               get_lab (the (root lr))) \<or> (ll = lr \<and> (l,r) mem R))
             else (l,r) mem R)
           (''labeled rule '' +#+ shows_rule(l,r) +@+
             shows_string '' is neither decreasing nor a labeled version of original rule''))"

lemma split_pairs_all: "(\<forall> x \<in> r :: 'a ars. P x) = (\<forall> a b . (a,b) \<in> r \<longrightarrow> P (a,b))"
by (auto)

lemma check_sl_only_lab_and_decr_sound[simp]:
  "isOK(check_sl_only_lab_and_decr gr R lR) = R_subset_lab_decr gr (set R) (set lR)"
proof - 
  {
    fix a b
    have "((unlab_term a = unlab_term b \<longrightarrow>
            \<not> is_var a \<and> \<not> is_var b \<and> (get_lab (the (root a)), get_lab (the (root b))) \<in> gr \<or> a = b \<and> (unlab_term b, unlab_term b) \<in> set R) \<and>
           (unlab_term a \<noteq> unlab_term b \<longrightarrow> (unlab_term a, unlab_term b) \<in> set R))
            =
           ((unlab_term a, unlab_term b) \<in> set R \<and> (a = b \<or> unlab_term a \<noteq> unlab_term b) \<or>
           unlab_term a = unlab_term b \<and> \<not> is_var a \<and> \<not> is_var b \<and> (get_lab (the (root a)), get_lab (the (root b))) \<in> gr)"
      by (auto)
  }
  thus ?thesis unfolding check_sl_only_lab_and_decr_def R_subset_lab_decr_def decr_of_ord_def
    by (simp add: Let_def, simp only: split_pairs_all, simp add: mem_iff mem_def[symmetric, of gr], blast)
qed

lemma check_sl_lab_trs_sound[simp]:
  assumes "isOK(check_sl_lab_trs I L C lP P)"
  shows "lab_trs I L (set C) (set P) \<subseteq> set lP"
proof -
  from assms have check: "\<forall> (l,r) \<in> set P. isOK(check_sl_rule I L C False lP (l,r))"
    unfolding check_sl_lab_trs_def by auto
  have "\<forall> (l,r) \<in> set P. lab_rule I L (set C) (l,r) \<subseteq> set lP" 
  proof (clarify)
    fix l r ll lr
    assume lr: "(l,r) \<in> set P" and llr: "(ll,lr) \<in> lab_rule I L (set C) (l,r)"
    from lr check have "isOK(check_sl_rule I L C False lP (l,r))" by auto 
    hence "lab_rule I L (set C) (l,r) \<subseteq> set lP" by auto
    with lr llr show "(ll,lr) \<in> set lP" by auto
  qed
  thus ?thesis by auto
qed

lemma check_sl_model_lab_trs_sound[simp]:
  assumes "isOK(check_sl_model_lab_trs I L C lR R)"
  shows "lab_trs I L (set C) (set R) \<subseteq> set lR \<and> model I L (set C) (set R)"
proof -
  from assms have check: "\<forall> (l,r) \<in> set R. isOK(check_sl_rule I L C True lR (l,r))"
    unfolding check_sl_model_lab_trs_def by auto
  have "\<forall> (l,r) \<in> set R. model_rule I L (set C) l r \<and> lab_rule I L (set C) (l,r) \<subseteq> set lR" 
  proof (clarify)
    fix l r
    assume lr: "(l,r) \<in> set R" 
    from lr check have "isOK(check_sl_rule I L C True lR (l,r))" by auto 
    thus "model_rule I L (set C) l r \<and> lab_rule I L (set C) (l,r) \<subseteq> set lR" by auto
  qed
  thus ?thesis unfolding model_def by auto
qed

definition
  check_unlab_proc ::
    "(('f::show,'l::show)lab,'v::show)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check"
where
  "check_unlab_proc P R ulP ulR \<equiv> do {
     let ULR = unlab_of_trsI R;
     check_subseteq (map_trsI unlab P) ulP;
     check_subseteq ULR ulR;
     check_subseteq ulR ULR
   } <+? (\<lambda>rl. ''problem with rule '' +#+ shows_rule rl +@+ shows_nl)"

lemma check_unlab_proc_sound: 
  "sound_lab_proc_impl (\<lambda> P R m rmap ulP ulR rm. m = Suc rm \<and> isOK(check_unlab_proc P R ulP ulR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m ulP ulR rm
  assume m: "m = Suc rm" 
    and  check: "isOK(check_unlab_proc P R ulP ulR)"
    and  finite: "finite_lab_dpp (set ulP, set ulR) rm"
  hence id: "unlab_of_trs (set R) = set ulR" and subs: "unlab_trs (set P) \<subseteq> set ulP" 
    unfolding check_unlab_proc_def by (auto simp: Let_def unlab_of_trsI_sound)
  {
    fix s t \<sigma>
    assume "min_lab_ichain (unlab_trs (set P), set ulR) rm s t \<sigma>"
    with subs have "min_lab_ichain (set ulP, set ulR) rm s t \<sigma>" by auto
  }
  hence fin: "finite_lab_dpp (unlab_trs (set P), set ulR) rm" using finite m unfolding finite_lab_dpp_def by blast
  show "finite_lab_dpp (set P, set R) m"
    by (simp only: m, rule unlab_proc_sound, auto simp only: id, rule fin)
qed


definition
  check_sem_lab_proc ::
    "('f::show,'l::show,'v::show)sl_check2
     \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> ('l + ('f, 'l) lab list) ars \<Rightarrow> ('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check"
where
  "check_sem_lab_proc check_lab check_model_lab gr I L P R lP lR \<equiv> do {
     check_lab lP P
       <+? (\<lambda>s. ''not all labeled DPs present:'' +#+ shows_nl +@+ s);
     check_model_lab lR R
       <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s);
     check_sl_only_lab_and_decr gr R lR 
       <+? (\<lambda>s. s +@+ shows_nl)
   }"

definition
  check_sem_lab_tt ::
    "('f::show,'l::show,'v::show)sl_check2
     \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> ('l + ('f, 'l) lab list) ars
     \<Rightarrow> ('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check"
where
  "check_sem_lab_tt check_lab check_model_lab gr I L R lR \<equiv> do {
     check_model_lab lR R;
     check_sl_only_lab_and_decr gr R lR
   } <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s)"


definition
  check_sem_lab_rel_tt ::
    "('f::show,'l::show,'v::show)sl_check2
     \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> ('l + ('f, 'l) lab list) ars
     \<Rightarrow> ('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 
     (('f,'l)lab,'v)reltrsL \<Rightarrow> 
     (('f,'l)lab,'v)reltrsL \<Rightarrow> 
     shows check"
where
  "check_sem_lab_rel_tt check_lab check_model_lab gr I L RS lRS \<equiv> do {
     check_model_lab (fst lRS) (fst RS);
     check_model_lab (snd lRS) (snd RS)
   } <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s)"

locale sl_interpr_impl = sl_interpr +
  fixes check_lab :: "('b,'c,'v::show)sl_check2"
   and  check_model_lab :: "('b,'c,'v::show)sl_check2"
 assumes check_lab: "isOK(check_lab lR R) \<Longrightarrow> lab_trs I L C (set R) \<subseteq> set lR"
   and   check_model_lab: "isOK(check_model_lab lR R) \<Longrightarrow> lab_trs I L C (set R) \<subseteq> set lR \<and> model I L C (set R)"
begin 

lemma check_sem_lab_tt_sound: 
  "sound_lab_tt_impl (\<lambda> R m lR rm. rm = Suc m \<and> isOK(check_sem_lab_tt check_lab check_model_lab gr I L R lR))"
unfolding sound_lab_tt_impl_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume "isOK(check_sem_lab_tt check_lab check_model_lab gr I L R rR)" and SN: "decrSN (set rR) rm" and m: "rm = Suc m"
  hence cond: "R_subset_lab_decr gr (set R) (set rR) \<and> model I L C (set R) \<and> Lab_trs (set R) \<subseteq> set rR" unfolding check_sem_lab_tt_def 
    using check_model_lab by (auto simp: check_model_lab)
  with lab_sound_tt SN show "decrSN (set R) m" unfolding sound_lab_tt_def by (auto simp: m)
qed

lemma check_sem_lab_rel_tt_sound: 
  "sound_rel_tt_impl (\<lambda> RS lRS. isOK(check_sem_lab_rel_tt check_lab check_model_lab gr I L RS lRS))"
unfolding sound_rel_tt_impl_def
proof (intro allI impI)
  fix R S rR rS
  assume "isOK(check_sem_lab_rel_tt check_lab check_model_lab gr I L (R,S) (rR,rS))" and SN: "rel_SN_alt (set rR, set rS)"
  hence mR: "model I L C (set R)" and R: "Lab_trs (set R) \<subseteq> set rR" and
        mS: "model I L C (set S)" and S: "Lab_trs (set S) \<subseteq> set rS" unfolding check_sem_lab_rel_tt_def 
    using check_model_lab by (auto simp: check_model_lab)
  from mR mS have "model I L C (set R \<union> set S)" unfolding model_def by auto
  from rel_SN_lab_imp_rel_SN[OF rel_SN_alt_mono[OF R S SN] this]
  show "rel_SN_alt (set R, set S)" .
qed

lemma check_sem_lab_proc_sound: 
  "sound_lab_proc_impl (\<lambda> P R m rmap lP lR rm. rm = Suc m \<and> isOK(check_sem_lab_proc check_lab check_model_lab gr I L P R lP lR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m lP lR rm

  assume check: "isOK(check_sem_lab_proc check_lab check_model_lab gr I L P R lP lR)"
    and  finite: "finite_lab_dpp (set lP, set lR) rm"
    and  m: "rm = Suc m"
  hence "Lab_trs (set P) \<subseteq> set lP \<and> Lab_trs (set R) \<subseteq> set lR \<and> model I L C (set R) \<and> R_subset_lab_decr gr (set R) (set lR)" using check_lab check_model_lab 
    unfolding check_sem_lab_proc_def by auto
  hence "sem_lab_proc (set P) (set R) m (set lP) (set lR) rm" unfolding sem_lab_proc_def by (auto simp: m)
  with finite sem_lab_proc_sound show "finite_lab_dpp (set P, set R) m" unfolding sound_lab_proc_def by blast
qed
end

lemma sl_helper:  "{x | x. x < Suc (get_largest_element sli)} = set [0..<Suc (get_largest_element sli)]" by auto

interpretation sl_finite_impl: sl_interpr_impl "{x | x . x < Suc (get_largest_element sli)}" "get_largest_element sli" "sl_inter_to_inter sli" "{}" L  
  "check_sl_lab_trs (sl_inter_to_inter sli) L [0 ..< Suc (get_largest_element sli)]"
  "check_sl_model_lab_trs (sl_inter_to_inter sli) L [0 ..< Suc (get_largest_element sli)]" 
  by (unfold_locales, simp only: sl_helper, simp only: check_sl_lab_trs_sound,
     simp only: sl_helper, simp only: check_sl_model_lab_trs_sound, auto)

definition
  create_sem_lab_tt ::
    "('f::show,'l::show)sl_inter \<Rightarrow> ('f,nat,'l)label \<Rightarrow> 
     (('f,'l)lab,'v::show)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check" 
where
  "create_sem_lab_tt sli L r lr \<equiv> do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     check_sem_lab_tt    
       (check_sl_lab_trs I L C) 
       (check_sl_model_lab_trs I L C) 
       (\<lambda> x. False)
       I L r lr
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"


definition
  create_sem_lab_rel_tt ::
    "('f::show,'l::show)sl_inter \<Rightarrow> ('f,nat,'l)label \<Rightarrow> 
     (('f,'l)lab,'v::show)reltrsL \<Rightarrow> 
     (('f,'l)lab,'v)reltrsL \<Rightarrow> 
     shows check" 
where
  "create_sem_lab_rel_tt sli L rs lrs \<equiv> do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     check_sem_lab_rel_tt    
       (check_sl_lab_trs I L C) 
       (check_sl_model_lab_trs I L C) 
       (\<lambda> x. False)
       I L rs lrs
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"

definition
  create_sem_lab_proc ::
    "('f::show,'l::show)sl_inter \<Rightarrow> ('f,nat,'l)label \<Rightarrow> 
     (('f,'l)lab,'v::show)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check" 
where
  "create_sem_lab_proc sli L r1 r2 r3 r4 \<equiv> do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     check_sem_lab_proc 
       (check_sl_lab_trs I L C) 
       (check_sl_model_lab_trs I L C) 
       (\<lambda>x. False)
       I L r1 r2 r3 r4
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"

lemma create_sem_lab_proc_sound:
  "sound_lab_proc_impl (\<lambda> P R m rmap lP lR rm. rm = Suc m \<and> isOK(create_sem_lab_proc sli L P R lP lR))"
proof -
  have id: "(\<lambda> x. False) = {}" by (rule ext, auto)
  show ?thesis
    using sl_finite_impl.check_sem_lab_proc_sound 
    by (simp add: Let_def create_sem_lab_proc_def id, blast)
qed

lemma create_sem_lab_tt_sound:
  "sound_lab_tt_impl (\<lambda> R m lR rm. rm = Suc m \<and> isOK(create_sem_lab_tt sli L R lR))"
proof -
  have id: "(\<lambda> x. False) = {}" by (rule ext, auto)
  show ?thesis using sl_finite_impl.check_sem_lab_tt_sound
    unfolding sound_lab_tt_impl_def
    by (simp add: Let_def create_sem_lab_tt_def id, blast)
qed

lemma create_sem_lab_rel_tt_sound:
  "sound_rel_tt_impl (\<lambda> RS lRS. isOK(create_sem_lab_rel_tt sli L RS lRS))"
proof -
  have id: "(\<lambda> x. False) = {}" by (rule ext, auto)
  show ?thesis using sl_finite_impl.check_sem_lab_rel_tt_sound
    unfolding sound_rel_tt_impl_def
    by (simp add: Let_def create_sem_lab_rel_tt_def id, blast)
qed
  
end
