(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory UsableRules
imports RedPair Substitution Edg
begin

fun
  rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "rewrite R s = concat (map (\<lambda> (l, r) . concat (map (\<lambda> p.
     (case match (s |_ p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>])
   ) (poss_list s)) ) R)"


lemma rewrite_complete:
  fixes l r::"('f,'v)term"
  assumes "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" and "vars_term l \<supseteq> vars_term r"
  shows "t \<in> set (rewrite R s)" 
proof -
  let ?C = "ctxt_of_pos_term p s"
  from assms have lr: "(l,r) \<in> set R" 
    and p: "p \<in> poss s" 
    and s: "?C\<langle>l \<cdot> \<sigma>\<rangle> = s"
    and t: "?C\<langle>r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from s p have "?C\<langle>l \<cdot> \<sigma>\<rangle> = ?C\<langle>s|_p\<rangle>" by (simp add: ctxt_supt_id)
  hence "l \<cdot> \<sigma> = s|_p" by auto
  from this match_complete obtain \<delta> where 
    match: "match (s|_p) l = Some \<delta>" and 
           "\<forall> x \<in> vars_term l. get_subst \<sigma> x = get_subst \<delta> x" by blast
  with assms lr have vars: "\<forall> x \<in> vars_term r. get_subst \<sigma> x = get_subst \<delta> x" by blast
  have "r \<cdot> \<sigma> = r \<cdot> (\<sigma>|_(vars_term r))" by (rule coincidence_lemma)
  also have "\<dots> = r \<cdot> (\<delta>|_(vars_term r))" 
  proof -
    have "\<sigma>|_(vars_term r) = \<delta>|_(vars_term r)" 
    proof (cases \<sigma>, cases \<delta>)
      fix sigma delta
      assume s: "\<sigma> = Subst sigma" and d: "\<delta> = Subst delta"
      with vars have vars2: "\<forall> x \<in> vars_term r. sigma x = delta x" by auto
      let ?P = "\<lambda> y. (\<lambda> x. if x \<in> vars_term r then sigma x else Var x) y = 
	(\<lambda> x. if x \<in> vars_term r then delta x else Var x) y"
      have "\<And> y. ?P y"	
      proof -
	fix y
	from vars2 show "?P y" by (cases "y \<in> vars_term r", auto)
      qed      
      with s d show ?thesis by auto  
    qed
    thus ?thesis by auto
  qed
  also have "\<dots> = r \<cdot> \<delta>" by (rule coincidence_lemma[symmetric])
  finally  have nearlyDone: "(l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ ?C\<langle>r\<cdot>\<sigma>\<rangle>])" using lr p match t by auto 
  hence "t \<in> set ((\<lambda>(l, r). concat  (map 
     (\<lambda>p. case match (s|_p) l  of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>])
                          (poss_list s)))  (l,r))" (is "t \<in> (?P (l,r))")
    by auto
  with lr have "\<exists> lr \<in> set R. t \<in> (?P lr)" by blast
  thus ?thesis by auto
qed

lemma rewrite_sound: assumes "t \<in> set (rewrite R s)"
  shows "(s,t) \<in> rstep (set R)"
proof -
  from assms 
  have "\<exists> l r p.  (l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])" 
    by (auto simp: mem_iff, blast)
  from this obtain l r p where one:
    "(l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])"
    by auto
  from this obtain \<sigma> where two: "match (s|_p) l = Some \<sigma> \<and> t \<in> {(ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>}"
    by (cases "match (s|_p) l", auto)
  hence match: "l \<cdot> \<sigma> = s|_p" using match_sound by auto 
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  thus "(s,t) \<in> rstep (set R)" unfolding rstep_iff_rstep_r_p_s by blast
qed



fun i_trans_check :: "('f,'v)rules \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "i_trans_check R ur us (Fun f ts) = ((f,length ts) \<in> us \<and> \<not> (\<exists> (l,r) \<in> set R - ur. matchCapBelow l R (Fun f ts)))"

declare i_trans_check.simps[simp del]

function i_trans :: "(('f,'v)rules \<times> ('f,'v)trs \<times> ('f \<times> nat)set \<times> ('f \<times> 'f \<times> nat)) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "i_trans _ (Var x) = (Var x)" 
   |  "i_trans (R, ur, us, cn) (Fun f ts) = (if SN_elt (rstep (set R)) (Fun f ts) then (
         if i_trans_check R ur us (Fun f ts) then Fun f (map (i_trans (R, ur, us, cn)) ts) 
                   else (comb cn) (Fun f (map (i_trans (R, ur, us, cn)) ts) # map (i_trans (R, ur, us, cn)) (rewrite R (Fun f ts)))
  )  else Fun f ts)"
by (pat_completeness, auto)


termination i_trans
proof -
  {
    fix R :: "('f,'v)rules" 
    and ur :: "('f,'v)trs" 
    and us :: "('f \<times> nat)set" 
    and cn :: "'f \<times> 'f \<times> nat"
    have "\<forall> t. i_trans_dom ((R,ur,us,cn),t)"
    proof (clarify)
      fix t
      let ?Ruc = "(R,ur,us,cn)"
      have "SN (SN_rel_supt (rstep (set R)))" by (rule SN_SN_rel_supt_rstep)
      hence "wf ((SN_rel_supt (rstep (set R)))\<inverse>)" by (simp add: SN_iff_wf)
      thus "i_trans_dom (?Ruc, t)"
      proof (induct)
	fix t    
	assume ind: "\<And>s. (s, t) \<in> (SN_rel_supt (rstep (set R)))\<inverse> \<Longrightarrow> i_trans_dom (?Ruc, s)"
	show "i_trans_dom (?Ruc, t)"
	proof 
	  fix s
	  assume step: "i_trans_rel s (?Ruc, t)"
	  thus "i_trans_dom s"
	  proof
	    fix R' :: "('f,'v)rules" and ur' :: "('f,'v)trs" and us' :: "('f \<times> nat)set" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(R',ur',us',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set ts"
	    assume SN: "SN_elt (rstep (set R')) (Fun f ts)"
	    from t' t SN have "t \<rhd> t'" and "SN_elt (rstep (set R)) t" by auto
	    hence "(t',t) \<in> (SN_rel_supt (rstep (set R)))\<inverse>" unfolding SN_rel_supt_def supt_def SN_rel_def by auto
	    with ind s t show "i_trans_dom s" by auto
	  next
	    fix R' :: "('f,'v)rules" and ur' :: "('f,'v)trs" and us' :: "('f \<times> nat)set" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(R',ur',us',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set ts"
	    assume SN: "SN_elt (rstep (set R')) (Fun f ts)"
	    from t' t SN have "t \<rhd> t'" and "SN_elt (rstep (set R)) t" by auto
	    hence "(t',t) \<in> (SN_rel_supt (rstep (set R)))\<inverse>" unfolding SN_rel_supt_def supt_def SN_rel_def by auto
	    with ind s t show "i_trans_dom s" by auto
	  next
	    fix R' :: "('f,'v)rules" and ur' :: "('f,'v)trs" and us' :: "('f \<times> nat)set" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(R',ur',us',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set (rewrite R' (Fun f ts))"
	    assume SN: "SN_elt (rstep (set R')) (Fun f ts)"
	    from t t' rewrite_sound have step: "(t, t') \<in> rstep (set R)" by blast
	    from SN t have "SN_elt (rstep (set R)) t" by auto
	    with step have "(t',t) \<in> (SN_rel_supt (rstep (set R)))\<inverse>" unfolding SN_rel_supt_def
	      SN_rel_def by auto
	    with ind s t show "i_trans_dom s" by auto
	  qed
	qed
      qed
    qed
  }
  thus ?thesis by force
qed


fun i_trans_subst :: "(('f,'v)rules \<times> ('f,'v)trs \<times> ('f \<times> nat)set \<times> ('f \<times> 'f \<times> nat)) \<Rightarrow> ('f,'v)subst \<Rightarrow> ('f,'v)subst"
  where "i_trans_subst params (Subst \<sigma>) = (Subst (\<lambda> x. i_trans params (\<sigma> x)))" 

fun ur_closed_term_af :: "('f,'v)term \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_term_af (Var x) _ _ _ _ = True"
   |  "ur_closed_term_af (Fun f ts) R ur us \<pi> = ((f,length ts) \<in> us \<and> (\<forall> i < length ts. \<pi> f (length ts) i \<longrightarrow> ur_closed_term_af (ts ! i) R ur us \<pi>) \<and> 
  (\<forall> (l,r) \<in> set R. matchCapBelow l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur) 
)" 

abbreviation ur_closed_af :: "('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)rules \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_af ur us R \<pi> \<equiv> 
   (\<forall> (l,r) \<in> ur. ur_closed_term_af r R ur us \<pi>)"

notation Edg.equiv_class ("[_]")

context af_redpair
begin

lemma i_transI: assumes ur: "ur_closed_term_af t R ur us \<pi>" and sn: "SN_elt (rstep (set R)) (t \<cdot> \<sigma>)"
  shows "(t \<cdot> (i_trans_subst (R, ur, us, com) \<sigma>), i_trans (R, ur, us, com) (t \<cdot> \<sigma>)) \<in> NS^*"
using assms
proof (induct t, cases \<sigma>, simp)
  case (Fun f ts)  
  hence sClosed: "\<And> i. i < length ts \<and> \<pi> f (length ts) i \<Longrightarrow> ur_closed_term_af (ts ! i) R ur us \<pi>" by auto
  let ?Ruc = "(R, ur, us, com)"
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_elt (rstep (set R)) (s \<cdot> \<sigma>)" 
    by (simp add: SN_imp_SN_arg)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have snt: "SN_elt (rstep (set R)) (Fun f ts)" by (rule SNinstance_imp_SN)
  have "i_trans ?Ruc (Fun f ts \<cdot> \<sigma>)
       = i_trans ?Ruc (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" (is "?l = _") by auto
  also have "\<dots> = Fun f (map (i_trans ?Ruc) (map (\<lambda> t. t \<cdot> \<sigma>) ts))"
  proof -
    from `ur_closed_term_af (Fun f ts) R ur us \<pi>`
    have inst: "\<forall> (l,r) \<in> set R. matchCapBelow l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur" and us: "(f, length ts) \<in> us" by auto
    have "i_trans_check R ur us (Fun f ts \<cdot> \<sigma>)" (is ?check)
    proof (rule ccontr)
      assume "\<not> ?check"
      from this us obtain l r \<delta> where nUsable: "(l,r) \<in> set R - ur" and elem: "l \<cdot> \<delta> \<in> [HFun f (map (\<lambda> t. cap (set R) (t \<cdot> \<sigma>)) ts)]" 
	(is "_ \<in> [?hs]")
	by (auto simp:i_trans_check.simps gcMatch_def)
      have "\<And> t. [cap (set R) (t \<cdot> \<sigma>)] \<subseteq> [cap (set R) t]" by (rule capInstanceSubset)
      hence "[?hs] \<subseteq> [HFun f (map (cap (set R)) ts)]" (is "_ \<subseteq> [?h]") by auto
      with elem have "l \<cdot> \<delta> \<in> [?h]" by auto
      hence "gcMatch ?h l" unfolding gcMatch_def ..
      with inst nUsable show False by auto
    qed
    with Fun show ?thesis by auto
  qed
  finally have eq1: "?l = Fun f (map (\<lambda> t. i_trans ?Ruc (t \<cdot> \<sigma>)) ts)" (is "_ = ?r") by simp
  let ?tis = "\<lambda> t. t \<cdot> (i_trans_subst ?Ruc \<sigma>)"
  let ?its = "\<lambda> t. i_trans ?Ruc (t \<cdot> \<sigma>)"
  have "(Fun f (map ?tis ts), Fun f (map ?its ts)) \<in> NS^*" 
  proof -
    have ind: "\<forall> i < length ts. \<pi> f (length ts) i \<longrightarrow> (?tis (ts ! i), ?its (ts ! i)) \<in> NS^*"
    proof (rule allI, (rule impI)+)
      fix i
      assume ilen: "i < length ts" and ipi: "\<pi> f (length ts) i"
      with sns have "ts ! i \<in> set ts" and "SN_elt (rstep (set R)) (ts ! i \<cdot> \<sigma>)" by auto
      with Fun sClosed ilen ipi show "(?tis (ts ! i), ?its (ts ! i)) \<in> NS^*" by auto
    qed
    from wmNS have cc: "one_step_ctxt_closed (NS^*)" using rtrancl_one_step_ctxt_closed  by auto
    show ?thesis 
    proof (rule af_steps_imp_orient, rule trans_rtrancl, rule refl_rtrancl, rule cc, simp)
      show "\<forall>i<length (map ?tis ts). \<pi> f (length ts) i \<longrightarrow> (map ?tis ts ! i, map ?its ts ! i) \<in> \<succeq>\<^sup>*" 
	using ind by auto
    next
      show "\<forall>bef s t aft. length (map ?tis ts) = Suc (length bef + length aft) \<longrightarrow> \<pi> f (length ts) (length bef) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> \<succeq>\<^sup>*"
	using af_compat unfolding af_compatible_def by force
    qed
  qed
  also have "Fun f (map ?tis ts) = Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>" by auto
  finally show ?case using eq1  by auto
qed
end

lemma i_transII:
  assumes sn: "SN_elt (rstep (set R)) (t \<cdot> \<sigma>)" 
  shows "(i_trans (R, ur, us, cn) (t \<cdot> \<sigma>), t \<cdot> i_trans_subst (R, ur, us, cn) \<sigma>) \<in> (rstep (ce_trs cn))^*
  \<and> (t \<notin> wf_terms us \<longrightarrow> (i_trans (R, ur, us, cn) (t \<cdot> \<sigma>), t \<cdot> i_trans_subst (R, ur, us, cn) \<sigma>) \<in> (rstep (ce_trs cn))^+) " 
  (is "(i_trans ?Ruc _,_) \<in> ?rel \<and> (_ \<longrightarrow> _ \<in> ?Rel)  ")
using sn
proof (induct t)
  case (Var x) thus ?case using wf_terms.Var by (cases \<sigma>) auto
next
  case (Fun f ts)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_elt (rstep (set R)) (s \<cdot> \<sigma>)"
    by (simp add: SN_imp_SN_arg)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have snt: "SN_elt (rstep (set R)) (Fun f ts)" by (rule SNinstance_imp_SN)
  let ?ts = "map (\<lambda> t. i_trans ?Ruc (t \<cdot> \<sigma>)) ts"
  let ?ss = "map (\<lambda> t. t \<cdot> i_trans_subst ?Ruc \<sigma>) ts"
  have len: "length ?ts = length ?ss" and other: "\<forall> i. i< length ?ts \<longrightarrow> (?ts ! i, ?ss ! i) \<in> ?rel" using sns Fun  by auto
  from len other have "(Fun f ?ts, Fun f ?ss) \<in> ?rel" using rstep_rtrancl_all_ctxt_closed[where r = "ce_trs cn"] unfolding all_ctxt_closed_def by blast 
  hence nearlyDone: "(Fun f ?ts, Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?rel" (is "(_,?ti) \<in> _") by auto
  show ?case
  proof (cases "i_trans_check R ur us (Fun f ts \<cdot> \<sigma>)")
    case False
    with Fun have id: "i_trans ?Ruc (Fun f ts \<cdot> \<sigma>) = 
       (comb cn) (Fun f ?ts # map (i_trans ?Ruc) (rewrite R (Fun f ts \<cdot> \<sigma>)))" (is "?it = (comb cn) ?list") 
      by (simp add: o_def) 
    have "Fun f ?ts \<in> set ?list" by simp
    hence "((comb cn) ?list, Fun f ?ts) \<in> (rstep (ce_trs cn))^+" by (rule ce_trs_sound) 
    with id nearlyDone show "?thesis" by auto
  next
    case True
    with Fun nearlyDone have part1: "(i_trans ?Ruc (Fun f ts \<cdot> \<sigma>), Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?rel"
      by (simp add: o_def) 
    show ?thesis 
    proof (rule conjI, rule part1, intro impI)
      assume not: "Fun f ts \<notin> wf_terms us"
      with Fun True have id: "i_trans ?Ruc (Fun f ts \<cdot> \<sigma>) = Fun f ?ts" by simp 
      from True have fus: "(f,length ts) \<in> us" by (simp add: i_trans_check.simps)
      from not obtain t where t: "t \<in> set ts" and tus: "t \<notin> wf_terms us" using wf_terms.Fun[OF fus] by blast
      from t obtain i where i: "i < length ?ts" and ti: "t = ts ! i" by (simp only: set_conv_nth, auto)
      from Fun(1)[OF t _] sns t ti tus have "(i_trans ?Ruc (ts ! i \<cdot> \<sigma>), ts ! i \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?Rel" by auto   
      with i have ind: "(?ts ! i, ?ss ! i) \<in> ?Rel" by auto
      have "(Fun f ?ts, Fun f ?ss) \<in> ?Rel" 
        by (rule rtrancl_trancl_into_trancl[OF len other i ind one_step_ctxt_closed_rstep])
      with id show "(i_trans ?Ruc (Fun f ts \<cdot> \<sigma>), Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?Rel" by auto
    qed
  qed
qed

context ce_redpair
begin

definition c :: "'a" where "c \<equiv> SOME c. True"
definition d :: "'a" where "d \<equiv> SOME d. True"
definition n :: nat where "n \<equiv> SOME n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c d.  ce_trs (c,d,m) \<subseteq> NS)"

lemma NS_ce_compat_n: "m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> NS"
proof -
  assume m: "n \<le> m"
  from mp[OF spec[OF someI_ex[OF NS_ce_compat]] m[unfolded n_def]] show "ce_trs (c,d,m) \<subseteq> NS" by blast
qed

lemma ce_orient: "m \<ge> n \<Longrightarrow> (rstep (ce_trs (c,d,m)))^* \<subseteq> NS^*"
  by (rule rtrancl_mono, rule trs_inclusion, rule wmNS, rule substNS, rule NS_ce_compat_n, simp)
end

context ce_af_redpair
begin  
lemma i_transIII: assumes 
  m: "m \<ge> n" and
  sn: "SN_elt (rstep (set R)) (Fun f ts)" and
  step: "(Fun f ts,s) \<in> rstep_r_p_s (set R) lr \<epsilon> \<sigma>" and
  ur: "i_trans_check R ur us (Fun f ts)" and
  ur_closed: "ur_closed_af ur us R \<pi>"
  shows "((i_trans (R,ur,us,(c,d,m)) (Fun f ts), i_trans (R,ur,us,(c,d,m)) s) \<in> NS^* O rstep ur O NS^* \<and> lr \<in> ur) \<and> 
    (fst lr \<notin> wf_terms us \<longrightarrow>  (i_trans (R,ur,us,(c,d,m)) (Fun f ts), i_trans (R,ur,us,(c,d,m)) s) \<in> (rstep (ce_trs (c,d,m)))^+ O rstep ur O NS^*)"
proof -
  let ?cn = "(c,d,m)"
  let ?Ruc = "(R,ur,us,?cn)"
  let ?I = "i_trans ?Ruc"
  let ?Is = "i_trans_subst ?Ruc \<sigma>"
  from step obtain l r where lr: "lr = (l,r)" and rule: "(l,r) \<in> set R" and left: "l \<cdot> \<sigma> = Fun f ts" and right: "r \<cdot> \<sigma> = s"
    unfolding rstep_r_p_s_def by force
  {
    assume "is_var l"
    with rule obtain x where "(Var x, r) \<in> set R" by (cases l, auto)
    hence "\<not> (SN_elt (rstep (set R)) (Fun f ts))" by (rule left_var_imp_not_SN) 
    hence False using sn by auto
  }
  then obtain g ss where l: "l = Fun g ss" by (cases l, auto)
  from ur have nmatch: "\<forall> (l,r) \<in> set R - ur. \<not> matchCapBelow l R (Fun f ts)" by (simp only: i_trans_check.simps, blast)
  from l left have fg: "f = g" and tsss: "ts = map (\<lambda> s. s \<cdot> \<sigma>) ss" (is "_ = ?ts") and nvar: "\<not> (is_var l)" by auto
  have "\<forall> i. ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
  proof 
    fix i
    show "ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
      by (rule capSound[where s = "ss ! i \<cdot> \<sigma>" and t = "ss ! i \<cdot> \<sigma>" and \<sigma> = \<iota>], auto) 
  qed
  hence "matchCapBelow (Fun f ss) R (Fun f (map (\<lambda> s. s \<cdot> \<sigma>) ss))"
    by (simp add: gcMatch_def, blast)
  with l fg tsss rule nmatch have urule: "(l,r) \<in> ur" by blast
  with left[symmetric] right[symmetric] subst_closureI[where \<sigma> = \<sigma>] ctxt_closureI[where C = Hole] 
  have urStep: "(Fun f ts, s) \<in> rstep ur" by force
  from urule ur_closed have fr: "ur_closed_term_af r R ur us \<pi>" by auto
  from left sn have snl: "SN_elt(rstep (set R)) (l \<cdot> \<sigma>)" by auto
  from i_transII[OF snl] left lr have oneB: "fst lr \<notin> wf_terms us \<longrightarrow> (?I (Fun f ts), l \<cdot> ?Is) \<in> (rstep (ce_trs ?cn))^+" by auto
  from i_transII[OF snl] ce_orient[OF m] have "(?I (l \<cdot> \<sigma>), l \<cdot> ?Is) \<in> NS^*" by blast
  with left have one: "(?I (Fun f ts), l \<cdot> ?Is) \<in> NS^*" by auto
  from step lr have "(Fun f ts, s) \<in> rstep (set R)" 
    unfolding rstep_iff_rstep_r_p_s by blast
  with right sn step_preserves_SN_elt[of _ _ "rstep (set R)"] have sns: "SN_elt (rstep (set R)) (r \<cdot> \<sigma>)" by auto
  have "(r \<cdot> ?Is , ?I (r \<cdot> \<sigma>)) \<in> NS^*"  by (rule i_transI[OF fr sns])
  with right have stepRight: "(r \<cdot> ?Is, ?I s) \<in> NS^*" by auto
  from urule have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep_r_p_s ur (l,r) \<epsilon> ?Is"
    unfolding rstep_r_p_s_def by auto
  hence "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep ur" unfolding rstep_iff_rstep_r_p_s by blast
  with stepRight have two: "(l \<cdot> ?Is, ?I s) \<in> rstep ur O NS^*" by auto
  from one two have part1: "(?I (Fun f ts), ?I s) \<in> NS^* O rstep ur O NS^*" by auto
  show ?thesis
  proof (rule conjI, rule conjI, rule part1, simp only: lr urule, intro impI)
    assume fus: "fst lr \<notin> wf_terms us"
    show "(?I (Fun f ts), ?I s) \<in> (rstep (ce_trs ?cn))^+ O rstep ur O NS^*" using mp[OF oneB fus] two by auto
  qed
qed
end

lemma i_transIV: assumes 
  sn: "SN_elt (rstep (set R)) (Fun f ts)" and
  step: "(Fun f ts,s) \<in> rstep (set R)" and
  ur: "\<not> i_trans_check R ur us (Fun f ts)" 
  shows "(i_trans (R,ur,us,cn) (Fun f ts), i_trans (R,ur,us,cn) s) \<in> (rstep (ce_trs cn))^+" (is "_ \<in> ?rel")
proof -
  let ?I = "i_trans (R,ur,us,cn)"
  from ur sn have id: "?I (Fun f ts) = (comb cn) (Fun f (map ?I ts) # map ?I (rewrite R (Fun f ts)))" (is "_ = (comb cn) ?list") 
    by auto
  from step[unfolded rstep_iff_rstep_r_p_s]
  obtain l r p \<sigma> where step2: "(Fun f ts,s) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" by blast
  with free_right_rewrite_imp_not_SN sn have "vars_term l \<supseteq> vars_term r" by blast
  with step2 have "s \<in> set (rewrite R (Fun f ts))" by (rule rewrite_complete)
  hence "(?I s) \<in> set ?list" by auto
  hence "(comb cn ?list, ?I s) \<in> ?rel" by (rule ce_trs_sound)
  with id show ?thesis by simp
qed

context ce_af_redpair
begin
lemma i_transV: 
  assumes 
  m: "m \<ge> n" and
  sn: "SN_elt (rstep (set R)) t" and
  step: "(t,s) \<in> rstep (set R)" and
  ur_closed: "ur_closed_af ur us R \<pi>" and
  orient: "ur \<subseteq> NS"
  shows "(i_trans (R,ur,us,(c,d,m)) t, i_trans (R,ur,us,(c,d,m)) s) \<in> NS^* \<and> ((t,s) \<in> rstep (ur - {lr | lr. fst lr \<notin> wf_terms us }) \<or>
         (i_trans (R,ur,us,(c,d,m)) t, i_trans (R,ur,us,(c,d,m)) s) \<in> (rstep (ce_trs (c,d,m)))^+ O NS^*)"
    (is "_ \<in> ?R \<and> (_ \<in> rstep ?ur \<or> _ \<in> ?Ce O NS^*)")
proof -
  let ?CeN = "?Ce O NS^*"
  let ?cn = "(c,d,m)"
  have cc1: "one_step_ctxt_closed ((rstep (ce_trs ?cn))^*)"
    using one_step_ctxt_closed_rstep[where r = "ce_trs ?cn"]
    and rtrancl_one_step_ctxt_closed by auto 
  from 
    rtrancl_one_step_ctxt_closed[OF wmNS]
    one_step_ctxt_closed_rstep[where r = ur]
    comp_one_step_ctxt_closed
  have ccR: "one_step_ctxt_closed ?R" by blast
  from trs_inclusion[OF wmNS substNS orient] have subset0: "rstep ur \<subseteq> NS" .
  hence subset1: "NS^* O rstep ur O NS^* \<subseteq> NS^*" by auto  
  from one_step_ctxt_closed_rstep[where r = "ce_trs ?cn"] cc1
  have ccC: "one_step_ctxt_closed ?Ce" 
    by (simp only: trancl_unfold_right, rule comp_one_step_ctxt_closed)
  have ccCN: "one_step_ctxt_closed ?CeN" 
    by (rule comp_one_step_ctxt_closed[OF ccR ccC])
  from ce_orient[OF m] have ce_or: "?Ce \<subseteq> NS^*" by auto
  let ?I = "i_trans (R,ur,us,?cn)"
  let ?goal = "\<lambda> t s. 
    (?I t, ?I s) \<in> ?R \<and> ((t,s) \<in> rstep ?ur \<or>  (?I t, ?I s) \<in> ?CeN)"
  from step[unfolded rstep_CC_CS] sn show "?goal t s"
  proof induct 
    fix t s C
    assume "(t,s) \<in> CS (set R)" and "SN_elt (rstep (set R)) C\<langle>t\<rangle>"
    thus "?goal C\<langle>t\<rangle> C\<langle>s\<rangle>"
    proof (induct C)
      case Hole
      thus ?case
      proof (induct)	
	fix l r \<sigma>
	assume lr: "(l,r) \<in> set R" and sn: "SN_elt (rstep (set R)) \<box>\<langle>l \<cdot> \<sigma>\<rangle>" (is "SN_elt _ ?ls")
	from lr have stepp: "(?ls, \<box>\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" (is "(_,?rs) \<in> _")
	  unfolding rstep_r_p_s_def by auto
	hence step: "(?ls, ?rs) \<in> rstep (set R)" unfolding rstep_iff_rstep_r_p_s by blast
	with sn have "\<not> (is_var ?ls)" using var_rewrite_imp_not_SN by blast
	from this obtain f ls where ls: "?ls = Fun f ls" by (cases ?ls, auto)
	with sn step stepp have snls: "SN_elt (rstep (set R)) (Fun f ls)" and 
	  stepls: "(Fun f ls, ?rs) \<in> rstep (set R)"  and
	  steppls: "(Fun f ls, ?rs) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" by auto 
	show "?goal ?ls ?rs"
	proof (cases "i_trans_check R ur us (Fun f ls)")
	  case True note outerTrue = this
          show ?thesis
          proof (cases "l \<in> wf_terms us")
            case True
            from i_transIII[OF m snls steppls outerTrue ur_closed]
	    have piece1: "(?I (Fun f ls), ?I ?rs) \<in> NS^*" and "(l,r) \<in> ur" using subset1 by blast+
            with True have lr: "(l,r) \<in> ?ur" by auto
            show ?thesis by (rule conjI, simp only: ls, rule piece1, rule disjI1, simp only: rstep_iff_rstep_r_c_s, 
              rule exI[of _ l], rule exI[of _ r], rule exI[of _ \<box>], intro exI,
              rule conjI, rule lr, unfold rstep_r_c_s_def, auto)
          next
            case False
            from subset0 have "(rstep (ce_trs ?cn))^+ O rstep ur O NS^* \<subseteq> ?CeN" by force
            with False i_transIII[OF m snls steppls outerTrue ur_closed, simplified fst_conv]
	    have piece1: "(?I (Fun f ls), ?I ?rs) \<in> ?CeN" by blast
            with ce_or have piece0: "(?I (Fun f ls), ?I ?rs) \<in> NS^*" by auto
            with ls piece1 show ?thesis by (auto)
          qed
	next
	  case False
	  from snls stepls this 
	  have steps: "(?I (Fun f ls), ?I ?rs) \<in> (rstep (ce_trs ?cn))^+" by (rule i_transIV)
          from steps ce_orient[OF m] have piece0: "(?I (Fun f ls), ?I ?rs) \<in> NS^*" by auto
          from steps have piece1: "(?I (Fun f ls), ?I ?rs) \<in> ?CeN" by auto
	  with ls piece0 show ?thesis by (auto)
	qed
      qed
    next
      case (More f bef D aft)
      hence snbta: "SN_elt (rstep (set R)) (Fun f (bef @ D\<langle>t\<rangle> # aft))" (is "SN_elt _ (Fun f ?bta)") by auto
      let ?Dt = "D\<langle>t\<rangle>"
      let ?Ds = "D\<langle>s\<rangle>"
      let ?bsa = "bef @ ?Ds # aft"
      let ?C = "More f bef D aft"
      from More ctxt_closureI[where C = ?C] have "(?C\<langle>t\<rangle>, ?C\<langle>s\<rangle>) \<in> rstep (set R)"
        unfolding rstep_CC_CS by auto
      hence step: "(Fun f ?bta, Fun f ?bsa) \<in> rstep (set R)" by auto
      show ?case 
      proof (cases "i_trans_check R ur us (Fun f ?bta)")
	case False
	from snbta step False 
	have steps: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> (rstep (ce_trs ?cn))^+" by (rule i_transIV)
        from steps ce_orient[OF m] have piece0: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> NS^*" by auto
        from steps have piece1: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> ?CeN" by auto
	  with piece0 show ?thesis by auto
      next
	case True
        hence fus: "(f, Suc (length bef + length aft)) \<in> us" by (simp add: i_trans_check.simps)
	have "?Dt \<in> set ?bta" by auto 
	with snbta have "SN_elt (rstep (set R)) ?Dt" by (rule SN_imp_SN_arg)
	with More have "?goal ?Dt ?Ds" by auto
        hence always: "(?I ?Dt, ?I ?Ds) \<in> ?R" and choice: "(?Dt, ?Ds) \<in> rstep ?ur \<or> (?I ?Dt, ?I ?Ds) \<in> ?CeN" by auto
        from always have piece0: "(Fun f (map ?I bef @ ?I ?Dt # map ?I aft), 
	        Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) \<in> ?R" (is "(?Il, ?Ir) \<in> _")
          using ccR unfolding one_step_ctxt_closed_def by auto
        from choice have ind: "(Fun f ?bta, Fun f ?bsa) \<in> rstep ?ur \<or> (?Il, ?Ir) \<in> ?CeN"
        proof
	  assume "(?Dt, ?Ds) \<in> rstep ?ur"
	  with one_step_ctxt_closed_rstep[where r = ?ur] show ?thesis unfolding one_step_ctxt_closed_def by auto
        next
          assume "(?I ?Dt, ?I ?Ds) \<in> ?CeN"
          with ccCN one_step_ctxt_closed_rstep[where r = ur] show ?thesis unfolding one_step_ctxt_closed_def by auto
        qed
	from True snbta have left: "?I (Fun f ?bta) = ?Il" by auto
	from step snbta have snbsa: "SN_elt (rstep (set R)) (Fun f ?bsa)" by (rule step_preserves_SN_elt)
	from More ctxt_closureI[where C = D] have "(?Dt,?Ds) \<in> rstep (set R)"
          unfolding rstep_CC_CS by auto
	hence subset: "[cap (set R) ?Ds] \<subseteq> [cap (set R) ?Dt]" by (rule capRewrite)		
	have "i_trans_check R ur us (Fun f ?bsa)"
	proof (simp add: i_trans_check.simps fus, rule ballI) 
	  fix lr
	  assume lr: "lr \<in> set R - ur"
	  show "\<not> (\<lambda>(l, r). gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l) lr"
	  proof (cases lr)
	    case (Pair l r)
	    from lr Pair have "(l,r) \<in> set R" by auto
	    from Pair lr have nMatch: 
	      "\<not> gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Dt # map (cap (set R)) aft)) l" (is "\<not> gcMatch ?cbta l") 
	      using True by (auto simp: i_trans_check.simps)
	    {
	      assume "gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l" 
	      from this subset have "gcMatch ?cbta l" by (rule gcMatchBelow)	    
	      with nMatch have False ..
	    }
	    with lr Pair show ?thesis by auto 
	  qed
	qed
	with snbsa  have "?I (Fun f ?bsa) = ?Ir" by auto	
	with left piece0 ind show ?thesis by (simp)
      qed
    qed
  qed
qed
end
no_notation Edg.equiv_class ("[_]")
	
context ce_af_redpair
begin

lemma i_transVI: assumes 
  m: "m \<ge> n" and
  sn: "SN_elt (rstep (set R)) t" and
  step: "(t,s) \<in> (rstep (set R))^*" and
  ur_closed: "ur_closed_af ur us R \<pi>" and
  orient: "ur \<subseteq> NS"
  shows "(i_trans (R,ur,us,(c,d,m)) t, i_trans (R,ur,us,(c,d,m)) s) \<in> NS^* \<and>
  ((i_trans (R,ur,us,(c,d,m)) t, i_trans (R,ur,us,(c,d,m)) s) \<in> 
  NS^* O rstep (ce_trs (c,d,m)) O NS^*
\<or> (t,s) \<in> (rstep (ur - {lr | lr. fst lr \<notin> wf_terms us}))^*)" (is "?goal t s")
using step 
proof (induct rule: rtrancl_induct, simp)
  case (step s u)
  let ?cn = "(c,d,m)"
  let ?ur = "rstep ur"
  let ?urm = "rstep (ur - {lr | lr. fst lr \<notin> wf_terms us})"
  let ?c = "rstep (ce_trs ?cn)"
  let ?R = "NS^* O ?c O NS^*"
  let ?I = "i_trans (R,ur,us,?cn)"
  from step(1) sn have "SN_elt (rstep (set R)) s" 
    by (rule steps_preserve_SN_elt[of _ _ "rstep (set R)"])
  from i_transV[OF m this step(2) ur_closed orient] have isteps2: "(?I s, ?I u) \<in> NS^*" and choice: "((s,u) \<in> ?urm \<or> (?I s, ?I u) \<in> ?c^+ O NS^*)" by auto
  from ce_orient[OF m] have subset2: "?c^+ \<subseteq> NS^*" by auto
  from subset2 step(3) have "(?I t, ?I s) \<in> NS^*" by auto
  with isteps2 have oneGoal: "(?I t, ?I u) \<in> NS^*" by auto
  thus "?goal t u" 
  proof (rule conjI) 
    from choice
    show "(?I t, ?I u) \<in> ?R \<or> (t,u) \<in> ?urm^*"
    proof
      assume su: "(s,u) \<in> ?urm" 
      from step(3) have "(?I t, ?I s) \<in> ?R \<or> (t,s) \<in> ?urm^*" ..
      thus ?thesis
      proof
      	assume "(t,s) \<in> ?urm^*"
  	with su have "(t,u) \<in> ?urm^*" by auto
  	thus ?thesis .. 
      next
  	assume "(?I t, ?I s) \<in> ?R"
        with isteps2 have "(?I t, ?I u) \<in> ?R O NS^*" by auto
        hence "(?I t, ?I u) \<in> ?R" by (simp only: O_assoc, auto)
    	thus ?thesis ..
      qed
    next
      assume "(?I s, ?I u) \<in> ?c^+ O NS^*"
      with ce_orient[OF m] have "(?I s, ?I u) \<in> NS^* O ?c O NS^*" by (auto simp: trancl_unfold_right)
      with `(?I t, ?I s) \<in> NS^*` have "(?I t, ?I u) \<in> NS^* O NS^* O ?c O NS^*" by auto
      hence "(?I t, ?I u) \<in> (NS^* O NS^*) O ?c O NS^*" by (simp only: O_assoc)
      thus ?thesis by auto
    qed
  qed
qed 
end


declare i_trans.simps[simp del]    

abbreviation ur_P_closed_af :: "('f,'v)rules \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool" 
  where "ur_P_closed_af R ur us P \<pi> \<equiv> \<forall> st \<in> P. ur_closed_term_af (snd st) R ur us \<pi>"


context ce_af_redpair
begin
lemma i_trans_sound: assumes
  m: "m \<ge> n"
  and chain: "min_ichain (P, set R) s t \<sigma>"
  and ur_closed: "ur_closed_af ur us R \<pi>"
  and ur_P_closed: "ur_P_closed_af R ur us P \<pi>"
  and orient: "ur \<subseteq> NS"
  shows  "ichain (P, NS) s t (\<lambda> i. i_trans_subst (R,ur,us,(c,d,m)) (\<sigma> i)) \<and> 
        (\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (ur - {lr | lr. fst lr \<notin> wf_terms us}))^* \<and> s (Suc i) \<in> wf_terms us \<or> 
         (t i \<cdot> i_trans_subst (R,ur,us,(c,d,m)) (\<sigma> i),
          s (Suc i) \<cdot> i_trans_subst (R,ur,us,(c,d,m)) (\<sigma> (Suc i))) \<in> 
          NS^* O (rstep (ce_trs (c,d,m))) O NS^* )"
proof -
  let ?cn = "(c,d,m)"
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?I = "i_trans (R, ur, us, ?cn)"
  let ?Is = "\<lambda> i. (i_trans_subst (R,ur,us,?cn) (\<sigma> i))"
  let ?sIsig = "\<lambda> i. (s i) \<cdot> ?Is i"
  let ?tIsig = "\<lambda> i. (t i) \<cdot> ?Is i"
  let ?ce = "rstep (ce_trs ?cn)"
  let ?ur = "rstep (ur - {lr | lr. fst lr \<notin> wf_terms us})"
  let ?R = "NS^* O ?ce O NS^*"
  from chain have allP: "\<forall> i. (s i, t i) \<in> P"
    and steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep (set R))^*"
    and snt: "\<forall> i. SN_elt (rstep (set R)) (?ts i)" by (auto simp: minimal_cond_def)
  have main: "\<forall> i. (?tIsig i, ?sIsig (Suc i)) \<in> NS^* \<and> ((?ts i, ?ss (Suc i)) \<in> ?ur^* \<and> s (Suc i) \<in> wf_terms us \<or> (?tIsig i, ?sIsig (Suc i)) \<in> ?R)" (is "\<forall> i. ?goal1 i \<and> ?goal2 i")
  proof
    fix i
    from steps snt have snti: "SN_elt (rstep (set R)) (?ts i)" and stepst: "(?ts i, ?ss (Suc i)) \<in> (rstep (set R))^*" by auto
    from this ur_closed orient have one: "(?I (?ts i), ?I (?ss (Suc i))) \<in> NS^*
      \<and> ((?I (?ts i), ?I (?ss (Suc i))) \<in> ?R \<or> (?ts i, ?ss (Suc i)) \<in> ?ur^* )" (is "_ \<and> (?IR \<or> ?oU)") 
      by (rule i_transVI[OF m])    
    from allP have "(s i, t i) \<in> P" by auto 
    with ur_P_closed have "ur_closed_term_af (t i) R ur us \<pi>" by auto
    from i_transI[OF this snti] ce_orient have It: "(t i \<cdot> ?Is i, ?I (?ts i)) \<in> NS^*" by simp
    with one have two: "(?tIsig i, ?I (?ss (Suc i))) \<in> NS^*" by auto    
    from stepst snti have snsi: "SN_elt (rstep (set R)) (?ss (Suc i))" by (rule steps_preserve_SN_elt[of _ _ "rstep(set R)"])
    have "(?I (?ss (Suc i)), s (Suc i) \<cdot> ?Is (Suc i)) \<in> ?ce^*" 
      and cep: "s (Suc i) \<notin> wf_terms us \<longrightarrow> (?I (?ss (Suc i)), s (Suc i) \<cdot> ?Is (Suc i)) \<in> ?ce^+" using  i_transII[OF snsi] by blast+
    with ce_orient[OF m] have Is: "(?I (?ss (Suc i)), ?sIsig (Suc i)) \<in> NS^*" by auto    
    with two have part1: "?goal1 i" by auto    
    from one have "?IR \<or> ?oU" ..
    hence "?goal2 i"
    proof
      assume ?IR
      with It Is have "(?tIsig i, ?sIsig (Suc i)) \<in> (NS^* O ?R) O NS^*" by auto
      hence "(?tIsig i, ?sIsig (Suc i)) \<in> (NS^* O NS^*) O ?ce O (NS^* O NS^*)" by (simp only: O_assoc)
      thus ?thesis by auto
    next
      assume "?oU"
      show ?thesis 
      proof (cases "s (Suc i) \<in> wf_terms us")
        case True
        with `?oU` show ?thesis by auto
      next
        case False
        have "?ce^+ = ?ce O ?ce^*" by (rule trancl_unfold_left)
        also have "\<dots> \<subseteq> ?ce O NS^*" using ce_orient[OF m] by auto
        finally have cep: "(?I (?ss (Suc i)), s (Suc i) \<cdot> ?Is (Suc i)) \<in> ?ce O NS^*" using cep False by auto
        with two show ?thesis by auto 
      qed
    qed
    with part1 
    show "?goal1 i \<and> ?goal2 i" ..
  qed 
  with chain allP snt rtrancl_mono[OF subset_rstep[of NS]] show ?thesis by auto
qed

lemma redPairUrSound :
  assumes rules: "ur \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and ur_closed: "ur_closed_af ur us R \<pi>"
  and ur_P_closed: "ur_P_closed_af R ur us P \<pi>"
  and minChain: "min_ichain (P,set R) s t \<sigma>"
  shows  "\<exists> j. min_ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  from minChain ur_closed ur_P_closed rules have chain: "ichain (P, NS) s t (\<lambda> i. i_trans_subst (R,ur,us,(c,d,n)) (\<sigma> i))" 
    (is "ichain _ _ _ ?Isig") using i_trans_sound[of n] by blast
  from redPairSound[OF _ nonStrict strict chain]
  have "\<exists>j. ichain (rP, NS) (shift s j) (shift t j) (shift ?Isig j)" by auto
  from this obtain j where ichain: "ichain (rP, NS) (shift s j) (shift t j) (shift ?Isig j)" ..
  have "min_ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" 
    using min_ichainI[where R' = "set R" and R = "set R"] using ichain minChain ci_subset_refl[of "set R"] by auto
  thus ?thesis ..
qed
end

context mono_ce_af_redpair
begin
lemma monoRedPairUrSound :
  assumes monotonic: "one_step_ctxt_closed S"
  and rNonStrict: "ur \<subseteq> NS \<union> S"
  and rStrict: "ur - {lr | lr. fst lr \<notin> wf_terms us} - rR \<subseteq> S"
  and nonStrict: "P \<subseteq> NS \<union> S"
  and strict: "P - {lr | lr. fst lr \<notin> wf_terms us} - rP \<subseteq> S"
  and subset: "rR \<subseteq> set R"
  and ur_closed: "ur_closed_af ur us R \<pi>"
  and ur_P_closed: "ur_P_closed_af R ur us P \<pi>"
  and minChain: "min_ichain (P,set R) s t \<sigma>"
  shows  "\<exists> j. min_ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  from S_ce_compat obtain n' where S_compat: "\<And> m. m \<ge> n' \<Longrightarrow> ce_trs (c,d,m) \<subseteq> S" by blast
  let ?NS = "NS \<union> S"
  let ?n = "max n (max n' ( ce_redpair.n ?NS))"
  from trs_inclusion[OF monotonic substS] S_compat have orientCe: "rstep (ce_trs (c,d,?n)) \<subseteq> \<succ>" by auto
  let ?S = "?NS^* O S O ?NS^*"
  let ?ur = "ur - {lr | lr. fst lr \<notin> wf_terms us}"
  let ?P = "P - {lr | lr. fst lr \<notin> wf_terms us}"
  let ?Isig = "(\<lambda> i. i_trans_subst (R,ur,us,(c,d,?n)) (\<sigma> i))"
  let ?c = "rstep (ce_trs (c,d,?n))"
  let ?ncn = "?NS^* O ?c O ?NS^*"  
  let ?NSS = "?NS^* \<union> ?S"
  have "?NS^* O ?S = (?NS^* O ?NS^*) O S O ?NS^*" by (simp only: O_assoc)
  hence ns_s: "?NS^* O ?S = ?S" by auto
  hence subset2: "?NS^* O ?S \<subseteq> ?S" by auto
  from rNonStrict have orient: "ur \<subseteq> ?NS" by auto
  have subst_NS: "subst_closed ?NS"
      by (rule subst_closed_union[OF substNS substS])
  have new_redpair: "ce_af_redpair ?S ?NS \<pi>"
  proof
    from compatible_SN'[OF compat SN]
    show "SN ?S" .
  next
    show "subst_closed ?NS" by (rule subst_NS)
  next
    show "subst_closed ?S"
      by (rule subst_closed_comp, rule subst_closed_rtrancl[OF subst_NS], rule subst_closed_comp[OF substS], rule subst_closed_rtrancl[OF subst_NS])
  next
    have "?NS O ?S \<subseteq> ?NS^* O ?S" by auto
    also have "\<dots> \<subseteq> ?S" by (rule subset2)
    finally show "?NS O ?S \<subseteq> ?S" .
  next
    from S_ce_compat
    show "\<exists> n. \<forall> m \<ge> n. (\<forall> c d. ce_trs (c,d,m) \<subseteq> ?NS)" by blast
  next
    show "one_step_ctxt_closed ?NS" using wmNS monotonic unfolding one_step_ctxt_closed_def by blast
  next
    show "af_compatible \<pi> ?NS" using af_compat unfolding af_compatible_def by blast
  qed
  obtain \<delta> where delta: "\<delta> = ?Isig" by auto
  from ce_af_redpair.i_trans_sound[OF new_redpair _ minChain ur_closed ur_P_closed orient, of ?n]
  have ichain: "ichain (P, ?NS) s t \<delta> \<and>
  (\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep ?ur)^* \<and> s (Suc i) \<in> wf_terms us \<or> 
         (t i \<cdot> \<delta> i,
          s (Suc i) \<cdot> \<delta> (Suc i)) \<in> 
          ?ncn)"
    (is "?one \<and> (\<forall> i. ?cas1 i \<or> ?cas2 i)") by (simp add: delta) 
  hence "?one" ..
  hence "\<exists> j. ichain (?P, ?ur) (shift s j) (shift t j) (shift \<sigma> j)"
  proof (cases "\<exists> j. \<forall> k. ?cas1 (k + j)")
    case True
    from this obtain j where all: "\<forall> k. ?cas1 (k + j)" by auto
    have all2: "\<forall> k. ?cas1 (k + Suc j)" 
    proof
      fix k
      show "?cas1 (k + Suc j)" using spec[OF all, of "Suc k"] by auto
    qed
    from all all2 minChain have "ichain (?P, ?ur) (shift s (Suc j)) (shift t (Suc j)) (shift \<sigma> (Suc j))" by simp 
    thus ?thesis ..
  next
    case False
    with ichain have infCe: "\<forall> j. \<exists> k. ?cas2 (k + j)" by blast
    from orientCe have "?ncn \<subseteq> ?NS^* O S O ?NS^*" by auto
    also have "\<dots> \<subseteq> ?NS^* O ?NS^* O ?NS^*" using rtrancl_mono by blast
    also have "\<dots> = ?NS^*" by auto
    finally have subset1: "?ncn \<subseteq> ?NS^*" .
    have "\<forall> i. (s i \<cdot> \<delta> i, t i \<cdot> \<delta> i) \<in> ?NS"
    proof 
      fix i
      show "(s i \<cdot> \<delta> i, t i \<cdot> \<delta> i) \<in> ?NS"
      proof (cases "(s i, t i) \<in> \<succ>")
	case True
	with substS show ?thesis unfolding subst_closed_def using subst_closureI by blast
      next
	case False
	with strict nonStrict minChain have "(s i, t i) \<in> \<succeq>" by auto
	with substNS show ?thesis unfolding subst_closed_def using subst_closureI by blast
      qed
    qed
    hence orientP: "\<forall> i. (s i \<cdot> \<delta> i, t i \<cdot> \<delta> i) \<in> ?NS^*" by auto
    have rstepNS: "rstep NS = NS" using subset_rstep[of NS] trs_inclusion[OF wmNS substNS] by blast
    have rstepS: "rstep S = S" using subset_rstep[of S] trs_inclusion[OF monotonic substS] by blast
    from rstepNS rstepS have rstep_NS: "rstep ?NS = ?NS" by (simp add: rstep_union)
    let ?NSS = "?NS^* \<union> ?S"
    have "?NS^* O ?S = (?NS^* O ?NS^*) O S O ?NS^*" by (simp only: O_assoc)
    hence ns_s: "?NS^* O ?S = ?S" by auto
    hence subset2: "?NS^* O ?S \<subseteq> ?S" by auto
    let ?f = "\<lambda> i. s i \<cdot> \<delta> i"
    from compatible_SN'[OF compat SN] have SN3: "SN_elt ?S (?f 0)" unfolding SN_def ..
    have "\<forall> i. (?f i, ?f (Suc i)) \<in> ?NSS"
    proof
      fix i
      from ichain have "(t i \<cdot> \<delta> i, ?f (Suc i)) \<in> ?NS^*" by (auto simp: rstep_NS)
      with spec[OF orientP, of i] show "(?f i, ?f (Suc i)) \<in> ?NSS" by auto
    qed
    from non_strict_ending[OF this subset2] SN3 obtain j where
      non_strict: "\<forall> i \<ge> j. (?f i, ?f (Suc i)) \<in> ?NS^* - ?S" by blast
    from spec[OF infCe, of j] obtain k where "(t (k+ j) \<cdot> \<delta> (k + j), ?f (Suc (k+j))) \<in> ?NS^* O ?c O ?NS^*" (is "(?t,?s) \<in> _") ..
    with orientCe have "(?t,?s) \<in> ?S" by auto
    with spec[OF orientP, of "k + j"] have "(?f (k+j), ?s) \<in> ?NS^* O ?S" by auto
    hence "(?f (k+j), ?s) \<in> ?S" by (simp only: ns_s)
    with spec[OF non_strict, of "k+j"] have False by auto
    thus ?thesis ..
  qed 
  from this obtain j where ichain: "ichain (?P,?ur) (shift s j) (shift t j) (shift \<sigma> j)" ..  
  from rNonStrict have rnon: "rR \<inter> ?ur \<subseteq> ?NS" by auto
  from nonStrict have nstri: "rP \<inter> ?P \<subseteq> ?NS" by auto
  from monoRedPairSound[OF monotonic rnon rStrict nstri strict ichain]
  have "\<exists> k. ichain (rP,rR) (shift (shift s j) k) (shift (shift t j) k) (shift (shift \<sigma> j) k)" (is "\<exists> k. ?chain k") .
  from this obtain k where kjChain: "?chain k" ..
  have shift: "\<And> f. shift (shift f j) k = shift f (k + j)" 
  proof (rule ext, simp)
    fix f x
    have "x + k + j = (x + ( k + j))" by auto
    thus "f (x + k + j) = f (x + ( k + j))" by (rule arg_cong)
  qed
  from kjChain have inf_chain: "ichain (rP,rR) (shift s (k + j)) (shift t (k + j)) (shift \<sigma> (k + j))" 
    by (simp only: shift[where f = s] shift[where f = t] shift[where f = \<sigma>])
  have "min_ichain (rP, rR) (shift s (k + j)) (shift t (k + j)) (shift \<sigma> (k + j))" 
    using min_ichainI[where R' = "set R" and R = "rR"] using inf_chain minChain ci_subsetI[OF subset]  by auto
  thus ?thesis ..
qed
end

definition \<pi>_full :: "'f af"
where "\<pi>_full _ _ _ = True"

lemma \<pi>_full_compatible: "af_compatible \<pi>_full r"
unfolding af_compatible_def \<pi>_full_def by auto


abbreviation redPairUrProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> bool"
 where "redPairUrProc S NS ur us P R rm rrm rP \<equiv> ur_closed_af ur us R \<pi>_full \<and> ur_P_closed_af R ur us (set P) \<pi>_full \<and> ur \<subseteq> NS \<and> (set rP) \<inter> (set P) \<subseteq> NS \<union> S \<and> (set P) - (set rP) \<subseteq> S"


abbreviation monoRedPairUrProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> bool"
 where "monoRedPairUrProc S NS ur us P R rm rP rR \<equiv> one_step_ctxt_closed S \<and> ur_closed_af ur us R \<pi>_full \<and> ur_P_closed_af R ur us (set P) \<pi>_full \<and> ur \<subseteq> NS \<union> S \<and> ur - {lr | lr. fst lr \<notin> wf_terms us} - (set rR) \<subseteq> S \<and> set P \<subseteq> NS \<union> S \<and> (set P) - {lr | lr. fst lr \<notin> wf_terms us} - (set rP) \<subseteq> S \<and> set rR \<subseteq> set R"

context ce_redpair
begin

lemma ce_af_redpair_\<pi>_full: "ce_af_redpair S NS \<pi>_full"
proof
  show "af_compatible \<pi>_full NS" by (rule \<pi>_full_compatible)
qed

lemma redPairUrProc_ci: "ci_only_pairs_proc_impl (redPairUrProc \<succ> \<succeq> ur us)"
unfolding ci_only_pairs_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume proc: "redPairUrProc \<succ> \<succeq> ur us P R rm rrm rP"
  show "\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (intro allI, intro impI)
    fix s t \<sigma>
    assume min: "min_ichain (set P, set R) s t \<sigma>"
    with proc have "\<exists> j. min_ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" using ce_af_redpair.redPairUrSound[OF ce_af_redpair_\<pi>_full] by auto
    thus "\<exists> j. ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" by force
  qed
qed
end


context mono_ce_redpair
begin

lemma mono_ce_af_redpair_\<pi>_full: "mono_ce_af_redpair S NS \<pi>_full"
proof
  show "af_compatible \<pi>_full NS" by (rule \<pi>_full_compatible)
qed

lemma monoRedPairUrProc_ci: "ci_proc_impl (monoRedPairUrProc \<succ> \<succeq> ur us)"
unfolding ci_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP rR
  assume proc: "monoRedPairUrProc \<succ> \<succeq> ur us P R rm rP rR"
  show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (intro conjI, simp add: proc ci_subsetI[of "set rR" "set R"], intro allI, intro impI)
    fix s t \<sigma>
    assume min: "min_ichain (set P, set R) s t \<sigma>"
    with proc have "\<exists> j. min_ichain (set rP, set rR) (shift s j) (shift t j) (shift \<sigma> j)" using mono_ce_af_redpair.monoRedPairUrSound[OF mono_ce_af_redpair_\<pi>_full] by auto
    thus "\<exists> j. ichain (set rP, set rR) (shift s j) (shift t j) (shift \<sigma> j)" by force
  qed
qed
end


abbreviation redPairUrAfProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> 'f af \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> bool"
 where "redPairUrAfProc S NS ur us \<pi> P R rm rrm rP \<equiv> ur_closed_af ur us R \<pi> \<and> ur_P_closed_af R ur us (set P) \<pi> \<and> ur \<subseteq> NS \<and> (set rP) \<inter> (set P) \<subseteq> NS \<union> S \<and> (set P) - (set rP) \<subseteq> S"

context ce_af_redpair 
begin

lemma redPairUrAfProc_ci: "ci_only_pairs_proc_impl (redPairUrAfProc \<succ> \<succeq> ur us \<pi>)"
unfolding ci_only_pairs_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume proc: "redPairUrAfProc \<succ> \<succeq> ur us \<pi> P R rm rrm rP"
  show "\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (intro allI, intro impI)
    fix s t \<sigma>
    assume min: "min_ichain (set P, set R) s t \<sigma>"
    with proc have "\<exists> j. min_ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" using redPairUrSound by auto
    thus "\<exists> j. ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" by force
  qed
qed
end


fun matchCapBelowI :: "('f,'v)term \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "matchCapBelowI l R (Fun f ts) = gcMatchI (HFun f (map (capI R) ts)) l"

lemma matchCapBelowISound[simp]:  "matchCapBelowI l R (Fun f ts) = matchCapBelow l R (Fun f ts)"
proof - 
  have cap: "capI R = cap (set R)"
    by (rule ext, simp add: capISound)
  show ?thesis by (simp only: cap gcMatchISound matchCapBelowI.simps matchCapBelow.simps)
qed

declare matchCapBelowI.simps[simp del]

fun matchCapRMBelow :: "('f,'v)term \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "matchCapRMBelow l rm (Fun f ts) = gcMatchI (HFun f (map (capRM rm) ts)) l"

lemma matchCapRMBelowSound[simp]:  "matchCapRMBelow l (computeRuleMap R) (Fun f ts) = matchCapBelow l R (Fun f ts)"
proof - 
  have cap: "capRM (computeRuleMap R) = cap (set R)"
    by (rule ext, simp add: capRMSound)
  show ?thesis by (simp only: cap gcMatchISound matchCapRMBelow.simps matchCapBelow.simps)
qed

declare matchCapRMBelow.simps[simp del]

fun
  check_ur_closed_term_af :: "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> 'f af \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_ur_closed_term_af _ _ _ (Var x) = OK" |
  "check_ur_closed_term_af R ur \<pi> (Fun f ts) = do {
     let n = length ts;
     check_allm_index (\<lambda>t i. if \<pi> f n i then check_ur_closed_term_af R ur \<pi> t else OK) ts;
     check_allm (\<lambda>lr.
       check (lr mem ur \<or> \<not> matchCapBelowI (fst lr) R (Fun f ts))
         (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
     ) R
   }"

fun
  check_ur_closed_term_rm_af :: "(('f \<times> nat) \<Rightarrow> ('f, 'v) rules) \<Rightarrow> ('f::show, 'v::show) rules \<Rightarrow> 'f af \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_ur_closed_term_rm_af _ _ _ (Var x) = OK" |
  "check_ur_closed_term_rm_af rm ur \<pi> (Fun f ts) = do {
     let n = length ts;
     check_allm_index (\<lambda>t i. if \<pi> f n i then check_ur_closed_term_rm_af rm ur \<pi> t else OK) ts;
     check_allm (\<lambda>lr.
       check (lr mem ur \<or> \<not> matchCapRMBelow (fst lr) (Some rm) (Fun f ts)) 
        (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
     ) (rm (f,n))
   }"

lemma check_ur_closed_term_af_sound: 
  assumes "isOK(check_ur_closed_term_af R ur \<pi> t)" 
    and "funas_term t \<subseteq> us"
  shows "ur_closed_term_af t R (set ur) us \<pi>"
using assms proof (induct t)
  case (Fun f ts)
  from Fun(3) have fus: "(f, length ts) \<in> us" by auto
  from Fun(3) have "\<forall> i < length ts. funas_term (ts ! i) \<subseteq> us" by force
  with Fun(1) Fun(2)
    have "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapBelow l R (Fun f ts)" 
    and "(\<forall>i<length ts. \<pi> f (length ts) i \<longrightarrow> ur_closed_term_af (ts ! i) R (set ur) us \<pi>)" by (auto simp: Let_def mem_iff)
  thus ?case using fus by auto
qed simp

lemma check_ur_closed_term_rm_af_sound:
  assumes check: "isOK(check_ur_closed_term_rm_af rm ur \<pi> t)" 
    and tus: "funas_term t \<subseteq> us"
    and ruleMap: "computeRuleMap R = Some rm"
  shows "ur_closed_term_af t R (set ur) us \<pi>"
using check tus proof (induct t)
  case (Fun f ts)
  from Fun(3) have fus: "(f, length ts) \<in> us" by auto
  from Fun(3) have "\<forall> i < length ts. funas_term (ts ! i) \<subseteq> us" by force
  with Fun(1) Fun(2)
  have rls: "\<forall> (l,r) \<in> set (rm (f,length ts)). (l,r) \<in> set ur \<or> \<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)"
    and ind: " (\<forall>i<length ts. \<pi> f (length ts) i \<longrightarrow> ur_closed_term_af (ts ! i) R (set ur) us \<pi>)" by (auto simp: Let_def mem_iff ruleMap) 
  from ruleMap computeRuleMapSound2[where R = R]  have noLeftVar: "\<forall> (l,r) \<in> set R. root l \<noteq> None" by auto
  from ruleMap computeRuleMapSound[where R = R and rm = rm and f = f] have fLookup: "set (rm (f,length ts)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = length ts}" by auto
  have "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" (is "\<forall> lr \<in> set R. ?P lr") 
  proof
    fix lr
    assume lr: "lr \<in> set R"
    show "?P lr"
    proof (cases lr)
      case (Pair l r)
      show ?thesis
      proof (cases "(l,r) \<in> set (rm (f,length ts))")
	case True
	with Pair rls show ?thesis by auto
      next
	case False
	from noLeftVar lr Pair obtain g ss where l: "l = Fun g ss" by (cases l, auto)
	with fLookup False lr Pair have "\<not> (f = g \<and> length ss = length ts)" by auto
	with l ruleMap have "\<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" by (auto simp: matchCapRMBelow.simps gcMatch_def)
	with Pair rls show ?thesis by auto
      qed
    qed
  qed
  with ind fus show ?case by auto
qed simp 

fun
  check_ur_P_closed_af :: "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> 'f af \<Rightarrow> shows check"
where
  "check_ur_P_closed_af R ur P \<pi> = do {
     check_allm (\<lambda>lr. check_ur_closed_term_af R ur \<pi> (snd lr)) ur
       <+? (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s);
     check_allm (\<lambda>st. check_ur_closed_term_af R ur \<pi> (snd st)) P
       <+? (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
   }"

lemma check_ur_P_closed_sound_af:
  assumes check: "isOK(check_ur_P_closed_af R ur P \<pi>)"
    and us: "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P))) \<subseteq> us"
  shows "ur_closed_af (set ur) us R \<pi>" and "ur_P_closed_af R (set ur) us (set P) \<pi>"
proof -
  {
    fix l r
    assume lr: "(l,r) \<in> set ur"
    with us have rus: "funas_term r \<subseteq> us" by force
    from check_ur_closed_term_af_sound[OF _ rus] check lr
    have "ur_closed_term_af r R (set ur) us \<pi>" by auto
  } thus "ur_closed_af (set ur) us R \<pi>" by auto
next
  {
    fix l r
    assume lr: "(l,r) \<in> set P"
    with us have rus: "funas_term r \<subseteq> us" by force
    from check_ur_closed_term_af_sound[OF _ rus] check lr
    have "ur_closed_term_af r R (set ur) us \<pi>" by auto
  } thus "ur_P_closed_af R (set ur) us (set P) \<pi>" by auto
qed

declare check_ur_P_closed_af.simps[simp del]

fun
  check_ur_P_closed_rm_af ::
    "('f::show, 'v::show) rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> 'f af \<Rightarrow> shows check"
where
  "check_ur_P_closed_rm_af R rmo ur P \<pi> =
     (case rmo of
       None \<Rightarrow> check_ur_P_closed_af R ur P \<pi>
     | Some rm \<Rightarrow> do {
       check_allm (\<lambda>lr. check_ur_closed_term_rm_af rm ur \<pi> (snd lr)) ur
         <+? (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s);
       check_allm (\<lambda>st. check_ur_closed_term_rm_af rm ur \<pi> (snd st)) P
         <+? (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
     })"

lemma check_ur_P_closed_rm_af_sound:
  assumes check: "isOK(check_ur_P_closed_rm_af R (computeRuleMap R) ur P \<pi>)"
    and us: "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P))) \<subseteq> us"
  shows "ur_closed_af (set ur) us R \<pi> \<and> ur_P_closed_af R (set ur) us (set P) \<pi>"
using check
proof (cases "computeRuleMap R", simp add: check_ur_P_closed_sound_af[OF _ us])
  case (Some rm)
  {
    fix l r
    assume lr: "(l,r) \<in> set ur"
    with us have rus: "funas_term r \<subseteq> us" by force
    from check_ur_closed_term_rm_af_sound[OF _ rus Some] Some check lr
    have "ur_closed_term_af r R (set ur) us \<pi>" by auto
  } note part1 = this
  {
    fix l r
    assume lr: "(l,r) \<in> set P"
    with us have rus: "funas_term r \<subseteq> us" by force
    from check_ur_closed_term_rm_af_sound[OF _ rus Some] Some check lr
    have "ur_closed_term_af r R (set ur) us \<pi>" by auto
  } note part2 = this
  from part1 part2 show ?thesis by auto
qed

declare check_ur_P_closed_rm_af.simps[simp del]

fun
  check_redpair_ur_proc ::
    "('f::show, 'v::show)redpair_impl \<Rightarrow>
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_redpair_ur_proc rp P rP r rm ur  = do {
     let (Sp :: ('f,'v) rules ,NSp) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) P;
     rp_valid rp;
     check_ur_P_closed_rm_af r rm ur P \<pi>_full
       <+? (\<lambda>s. ''problem with closure properties of usable rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) ur
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSp
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_subseteq NSp rP
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     rp_desc rp +@+ shows_nl +@+ ''and the following usable rules'' +#+ shows_nl +@+ shows_rules ur +@+ shows_nl +@+ s)"



fun
  check_redpair_ur_af_proc ::
    "('f::show, 'v::show)redpair_impl \<Rightarrow> 
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_redpair_ur_af_proc rp P rP r rm ur = do {
     let (Sp :: ('f,'v) rules ,NSp) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) P;
     rp_valid rp;
     check_ur_P_closed_rm_af r rm ur P (rp_af rp)
       <+? (\<lambda>s. ''problem with closure properties of usable rules w.r.t. argument filter'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) ur
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSp
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_subseteq NSp rP
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     rp_desc rp +@+ shows_nl +@+ ''and the following usable rules'' +#+ shows_nl +@+ shows_rules ur +@+ shows_nl +@+ s)"


fun
  check_mono_redpair_ur_proc ::
    "('f::show, 'v::show)redpair_impl \<Rightarrow> 
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_mono_redpair_ur_proc rp P rP R rm rR ur = do {
     let us = concat (map (funas_term_list o snd) (P @ ur));
     let filt = filter (\<lambda> lr. wf_terms_list us (fst lr));
     let (Sp :: ('f,'v) rules , NSp) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) P;
     let (Sr :: ('f,'v) rules , NSr) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) ur;
     rp_valid rp;
     rp_mono rp
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_ur_P_closed_rm_af R rm ur P \<pi>_full
       <+? (\<lambda>s. ''problem with closure properties of usable rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSr
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSp
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_subseteq rR R
       <+? (\<lambda>s. ''problem: additional rule(s) to TRS added'' +#+ shows_nl);
     check_subseteq (filt NSr) rR
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error);
     check_subseteq (filt NSp) rP
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
   } <+? (\<lambda>s. ''could not apply the (monotonic) reduction pair processor with the following'' +#+ shows_nl +@+
     rp_desc rp +@+ shows_nl +@+ ''and the following usable rules'' +#+ shows_rules ur +@+ shows_nl +@+ s)"


context generic_redpair_impl
begin
lemma check_redpair_ur_proc_sound :
  shows "ci_only_pairs_proc_impl (\<lambda> P R rm rrm rP. isOK(check_redpair_ur_proc (generate_redpair rp) P rP R rm ur))"
unfolding ci_only_pairs_proc_impl_def
proof (intro allI, intro impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix P R rP 
  assume check2: "isOK (check_redpair_ur_proc ?rp P rP R (computeRuleMap R) ur)"
  let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P)))"
  show "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (cases "List.partition (\<lambda> dp. isOK(rp_strict ?rp dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check2 have valid: "isOK(rp_valid ?rp)" and cond:
      "isOK(check_allm (rp_strict ?rp) Sp)" and "isOK(check_allm (rp_non_strict ?rp) (NSp @ ur))" by (auto simp: Let_def mem_iff) 
    from generate_redpair[OF this] obtain S NS where redpair: "ce_af_redpair S NS ?af" and S: "set Sp \<subseteq> S" and NS: "set (NSp @ ur) \<subseteq> NS" by auto
    from Pair check2 have ur: "isOK(check_ur_P_closed_rm_af R (computeRuleMap R) ur P \<pi>_full)" by auto
    have ur: "ur_closed_af (set ur) ?us R \<pi>_full \<and> ur_P_closed_af R (set ur) ?us (set P) \<pi>_full " using check_ur_P_closed_rm_af_sound[OF ur] by auto
    from Pair check2 S NS have "set P - set NSp \<subseteq> S \<and> ?NSp \<inter> set P \<subseteq> NS \<union> S \<and> set NSp \<subseteq> set rP" by (auto)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
    have proc: "redPairUrProc S NS (set ur) ?us P R (computeRuleMap R) rrm rP" using ur stri NS by auto
    from redpair interpret ce_af_redpair S NS ?af .
    show ?thesis using redPairUrProc_ci proc unfolding ci_only_pairs_proc_impl_def by auto
  qed
qed

lemma check_redpair_ur_af_proc_sound :
  shows "ci_only_pairs_proc_impl (\<lambda> P R rm rrm rP. isOK(check_redpair_ur_af_proc (generate_redpair rp) P rP R rm ur))"
unfolding ci_only_pairs_proc_impl_def
proof (intro allI, intro impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix P R rP 
  assume check2: "isOK (check_redpair_ur_af_proc ?rp P rP R (computeRuleMap R) ur)"
  let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P)))"
  show "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (cases "List.partition (\<lambda> dp. isOK(rp_strict ?rp dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check2 have "isOK(rp_valid ?rp)" and cond:
      "isOK(check_allm (rp_strict ?rp) Sp)" and "isOK(check_allm (rp_non_strict ?rp) (NSp @ ur))" by (auto simp: Let_def mem_iff) 
    from generate_redpair[OF this] obtain S NS where redpair: "ce_af_redpair S NS ?af" and S: "set Sp \<subseteq> S" and NS: "set (NSp @ ur) \<subseteq> NS" by auto
    from Pair check2 have ur: "isOK(check_ur_P_closed_rm_af R (computeRuleMap R) ur P ?af)" by auto
    have ur: "ur_closed_af (set ur) ?us R ?af \<and> ur_P_closed_af R (set ur) ?us (set P) ?af" using check_ur_P_closed_rm_af_sound[OF ur] by auto
    from Pair check2 S NS have "set P - set NSp \<subseteq> S \<and> ?NSp \<inter> set P \<subseteq> NS \<union> S \<and> set NSp \<subseteq> set rP" by (auto)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
    have proc: "redPairUrAfProc S NS (set ur) ?us ?af P R (computeRuleMap R) rrm rP" using ur stri NS by auto
    from redpair interpret ce_af_redpair S NS ?af .
    show ?thesis using redPairUrAfProc_ci proc unfolding ci_only_pairs_proc_impl_def by auto
  qed
qed

lemma check_mono_redpair_ur_proc_sound :
  shows "ci_proc_impl (\<lambda> P R rm rP rR. isOK(check_mono_redpair_ur_proc (generate_redpair rp) P rP R rm rR ur))"
unfolding ci_proc_impl_def
proof (intro allI, intro impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix P R rP  rR
  assume check2: "isOK (check_mono_redpair_ur_proc ?rp P rP R (computeRuleMap R) rR ur)"
  let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P)))"
  let ?usi = "concat (map (funas_term_list o snd) (P @ ur))"
  have us: "set ?usi = ?us" by auto
  let ?filt = "filter (\<lambda> lr. wf_terms_list ?usi (fst lr))"
  {
    fix X
    have "set (?filt X) = set X \<inter> {lr | lr. wf_terms_list ?usi (fst lr)}" by auto
    also have "\<dots> = set X \<inter> {lr | lr. fst lr \<in> wf_terms ?us}" by (simp only: wf_terms_list_sound us)
    finally have "set (?filt X) = set X \<inter> {lr | lr. fst lr \<in> wf_terms ?us}" .
  } note filt = this
  show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (cases "List.partition (\<lambda> dp. isOK(rp_strict ?rp dp)) P")
    case (Pair Sp NSp) note pPairs = this
    let ?NSp = "set NSp \<inter> set rP"
    show ?thesis 
    proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) ur")
      case (Pair Sr NSr)
      from pPairs Pair check2 have subset: "set rR \<subseteq> set R" and  mono: "isOK(rp_mono ?rp)" and valid: "isOK(rp_valid ?rp)" and cond1:
        "isOK(check_allm (rp_strict ?rp) (Sp @ Sr))" and cond2: "isOK(check_allm (rp_non_strict ?rp) (NSp @ NSr))" 
      and cond: "set (?filt NSp) \<subseteq> set rP \<and> set (?filt NSr) \<subseteq> set rR" by (auto simp: Let_def mem_iff)    
      from generate_redpair[OF valid cond1 cond2] mono obtain S NS where redpair: "mono_ce_af_redpair S NS ?af" and S: "set Sp \<union> set Sr \<subseteq> S" and NS: "set NSp \<union> set NSr \<subseteq> NS" 
        and mono: "one_step_ctxt_closed S" by auto
      from check2 have ur: "isOK(check_ur_P_closed_rm_af R (computeRuleMap R) ur P \<pi>_full)" by auto
      have ur1: "ur_closed_af (set ur) ?us R \<pi>_full" and ur2: "ur_P_closed_af R (set ur) ?us (set P) \<pi>_full" using check_ur_P_closed_rm_af_sound[OF ur] by auto
      from cond pPairs S NS have condP:
        "set P - set NSp \<subseteq> S \<and> 
         set NSp \<inter> set P \<subseteq> NS \<union> S   \<and> 
         set (?filt NSp) \<subseteq> set rP"
        by auto
      hence rP: "set NSp - {lr | lr. fst lr \<notin> wf_terms (?us)} \<subseteq> set rP" by (simp only: filt) auto
      from condP have P: "set P \<subseteq> NS \<union> S" by blast
      from condP rP P have rP: "set P - {lr | lr. fst lr \<notin> wf_terms ?us} - set rP \<subseteq> S" by blast
      from cond Pair S NS have condR:
      "set ur - set NSr \<subseteq> S \<and> 
      set NSr \<inter> set ur \<subseteq> NS \<union> S   \<and> 
      set (?filt NSr) \<subseteq> set rR"
	by auto
      hence rR: "set NSr - {lr | lr. fst lr \<notin> wf_terms ?us} \<subseteq> set rR"
        by (simp only: filt) auto
      from condR have R: "set ur \<subseteq> NS \<union> S" by blast
      from condR rR R have rR: "set ur - {lr | lr. fst lr \<notin> wf_terms ?us} - set rR \<subseteq> S" by blast
      from ur1 ur2 
      have proc: "monoRedPairUrProc S NS (set ur) ?us P R (computeRuleMap R) rP rR"
        by (simp only: mono subset R rR P rP, auto)
      from redpair interpret mono_ce_af_redpair S NS ?af .
      show ?thesis using monoRedPairUrProc_ci proc ci_subsetI[OF subset] mono unfolding ci_proc_impl_def by simp
    qed
  qed
qed
end


end
